<?xml version="1.0" encoding="euc-jp"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:fs="http://www-tsujii.is.s.u-tokyo.ac.jp/lilfes"
                xmlns="http://www.w3.org/1999/xhtml"
                xmlns:html="http://www.w3.org/1999/xhtml"
                version="1.0">

<xsl:comment>
######################################################################
Copyright (c) 2005, MIYAO Yusuke
You may distribute this file under the terms of the Artistic License.

Name: fs.xsl
Author: MIYAO Yusuke (yusuke@is.s.u-tokyo.ac.jp)
Stylesheet for browsing feature structures and trees
######################################################################
</xsl:comment>

<xsl:comment>
############################################################
##
##  Conversion of feature structures
##
############################################################
</xsl:comment>

<xsl:template match="fs:fs">
  <xsl:choose>
  <xsl:when test="count(*)&lt;2">
    <!-- atom -->
    <xsl:apply-templates />  
  </xsl:when>
  <xsl:when test="count(*)=3 and fs:type[position()=1]/text()='cons' and
                  count(fs:feat[@edge='hd'])=1 and count(fs:feat[@edge='tl'])=1">
    <!-- list -->
    <table class="list">
      <tr>
        <td>&lt;</td>
        <xsl:apply-templates select="." mode="list" />
        <td>&gt;</td>
      </tr>
    </table>
  </xsl:when>
  <xsl:otherwise>
    <!-- complex feature structures -->
    <xsl:variable name="num_dtrs" select="count(*)" />
    <table class="fs">
      <xsl:for-each select="*">
        <tr>
          <xsl:if test="position()=1">
            <td class="lprn">
              <xsl:attribute name="rowspan">
                <xsl:value-of select="$num_dtrs"/>
              </xsl:attribute>
            </td>
          </xsl:if>
          <xsl:apply-templates select="." mode="fs" />
          <xsl:if test="position()=1">
            <td class="rprn">
              <xsl:attribute name="rowspan">
                <xsl:value-of select="$num_dtrs"/>
              </xsl:attribute>
            </td>
          </xsl:if>
        </tr>
      </xsl:for-each>
    </table>
  </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<xsl:template match="fs:fs" mode="shared_id">
  <xsl:variable name="shared_id" select="@shared_id" />
  <xsl:if test="$shared_id != ''">
    <xsl:choose>
    <xsl:when test="node()"> 
      <span class="shared_id">
        <a>
          <xsl:attribute name="name">shared_id<xsl:value-of select="$shared_id" /></xsl:attribute>
          <xsl:value-of select="$shared_id" />
        </a>
      </span>
    </xsl:when>
    <xsl:otherwise>
      <span class="shared_id">
        <a>
          <xsl:attribute name="href">#shared_id<xsl:value-of select="$shared_id" /></xsl:attribute>
          <xsl:value-of select="$shared_id" />
        </a>
      </span>
    </xsl:otherwise>
    </xsl:choose>
  </xsl:if>
</xsl:template>

<xsl:template match="fs:fs" mode="list">
  <xsl:for-each select="fs:feat[@edge='hd']">
    <!-- head of list -->
    <xsl:choose>
    <xsl:when test="fs:fs/@shared_id!=''">
      <td><xsl:apply-templates select="fs:fs" mode="shared_id" /></td>
      <td><xsl:apply-templates select="fs:fs" /></td>
    </xsl:when>
    <xsl:otherwise>
      <td><xsl:apply-templates select="*" /></td>
    </xsl:otherwise>
    </xsl:choose>
  </xsl:for-each>
  <xsl:for-each select="fs:feat[@edge='tl']">
    <!-- tail of list -->
    <xsl:choose>
    <xsl:when test="fs:fs/@shared_id!=''">
      <!-- tail is shared: print as ordinary feature structures -->
      <td>|<xsl:apply-templates select="*" mode="shared_id" /></td>
      <td><xsl:apply-templates select="." /></td>
    </xsl:when>
    <xsl:when test="count(*)=1 and fs:type/text()='nil'">
      <!-- nil: do nothing -->
    </xsl:when>
    <xsl:when test="count(fs:fs/*)=3 and fs:fs/fs:type[position()=1]/text()='cons' and
		    count(fs:fs/fs:feat[@edge='hd'])=1 and
		    count(fs:fs/fs:feat[@edge='tl'])=1">
      <!-- list elements -->
      <td>,</td>
      <td><xsl:apply-templates select="*" mode="list" /></td>
    </xsl:when>
    <xsl:otherwise>
      <!-- tail is not specified -->
      <td>|</td>
      <td><xsl:apply-templates select="." /></td>
    </xsl:otherwise>
    </xsl:choose>
  </xsl:for-each>
</xsl:template>

<xsl:template match="fs:feat" mode="fs">
  <td>
<!--
  <span class="edge_fs"><xsl:value-of select="@edge" /></span>
  <xsl:if test="./fs:fs/@shared_id != ''">
    <span class="shared_id"><xsl:value-of select="./fs:fs/@shared_id" /></span>
-->
  <span class="edge_fs"><xsl:value-of select="@edge" /></span>
  <xsl:apply-templates select="fs:fs" mode="shared_id" />
  </td>
  <td>
    <xsl:apply-templates />
  </td>
</xsl:template>

<xsl:template match="*" mode="fs">
  <td colspan="2">
    <xsl:apply-templates select="." />
  </td>
</xsl:template>

<xsl:template match="fs:type">
  <span class="type">
    <xsl:apply-templates />
  </span>
</xsl:template>

<xsl:template match="fs:decimal">
  <span class="decimal">
    <xsl:apply-templates />
  </span>
</xsl:template>

<xsl:template match="fs:float">
  <span class="float">
    <xsl:apply-templates />
  </span>
</xsl:template>

<xsl:template match="fs:doublequoted">
  <span class="doublequoted">
    <xsl:apply-templates />
  </span>
</xsl:template>

<xsl:template match="fs:singlequoted">
  <span class="singlequoted">
    <xsl:apply-templates />
  </span>
</xsl:template>

<xsl:template match="html:*">
  <xsl:copy>
  <xsl:copy-of select="@*"/> 
  <xsl:apply-templates />
  </xsl:copy>
</xsl:template>

<xsl:comment>
############################################################
##
##  Conversion of tree structures
##
############################################################
</xsl:comment>

<!-- Root node of a tree -->
<xsl:template match="fs:tree">
  <table class="tree">
    <xsl:call-template name="draw_tree" />
  </table>
</xsl:template>

<!-- daughter node -->
<xsl:template name="draw_dtr">
  <xsl:param name="line1">space</xsl:param>
  <xsl:param name="line2">space</xsl:param>
  <xsl:variable name="edge_label" select="@edge" />
  <xsl:variable name="num_trees" select="count(fs:tree)" />
  <xsl:choose>
  <!-- no tree node -->
  <xsl:when test="$num_trees = 0">
    <tr>
      <td>
        <xsl:attribute name="class"><xsl:value-of select="$line1" /></xsl:attribute>
      </td>
      <td>
        <xsl:attribute name="class"><xsl:value-of select="$line2" /></xsl:attribute>
        <span class="edge_tree"><xsl:value-of select="$edge_label" /></span>
      </td>
    </tr>
    <tr>
      <td class="term" colspan="2">
        <div class="term">
          <xsl:apply-templates />
        </div>
      </td>
    </tr>
  </xsl:when>
  <!-- tree node -->
  <xsl:otherwise>
    <xsl:for-each select="fs:tree">
      <xsl:call-template name="draw_tree">
        <xsl:with-param name="line1" select="$line1" />
        <xsl:with-param name="line2" select="$line2" />
        <xsl:with-param name="edge_label" select="$edge_label" />
      </xsl:call-template>
    </xsl:for-each>
  </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!-- tree node -->
<xsl:template name="draw_tree">
  <xsl:param name="line1">space</xsl:param>
  <xsl:param name="line2">space</xsl:param>
  <xsl:param name="edge_label"></xsl:param>
  <xsl:variable name="num_dtrs" select="count(fs:dtr)" />
  <xsl:choose>
  <!-- no daughter nodes -->
  <xsl:when test="$num_dtrs = 0">
    <tr>
      <td>
        <xsl:attribute name="class"><xsl:value-of select="$line1" /></xsl:attribute>
      </td>
      <td>
        <xsl:attribute name="class"><xsl:value-of select="$line2" /></xsl:attribute>
        <span class="edge_tree"><xsl:value-of select="$edge_label" /></span>
      </td>
    </tr>
    <tr>
      <td class="term" colspan="2">
        <div class="term">
          <xsl:apply-templates />
        </div>
      </td>
    </tr>
  </xsl:when>
  <!-- unary tree -->
  <xsl:when test="$num_dtrs = 1">
    <xsl:variable name="n">
      <xsl:call-template name="count_daughters">
        <xsl:with-param name="target" select="." />
      </xsl:call-template>
    </xsl:variable>
    <tr>
      <td>
        <xsl:attribute name="colspan"><xsl:value-of select="$n" /></xsl:attribute>
        <xsl:attribute name="class"><xsl:value-of select="$line1" /></xsl:attribute>
      </td>
      <td>
        <xsl:attribute name="colspan"><xsl:value-of select="$n" /></xsl:attribute>
        <xsl:attribute name="class"><xsl:value-of select="$line2" /></xsl:attribute>
        <span class="edge_tree"><xsl:value-of select="$edge_label" /></span>
      </td>
    </tr>
    <tr>
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$n" /></xsl:attribute>
      </td>
      <td class="node">
        <xsl:attribute name="colspan"><xsl:value-of select="$n" /></xsl:attribute>
        <div class="node">
          <xsl:value-of select="./text()" /><xsl:apply-templates select="./html:*" />
        </div>
      </td>
    </tr>
    <tr>
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$n" /></xsl:attribute>
      </td>
      <td class="right">
        <xsl:attribute name="colspan"><xsl:value-of select="$n" /></xsl:attribute>
      </td>
    </tr>
    <xsl:for-each select="fs:dtr">
      <xsl:call-template name="draw_dtr">
        <xsl:with-param name="line1">space</xsl:with-param>
        <xsl:with-param name="line2">right</xsl:with-param>
      </xsl:call-template>
    </xsl:for-each>
  </xsl:when>
  <!-- n-ary (n>1) tree -->
  <xsl:otherwise>
    <tr>
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$num_dtrs" /></xsl:attribute>
        <xsl:attribute name="class"><xsl:value-of select="$line1" /></xsl:attribute>
      </td>
      <td>
        <xsl:attribute name="colspan"><xsl:value-of select="$num_dtrs" /></xsl:attribute>
        <xsl:attribute name="class"><xsl:value-of select="$line2" /></xsl:attribute>
        <span class="edge_tree"><xsl:value-of select="$edge_label" /></span>
      </td>
<!--
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$num_dtrs - 1" /></xsl:attribute>
      </td>
-->
    </tr>
    <tr>
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$num_dtrs" /></xsl:attribute>
      </td>
      <td class="node">
        <xsl:attribute name="colspan"><xsl:value-of select="$num_dtrs" /></xsl:attribute>
        <div class="node">
          <xsl:value-of select="./text()" /><xsl:apply-templates select="./html:*" />
        </div>
      </td>
<!--
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$num_dtrs - 1" /></xsl:attribute>
      </td>
-->
    </tr>
    <tr>
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$num_dtrs" /></xsl:attribute>
      </td>
      <td class="right"></td>
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$num_dtrs - 1" /></xsl:attribute>
      </td>
    </tr>
    <tr>
      <xsl:for-each select="fs:dtr">
        <xsl:variable name="dtr_id">
          <xsl:number level="single" count="fs:dtr" format="1" />
        </xsl:variable>
        <td class="dtr" colspan="2">
          <table class="tree">
            <xsl:choose>
            <xsl:when test="$dtr_id = 1">
              <xsl:call-template name="draw_dtr">
                <xsl:with-param name="line1">space</xsl:with-param>
                <xsl:with-param name="line2">left</xsl:with-param>
              </xsl:call-template>
            </xsl:when>
            <xsl:when test="$dtr_id = $num_dtrs">
              <xsl:call-template name="draw_dtr">
                <xsl:with-param name="line1">top</xsl:with-param>
                <xsl:with-param name="line2">right</xsl:with-param>
              </xsl:call-template>
            </xsl:when>
            <xsl:otherwise>
              <xsl:call-template name="draw_dtr">
                <xsl:with-param name="line1">top</xsl:with-param>
                <xsl:with-param name="line2">left</xsl:with-param>
              </xsl:call-template>
            </xsl:otherwise>
            </xsl:choose>
          </table>
        </td>
      </xsl:for-each>
    </tr>
  </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!-- count the number of non-unary daughters -->
<xsl:template name="count_daughters">
  <xsl:param name="target">.</xsl:param>
  <xsl:variable name="num" select="count($target/fs:dtr)" />
  <xsl:choose>
  <xsl:when test="$num = 0">
    1
  </xsl:when>
  <xsl:when test="$num = 1">
    <xsl:call-template name="count_daughters">
      <xsl:with-param name="target" select="$target/fs:dtr/fs:tree" />
    </xsl:call-template>
  </xsl:when>
  <xsl:otherwise>
    <xsl:value-of select="$num" />
  </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!--
<xsl:template match="fs:tree">
<xsl:variable name="width">
<xsl:call-template name="rec_dom_leaves_count">
  <xsl:with-param name="x" select="./fs:dtr" />
</xsl:call-template>
</xsl:variable>
<xsl:variable name="depth">
<xsl:call-template name="dtrs_depth">
  <xsl:with-param name="ts" select="./fs:dtr" />
</xsl:call-template>
</xsl:variable>

դο<xsl:value-of select="$width" /> words<br/>
ڤο<xsl:value-of select="$depth" />

<table class="tree" border="1" cellspacing="0">
  <xsl:call-template name="recursive_draw">
    <xsl:with-param name="n"><xsl:value-of select="$depth" /></xsl:with-param>
    <xsl:with-param name="d"><xsl:value-of select="$depth" /></xsl:with-param>
    <xsl:with-param name="t" select="." />
  </xsl:call-template>
</table>

</xsl:template>

<xsl:template name="recursive_draw">
  <xsl:param name="n" />
  <xsl:param name="d" />
  <xsl:param name="t" />
  <xsl:comment>
    ###################################
      input: fs:tree (constant: top of tree)
      1  n  d
    ###################################
  </xsl:comment>
  <xsl:choose>
    <xsl:when test="$n &lt; 1"></xsl:when>
    <xsl:otherwise>
      <xsl:if test="$n != $d">
        <tr height="16">
          <xsl:call-template name="in_rec_draw_line">
            <xsl:with-param name="t" select="$t" />
            <xsl:with-param name="n"><xsl:value-of select="$n" /></xsl:with-param>
            <xsl:with-param name="d"><xsl:value-of select="$d" /></xsl:with-param>
            <xsl:with-param name="c"><xsl:value-of select="$d" /></xsl:with-param>
          </xsl:call-template>
        </tr>
      </xsl:if>
      <tr>
        <xsl:call-template name="in_rec_draw">
          <xsl:with-param name="t" select="$t" />
          <xsl:with-param name="n"><xsl:value-of select="$n" /></xsl:with-param>
          <xsl:with-param name="d"><xsl:value-of select="$d" /></xsl:with-param>
        </xsl:call-template>
      </tr>
      <xsl:if test="$n &gt; 1">
      <tr height="16">
        <xsl:call-template name="in_rec_draw_line3">
          <xsl:with-param name="t" select="$t" />
          <xsl:with-param name="n"><xsl:value-of select="$n" /></xsl:with-param>
          <xsl:with-param name="d"><xsl:value-of select="$d" /></xsl:with-param>
          <xsl:with-param name="c"><xsl:value-of select="$d" /></xsl:with-param>
        </xsl:call-template>
      </tr>
      <tr height="1">
        <xsl:call-template name="in_rec_draw_line2">
          <xsl:with-param name="t" select="$t" />
          <xsl:with-param name="n"><xsl:value-of select="$n" /></xsl:with-param>
          <xsl:with-param name="d"><xsl:value-of select="$d" /></xsl:with-param>
        </xsl:call-template>
      </tr>
      </xsl:if>
      <xsl:call-template name="recursive_draw">
        <xsl:with-param name="n"><xsl:value-of select="$n - 1" /></xsl:with-param>
        <xsl:with-param name="d"><xsl:value-of select="$d" /></xsl:with-param>
        <xsl:with-param name="t" select="$t" />
      </xsl:call-template>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<xsl:template name="in_rec_draw">
  <xsl:param name="t" />
  <xsl:param name="n" />
  <xsl:param name="d" />
<xsl:comment>
#########################
  draw node
  bug that "leaves are not shown"
#########################
</xsl:comment>

  <xsl:choose>
    <xsl:when test="$d = $n">
      <xsl:variable name="lc">
        <xsl:call-template name="rec_dom_leaves_count">
          <xsl:with-param name="x" select="$t/fs:dtr" />
        </xsl:call-template>
      </xsl:variable>

      <xsl:variable name="a"><xsl:call-template name="rec_dom_leaves_count">
        <xsl:with-param name="x" select="$t/fs:dtr[1]"/>
      </xsl:call-template></xsl:variable>
      <xsl:variable name="b"><xsl:call-template name="rec_dom_leaves_count">
        <xsl:with-param name="x" select="$t/fs:dtr[last()]"/>
      </xsl:call-template></xsl:variable>
      
      <xsl:variable name="cols"><xsl:value-of select="$lc * 2 - $a - $b" /></xsl:variable>
      <xsl:variable name="cols1">
        <xsl:value-of select="floor($cols div 2)" />
      </xsl:variable>
      <xsl:variable name="cols2"><xsl:value-of select="$lc - $cols1" /></xsl:variable>

      <xsl:variable name="rows">
        <xsl:choose>
          <xsl:when test="$t/fs:dtr/fs:tree">
            <xsl:value-of select="($d - $n) * 4 + 1" />
          </xsl:when>
          <xsl:otherwise>1</xsl:otherwise>
        </xsl:choose>
      </xsl:variable>

      <xsl:choose>
      <xsl:when test="$cols &gt; 0">
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$a" /></xsl:attribute>
        <xsl:attribute name="rowspan"><xsl:value-of select="$rows" /></xsl:attribute>
        <img height="1" />
      </td>
      <xsl:if test="$cols1 &gt; 1">
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$cols1 - 1" /></xsl:attribute>
        <xsl:attribute name="rowspan"><xsl:value-of select="$rows" /></xsl:attribute>
        <img height="1" />
      </td>
      </xsl:if>
      <td class="node" colspan="2">
        <xsl:attribute name="rowspan"><xsl:value-of select="$rows" /></xsl:attribute>
        <xsl:value-of select="$t/text()" /><xsl:apply-templates select="$t/html:*" />
      </td>
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$cols2 - 1 + $b" /></xsl:attribute>
        <xsl:attribute name="rowspan"><xsl:value-of select="$rows" /></xsl:attribute>
        <img height="1" />
      </td>
      </xsl:when>
      <xsl:otherwise>
      <td class="node" colspan = "2">
        <xsl:attribute name="rowspan"><xsl:value-of select="$rows" /></xsl:attribute>
        <xsl:value-of select="$t/text()" /><xsl:apply-templates select="$t/html:*" />
      </td>
      </xsl:otherwise>
      </xsl:choose>

    </xsl:when>
    
    <xsl:when test="$t/*/fs:tree">
      <xsl:for-each select="$t/fs:dtr">
        <xsl:choose>
        <xsl:when test="./fs:tree">
        <xsl:call-template name="in_rec_draw">
          <xsl:with-param name="t" select="./fs:tree" />
          <xsl:with-param name="n"><xsl:value-of select="$n + 1" /></xsl:with-param>
          <xsl:with-param name="d"><xsl:value-of select="$d" /></xsl:with-param>
        </xsl:call-template>
        </xsl:when>
        <xsl:when test="$n + 1 = $d">
          <xsl:variable name="rows">
            <xsl:value-of select="($d - $n) * 4 + 1" />
          </xsl:variable>
          <td colspan="2" class="leaf">
            <xsl:attribute name="rowspan"><xsl:value-of select="$rows" /></xsl:attribute>
            <xsl:value-of select="." />
          </td>
        </xsl:when>
        </xsl:choose>
      </xsl:for-each>
    </xsl:when>
    <xsl:when test="$d - $n = 1">
      <xsl:for-each select="$t/fs:dtr">
        <xsl:variable name="rows">
          <xsl:value-of select="($d - $n) * 4 + 1" />
        </xsl:variable>
        <td colspan="2" class="leaf">
        <xsl:attribute name="rowspan"><xsl:value-of select="$rows" /></xsl:attribute>
        <xsl:apply-templates /></td>
      </xsl:for-each>
    </xsl:when>
    <xsl:otherwise>
    </xsl:otherwise>



  </xsl:choose>
</xsl:template>

<xsl:template name="in_rec_draw_line">
  <xsl:param name="t" />
  <xsl:param name="n" />
  <xsl:param name="d" />
  <xsl:param name="c" />
  <xsl:comment>
    ###################################
      input: fs:tree (constant: top of tree)
      t: top of tree
      n: ȤƤڤι⤵ɸ
      d: Τι⤵
      c: Ƶѿ; c == n ΤȤc = d1ޤǸ餹
    ###################################
  </xsl:comment>
<xsl:comment>(above)</xsl:comment>
  <xsl:choose>
    <xsl:when test="$d = $n">
      <xsl:variable name="left">
        <xsl:call-template name="rec_dom_leaves_count">
            <xsl:with-param name="x" select="$t/fs:dtr" />
        </xsl:call-template>
      </xsl:variable>

      <xsl:variable name="a"><xsl:call-template name="rec_dom_leaves_count">
        <xsl:with-param name="x" select="$t/fs:dtr[1]"/>
      </xsl:call-template></xsl:variable>
      <xsl:variable name="b"><xsl:call-template name="rec_dom_leaves_count">
        <xsl:with-param name="x" select="$t/fs:dtr[last()]"/>
      </xsl:call-template></xsl:variable>
      <xsl:variable name="lc"><xsl:value-of select="$left * 2 - $a - $b" /></xsl:variable>
      <xsl:variable name="cols1">
        <xsl:value-of select="floor($lc div 2)" />
      </xsl:variable>
      <xsl:variable name="cols2"><xsl:value-of select="$lc - $cols1" /></xsl:variable>

      <td class="right">
        <xsl:attribute name="colspan"><xsl:value-of select="$a + $cols1" /></xsl:attribute>
        <xsl:attribute name="style">
        <xsl:if test="@edge">
        width: 40px;
        </xsl:if>
        </xsl:attribute>
        <img height="1" />
      </td>

      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$cols2 + $b" /></xsl:attribute>
        <img height="1" />
        <span class="edge_tree"><xsl:value-of select="@edge" /></span>
      </td>
      
    </xsl:when>
    <xsl:when test="$t/*/fs:tree">
      <xsl:for-each select="$t/fs:dtr">
        <xsl:choose>
        <xsl:when test="./fs:tree">
        <xsl:call-template name="in_rec_draw_line">
          <xsl:with-param name="t" select="./fs:tree" />
          <xsl:with-param name="n"><xsl:value-of select="$n + 1" /></xsl:with-param>
          <xsl:with-param name="d"><xsl:value-of select="$d" /></xsl:with-param>
          <xsl:with-param name="c"><xsl:value-of select="$c - 1" /></xsl:with-param>
        </xsl:call-template>
        </xsl:when>
        <xsl:when test="$n + 1 = $d">
          <td class="right"><img width="1" /></td><td class="space"><img width="1" /></td>
        </xsl:when>
        </xsl:choose>
      </xsl:for-each>
    </xsl:when>
    <xsl:otherwise>
      <xsl:for-each select="$t/fs:dtr">
        <xsl:if test="$n + 1 = $d">
        <td class="right">
        <xsl:attribute name="style">
        <xsl:if test="@edge">
        width: 40px;
        </xsl:if>
        </xsl:attribute>
        <img height="1" />
        </td>
        <td class="space"><img height="1" />
          <span class="edge_tree"><xsl:value-of select="@edge" /></span>
        </td>
        </xsl:if>
      </xsl:for-each>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<xsl:template name="in_rec_draw_line3">
  <xsl:param name="t" />
  <xsl:param name="n" />
  <xsl:param name="d" />
  <xsl:param name="c" />
  <xsl:comment>
    ###################################
      input: fs:tree (constant: top of tree)
      t: top of tree
      n: ȤƤڤι⤵ɸ
      d: Τι⤵
      Ρɤβin_rec_draw_lineȤΰ㤤Ͼ
    ###################################
  </xsl:comment>
<xsl:comment></xsl:comment>
  <xsl:choose>
    <xsl:when test="$d = $n">
      <xsl:variable name="left">
        <xsl:call-template name="rec_dom_leaves_count">
            <xsl:with-param name="x" select="$t/fs:dtr" />
        </xsl:call-template>
      </xsl:variable>

      <xsl:variable name="a"><xsl:call-template name="rec_dom_leaves_count">
        <xsl:with-param name="x" select="$t/fs:dtr[1]"/>
      </xsl:call-template></xsl:variable>
      <xsl:variable name="b"><xsl:call-template name="rec_dom_leaves_count">
        <xsl:with-param name="x" select="$t/fs:dtr[last()]"/>
      </xsl:call-template></xsl:variable>
      
      <xsl:variable name="lc"><xsl:value-of select="$left * 2 - $a - $b" /></xsl:variable>
      <xsl:variable name="cols1">
        <xsl:value-of select="floor($lc div 2)" />
      </xsl:variable>
      <xsl:variable name="cols2"><xsl:value-of select="$lc - $cols1" /></xsl:variable>

      <td class="right">
        <xsl:attribute name="colspan"><xsl:value-of select="$a + $cols1" /></xsl:attribute>
        <img height="1" />
      </td>
      <td class="space">
        <xsl:attribute name="colspan"><xsl:value-of select="$cols2 + $b" /></xsl:attribute>
        <img height="1" />
      </td>
    </xsl:when>
    <xsl:when test="$t/*/fs:tree">
      <xsl:for-each select="$t/fs:dtr">
        <xsl:choose>
        <xsl:when test="./fs:tree">
        <xsl:call-template name="in_rec_draw_line3">
          <xsl:with-param name="t" select="./fs:tree" />
          <xsl:with-param name="n"><xsl:value-of select="$n + 1" /></xsl:with-param>
          <xsl:with-param name="d"><xsl:value-of select="$d" /></xsl:with-param>
        </xsl:call-template>
        </xsl:when>
        </xsl:choose>
      </xsl:for-each>
    </xsl:when>
  </xsl:choose>
</xsl:template>

<xsl:template name="in_rec_draw_line2">
  <xsl:param name="t" />
  <xsl:param name="n" />
  <xsl:param name="d" />
<xsl:comment>襳</xsl:comment>
  <xsl:comment>
    ###################################
      input: fs:tree (constant: top of tree)
    ###################################
  </xsl:comment>
  <xsl:choose>
    <xsl:when test="$d = $n">
      <xsl:variable name="lc">
        <xsl:call-template name="rec_dom_leaves_count">
            <xsl:with-param name="x" select="$t/fs:dtr" />
        </xsl:call-template>
      </xsl:variable>

      <xsl:variable name="wid">
        <xsl:call-template name="right_space">
          <xsl:with-param name="t" select="$t" />
        </xsl:call-template>
      </xsl:variable>

      <xsl:variable name="a"><xsl:call-template name="rec_dom_leaves_count">
        <xsl:with-param name="x" select="$t/fs:dtr[1]"/>
      </xsl:call-template></xsl:variable>
      <xsl:variable name="b"><xsl:call-template name="rec_dom_leaves_count">
        <xsl:with-param name="x" select="$t/fs:dtr[last()]"/>
      </xsl:call-template></xsl:variable>
      
    <td class="space">
      <xsl:attribute name="colspan"><xsl:value-of select="$a" /></xsl:attribute>
    </td>
      <xsl:variable name="left"><xsl:value-of select="$lc * 2 - $a - $wid" /></xsl:variable>
      <xsl:if test="$left &gt; 0">
        <td class="top">
        <xsl:attribute name="colspan"><xsl:value-of select="$left" /></xsl:attribute>
        <img height="1" /></td>
      </xsl:if>
    <td class="space">
      <xsl:attribute name="colspan"><xsl:value-of select="$wid" /></xsl:attribute>
      <img height="1" />
    </td>
    </xsl:when>
    <xsl:when test="$t/fs:dtr/fs:tree">
      <xsl:for-each select="$t/fs:dtr/fs:tree">
        <xsl:call-template name="in_rec_draw_line2">
          <xsl:with-param name="t" select="." />
          <xsl:with-param name="n"><xsl:value-of select="$n + 1" /></xsl:with-param>
          <xsl:with-param name="d"><xsl:value-of select="$d" /></xsl:with-param>
        </xsl:call-template>
      </xsl:for-each>
    </xsl:when>

  </xsl:choose>
</xsl:template>


<xsl:template match="html:br"><br/></xsl:template>

<xsl:template match="html:*"><xsl:copy><xsl:copy-of select="@*"/><xsl:apply-templates /></xsl:copy></xsl:template>




<xsl:template name="dom_leaves_count">
<xsl:param name="x" />
<xsl:comment>
  ##############################################
    fs:dtr ۤleavesο򥫥Ȥ
    dtr./fs:treeäƤкƵŪdom_leaves_countƤӽФؿrec_~Ƥ
    äƤʤLeafȤ"1"򤫤
  ##############################################
</xsl:comment>

<xsl:choose>
  <xsl:when test="$x/fs:tree">
    <xsl:call-template name="rec_dom_leaves_count">
      <xsl:with-param name="x" select="$x/fs:tree/fs:dtr" />
    </xsl:call-template>
  </xsl:when>
  <xsl:otherwise>
    <xsl:value-of select="1" />
  </xsl:otherwise>
</xsl:choose>
</xsl:template>

<xsl:template name="rec_dom_leaves_count">
<xsl:param name="x"/>
<xsl:comment>
  ##############################################
    fs:tree ľΤ٤ƤdtrФdom_leaves_count׻
    ­碌ؿ
    ϡdtr(s)
  ##############################################
</xsl:comment>
<xsl:choose>
  <xsl:when test="$x">
    <xsl:variable name="t">
      <xsl:call-template name="dom_leaves_count">
        <xsl:with-param name="x" select="$x[1]" />
      </xsl:call-template>
    </xsl:variable>
    <xsl:variable name="rec_result">
      <xsl:call-template name="rec_dom_leaves_count">
        <xsl:with-param name="x" select="$x[position() &gt; 1]" />
      </xsl:call-template>
    </xsl:variable>
    <xsl:value-of select="$rec_result + $t" />
  </xsl:when>
  <xsl:otherwise><xsl:value-of select="0" /></xsl:otherwise>
</xsl:choose>
</xsl:template>

<xsl:template name="dtrs_depth">
<xsl:param name="ts" />
<xsl:comment>
  ##############################################
    dtrs_depth: fs:dtr(s)οmax򤫤
    
  ##############################################
</xsl:comment>
  <xsl:choose>
    <xsl:when test="$ts">
      <xsl:variable name="x">
        <xsl:call-template name="dtr_depth">
          <xsl:with-param name="t" select="$ts[1]" />
        </xsl:call-template>
      </xsl:variable>
      <xsl:variable name="y">
        <xsl:call-template name="dtrs_depth">
          <xsl:with-param name="ts" select="$ts[position() &gt; 1]" />
        </xsl:call-template>
      </xsl:variable>
      <xsl:choose>
        <xsl:when test="$x &gt; $y">
          <xsl:value-of select="$x" />
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="$y" />
        </xsl:otherwise>
      </xsl:choose>
    </xsl:when>
    <xsl:otherwise><xsl:value-of select="0" /></xsl:otherwise>
  </xsl:choose>
</xsl:template>

<xsl:template name="dtr_depth">
  <xsl:param name="t" />
  <xsl:comment>
  ##############################################
    leaf nodeǤϹ⤵Ȥƣ֤Ƥ롣ϡdepthǷ׻
    ι⤵ȤΤʬθǤ롣
    Ȥ<fs:tree>NNP<fs:dtr>Japan</fs:dtr></fs:tree> Ȥɤʤ뤫
    狼ޤ󡣤֤⤵֤ƤȻפޤ
  ##############################################
  </xsl:comment>
  <xsl:choose>
    <xsl:when test="$t/fs:tree">
      <xsl:variable name="rec_result">
      <xsl:call-template name="dtrs_depth">
        <xsl:with-param name="ts" select="$t/fs:tree/fs:dtr" />
      </xsl:call-template>
      </xsl:variable>
      <xsl:value-of select="$rec_result + 1" />
    </xsl:when>
    <xsl:otherwise>
      <xsl:value-of select="2" />
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<xsl:template name="right_space">
  <xsl:param name="t" />
<xsl:comment>
  ###############################
  α¦ζοؿ
  ϡfs:tree
  ###############################
</xsl:comment>
  <xsl:choose>
    <xsl:when test="$t/fs:dtr[position() = last()]/fs:tree">
      <xsl:variable name="r">
        <xsl:call-template name="right_space">
          <xsl:with-param name="t" select="$t/fs:dtr[position() = last()]/fs:tree" />
        </xsl:call-template>
      </xsl:variable>
      <xsl:variable name="w">
        <xsl:call-template name="rec_dom_leaves_count">
          <xsl:with-param name="x" select="$t/fs:dtr[last()]/fs:tree/fs:dtr" />
        </xsl:call-template>
      </xsl:variable>
      <xsl:variable name="a">
        <xsl:call-template name="dom_leaves_count">
          <xsl:with-param name="x" select="$t/fs:dtr[last()]/fs:tree/fs:dtr[1]" />
        </xsl:call-template>
      </xsl:variable>
      <xsl:variable name="b">
        <xsl:call-template name="dom_leaves_count">
          <xsl:with-param name="x" select="$t/fs:dtr[last()]/fs:tree/fs:dtr[last()]" />
        </xsl:call-template>
      </xsl:variable>
      <xsl:value-of select="$r + ceiling(($w * 2 - $a - $b) div 2)" />
    </xsl:when>
    <xsl:otherwise>
      <xsl:value-of select="1" />
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>
-->

</xsl:stylesheet>
