/* $Id: errors.h,v 1.8 2005/02/12 07:11:54 yusuke Exp $
 *
 *    Copyright (c) 1997-1998, Makino Takaki
 *
 *    You may distribute this file under the terms of the Artistic License.
 *
 */
//
///  <ID>$Id: errors.h,v 1.8 2005/02/12 07:11:54 yusuke Exp $</ID>
//
///  <COLLECTION> lilfes-bytecode </COLLECTION>
//
///  <name>errors.h</name>
//
///  <overview>
///  <jpn>顼˴ؤ롼</jpn>
///  <eng>errors </eng>
///  </overview>

//
//  LightSpeed HPSG Parser Project
//
//  errors.h - Definition of error exception handling.
//
//  **LICENSE**
//
//  Change Log:
//
//  96/11/20  Created. / mak
/// <notiondef>
/// <name> errors_overview </name>
/// <title> 顼롼 </title>
/// 
/// <desc>
/// 
/// <p> <ref>errors.h</ref> ˤϡ顼ȯ˥顼ɽ뤿Υ饤֥꤬äƤޤ<br>
/// errorstream ϥ顼ȯ˥顼ƤϤޤ<br>
/// tracestream ϥǥХåѤνϤǡ̾ϤʤˤϤʤ褦ˤʤäƤޤDEBUG ơѿ DEB  true äƤȡϤޤ
/// </p>
/// </desc>
/// <see> errorstream, tracestream </see>
/// </notiondef>

#ifndef __errors_h
#define __errors_h

#include "lconfig.h"

#include "basic.h" // for string type
#include "lexer.h" 
#include <iostream>

namespace lilfes {

#ifdef DEBUG
extern bool DEB;
#else
const bool DEB = false;
#endif

enum endmsg_t { endmsg };

extern std::istream* input_stream;
extern std::ostream* output_stream;
extern std::ostream* error_stream;

/// <classdef>
/// <name> tracestream </name>

/// <overview>饹</overview>
/// <desc>
/// <p>
/// tracestream ϥǥХåѤνϤǡ̾ϤʤˤϤʤ褦ˤʤäƤޤDEBUG ơѿ DEB  true äƤȡϤޤ
/// </p>
/// </desc>
/// <see></see>

/// <body>

class tracestream
{
	std::string topline;
	bool show_topline;

public:
	tracestream(std::string itopline = "") : topline(itopline), show_topline(true) { }

	template <class T>
	tracestream& operator<<(T &t) { 
			if( DEB ) {
				if( show_topline )
				{
					*output_stream << topline;
					show_topline = false;
				}
				*output_stream << t;
			}
			return *this;
		}
	template <class T>
	tracestream& operator<<(const T &t) { 
			if( DEB ) {
				if( show_topline )
				{
					*output_stream << topline;
					show_topline = false;
				}
				*output_stream << t;
			}
			return *this;
		}
	void operator<< (const endmsg_t &) { 
			if( DEB ) {
				*output_stream << std::endl;
				show_topline = true;
			}
		}
};
/// </body></classdef>

tracestream &tracer();

#define TRACE(x) (lilfes::tracer() << x << lilfes::endmsg)

/// <classdef>
/// <name> errorstream </name>

/// <overview>饹</overview>
/// <desc>
/// <p><br>
/// <br>
/// errorstream errout("Error: ", " at somewhere"); <br>
/// errout &lt;&lt; "Undefined something '" << name << "'" << endmsg; <br>
/// Ƚ񤯤ȡ̤ <br>
/// Error: Undefined something 'hogehoge' at somewhere\n <br>
/// ɽޤ<br>
/// </p>
/// </desc>
/// <see></see>

/// <body>

class errorstream
{

  // 顼åcerrǤϤʤstringstreamƤѿǤ
  //static std::ostream *error_ptr;

  std::string topline;
  std::string endline;
  bool show_topline;
  bool display;

public:
	errorstream(std::string itopline = "", std::string iendline = "", bool idisplay = true) 
		: topline(itopline), endline(iendline), show_topline(true), display(idisplay) { }

	errorstream sourceinfo( const char *file, int line );
  /// <var>file</var>Ϳե̾<var>line</var>ͿԿϤޤ
	template <class T>
	errorstream operator<<(const T &t) 
		{ 
			if( display )
			{
				if( show_topline )
				{
					(*error_stream) << topline;
					show_topline = false;
				}
				(*error_stream) << t;
			}
			return *this;
		}
	void operator<<(const endmsg_t &) 
		{
			if( display )
			{
				(*error_stream) << endline << std::endl;
				show_topline = true;
			}
		}
// 	static void SwitchErr(std::ostream *ptr) {
// 	  error_ptr = ptr;
// 	}
};
/// </body></classdef>

/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> runerr </name>

/// <overview>¹Ի顼ν</overview>
/// <desc>
/// <jpn>
/// <p>
/// ¹Ի顼Ϥޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <see>errorstream</see>

/// <body>
errorstream &runerr();
/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> runwarn </name>

/// <overview>¹Իٹν</overview>
/// <desc>
/// <jpn>
/// <p>
/// ¹ԻٹϤޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <see>errorstream</see>

/// <body>
errorstream &runwarn();
/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> inerr </name>

/// <overview>ɤ߹ߥ顼ν</overview>
/// <desc>
/// <jpn>
/// <p>
/// ɤ߹ߥ顼Ϥޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <see>errorstream</see>
errorstream inerr(lexer_t l);
/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> inwarn </name>

/// <overview>ɤ߹߷ٹν</overview>
/// <desc>
/// <jpn>
/// <p>
/// ɤ߹߷ٹϤޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <see>errorstream</see>
errorstream inwarn(lexer_t l);
/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> internalerr </name>

/// <overview>顼ν</overview>
/// <desc>
/// <jpn>
/// <p>
/// 顼Ϥޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <see>errorstream</see>
errorstream &internalerr();
/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////

/// <funcdef>
/// <name> internalwarn </name>

/// <overview>ٹν</overview>
/// <desc>
/// <jpn>
/// <p>
/// ٹϤޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <see>errorstream</see>
errorstream &internalwarn();
/// </body></funcdef>


#ifdef DOASSERT
#define runerr() runerr().sourceinfo(__FILE__, __LINE__)
#define runwarn() runwarn().sourceinfo(__FILE__, __LINE__)
#define inerr(l) inerr(l).sourceinfo(__FILE__, __LINE__)
#define inwarn(l) inwarn(l).sourceinfo(__FILE__, __LINE__)
#endif
#define internalerr() internalerr().sourceinfo(__FILE__, __LINE__)
#define internalwarn() internalwarn().sourceinfo(__FILE__, __LINE__)

// transitional macros (obsolete)
#define RUNERR(e)  (runerr() << e << endmsg)
#define RUNWARN(e) (runwarn() << e << endmsg)
#define INERR(m,e)   (inerr((m)->getLexer()) << e << endmsg)
#define INWARN(m,e)  (inwarn((m)->getLexer()) << e << endmsg)
#define ABORT(x)   do { internalerr() << x << endmsg; abort();} while(false)
#define DMSG(x)   (internalwarn() << x << endmsg)

#if defined(DOASSERT)
#define ASSERT(x) if( !(x) ) { internalerr() << "Assertion failed: (" #x ")" << endmsg; abort(); } else
#else
#define ASSERT(x)
#endif


extern long int PASSCOUNT;
extern long int TRAILBACK;
extern long int Errors, Warning;

} // namespace lilfes

#endif /* __errors_h */
