/* $Id: ftype.h,v 1.2 2004/02/08 06:35:51 mak Exp $
 *
 *    Copyright (c) 1996-1998, Makino Takaki
 *
 *    You may distribute this file under the terms of the Artistic License.
 *
 */
///  <ID>$Id: ftype.h,v 1.2 2004/02/08 06:35:51 mak Exp $</ID>
//
///  <COLLECTION> lilfes-bytecode </COLLECTION>
//
///  <name>ftype.h</name>
//
///  <overview>
///  <jpn>¤롼</jpn>
///  <eng>feature structure routine </eng>
///  </overview>


#ifndef __ftype_h
#define __ftype_h

namespace lilfes {

class type;
class feature;

} // namespace lilfes

#include "lconfig.h"

#include "profile.h"
#include "module.h"

namespace lilfes {

////////////////////////////////////////
// Constants

// TYPE_GROW_STEP
//    When the array type::children has no rest for a new child, the array 
//    is expanded with the width TYPE_GROW_STEP. Must be 2^n.
const int TYPE_GROW_STEP = 4;

// FEAT_GROW_STEP
//    When the array type::feats has no rest for a new feature, the array 
//    is expanded with the width FEAT_GROW_STEP. Must be 2^n.
const int FEAT_GROW_STEP = 4;

// FEATTABLE_GROW_BITS
//    When the array type::feats has no rest for a new feature, the array 
//    is expanded with the width 2^FEAT_GROW_BITS.
const int FEATTABLE_GROW_BITS = 8;

// TYPETABLE_GROW_BITS
//    When the array type::feats has no rest for a new feature, the array 
//    is expanded with the width 2^FEAT_GROW_BITS.
const int TYPETABLE_GROW_BITS = 8;

#include "errors.h"
#include "hash.h"
#include "lstring.h"

#define FEATURE_POSTFIX "\\"

//////////////////////////////////////////////////////////////////////////////
//
// Global variable definition
/// <vardef>
/// <name> bot </name>
/// <overview>bottom </overview>
/// <desc>
/// <p>
/// LiLFeSȤ߹ߤηǡƤηsupertypeȤʤ뷿Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>
extern type * bot;		// Global variable for the bottom
/// </body></vardef>

//////////////////////////////////////////////////////////////////////////////
//
//  class feature
//
/// <classdef>
/// <name> feature </name>
/// <overview> 饹</overview>
/// <desc>
/// <p>
/// 򤢤魯饹Ǥ
/// </p>
/// </desc>
/// <see>type</see>
/// <body>
class feature
{
	const type *reptype;
	int priority;		// Priority of the feature (the order of features)
	tserial basetype;	// Serial No. of the feature's base type
	fserial serialno;	// Serial No. of this feature

	lstring name;
	lstring propername;
	
	static fserial builtin_limit;
	static IntegerHash<feature> hash;			// Hash table for name -> feature
	static VarArray<const feature *> serialtable;
											// Table for serial -> feature

public:
		feature(const char *repname, module *mod, int ip=0, bool exported = true);
  /// <var>repname</var>̾<br>
  /// <var>mod</var>°⥸塼̾<br>
  /// <var>ip</var>initial priority<br>
  /// <var>exported</var>reptypeexportedɤ<br>
		feature(const type *reptype, int ip=0);
  /// <var>reptype</var>ɽ<br>
  /// <var>ip</var>initial priority<br>
		~feature();
		
	tserial GetKey() const;	// Used in class HashTable
    /// HashTableѤΥ᥽åɤǤȤname֤ޤ<ref>hash.h</ref>
	const char *GetSimpleName() const { return name; }
  /// ̾ޤ̾ʤΤǤϤޤ̾ȤGetProperName()ѤƤ
	const char *GetProperName() const { return propername; }
  /// ̾ޤ
	const char *GetPrintName() const;
  /// ɽѤ̾ޤ
	const char *GetName() const  { return GetProperName(); }
  /// ̾ޤ
	int GetPriority() const { return priority; }
  /// ʬ°뷿Ǥ̤ͥޤ
	fserial GetSerialNo() const { return serialno; }
  /// ꥢֹޤ
	tserial GetBaseTypeS() const { return basetype; }
  /// ١סʤǽ줿ˤΥꥢֹޤ
	const type *GetBaseType() const;
  /// ١סʤǽ줿ˤޤ
	const type *GetRepType() const;
  /// ɽסʤɽ뷿ˤޤ
	void SetBaseType(const type *t);
  /// type <var>t</var>١סʤǽ줿ˤꤷޤ
	bool IsBuiltIn() const { return serialno < builtin_limit; }
  /// Ȥ߹ߺѤߤΤΤɤĴ٤ޤȤ߹ߤΤΤʤtrue桼Τʤfalse֤ޤ
	void Delete();
  /// feature֥Ȥޤ
	friend int compfeature(const feature *a, const feature *b) 
		{ int x = a->priority - b->priority; return x ? x : strcmp(a->name, b->name); }
	friend int compfeature(const void *a, const void *b) { return compfeature(*(const feature **)a, *(const feature **)b); }
  /// Ĥfeature֥ȤΥ١ǤνӤޤŪˤpriorityνǡpriorityƱʤ鼭ˤʤޤ
	static const feature* Search(const type *t);
	static const feature* Search(tserial t) { return hash.Find(t); }
  /// reptype<var>t</var>бfeature֥Ȥõޤ
	static const feature* Serial(fserial ser) { ASSERT(ser < serialtable.GetSize()); return serialtable[ser]; }
  /// ꥢֹ椫бfeature֥Ȥޤ
//	static const feature* New   (char *n, const type *app = bot, int ip=0) 
//			{ const feature *r = hash.Find(n); return r ? r : new feature(n, app, ip); }

	static fserial GetSerialCount() { return (fserial)serialtable.GetSize(); }
  /// ꥢֹ椬ĤޤǳƤƤ뤫Ĥޤfeature֥Ȥ¸ߤ뤫ޤ
	static void SetBuiltInLimit() { builtin_limit = GetSerialCount(); }
  /// λ¸ߤƤfeature֥ȤBuiltInǤȤޤ
	static void terminate();
  /// λνʸˤԤޤ
};
/// </body></classdef>

//////////////////////////////////////////////////////////////////////////////
//
//  class type
//

class FSP;	// forward
class machine;		// forward

//////////////////////////////////////////////////////////////////////////////
//
//  class type
/// <classdef>
/// <name> type </name>
/// <overview> 饹</overview>
/// <desc>
/// <p>
/// 򤢤魯饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>
class type {
	lstring name;
	lstring propername;
	module *mod; /// °⥸塼̾Ǥ
	bool exported;
	bool dynamic;
	bool multifile;
	bool deleted;
public:
	int nf;				// the number of features
  /// ηͭfeatureοɽޤ
	const feature ** feats;
  /// ηͭʤˤǤ
	const type ** apptypes;
  /// ηΤȤꤦ뷿ʤˡ㤨feats[4]Ȥꤦ뷿(appropriate type)apptypes[4]ǻȤǤ롣
	int nc;				// the number of child-types
  /// Ҥtypeοɽޤ
	const type ** children;
  /// ηλҤtypeʤˤǤ
	int np;				// the number of parent-types
  /// Ƥtypeοɽޤ
	const type ** parents;
  /// ƤtypeʤˤǤ
	tserial serialno;
  /// ηΥꥢֹǤ
#ifdef XCONSTR
	bool constrained;		// constrained?
  /// constraint¸ߤ뤫ɤ
	bool feat_constrained;	// feat_constrained?
  /// ηͭapptypeconstraint¸ߤ뤫ɤ
	bool fixed;				//
  /// type֥ȤѤߤɤ򤢤路ޤ
	bool _dummy; /// for padding
#endif
	int appcounter;
  /// ηapptypeȤƤĤηѤ줿
	short *featindex;		// Table of feature -> index
  /// fserialindexбݻfeatindex[Υꥢֹ] = index
	FSP *attr;			// using pointer since forwarding class def
#ifdef XCONSTR
	code *constraint;			// pointer that holds constraint code
  /// ηȤ˼¹Ԥ̿ᡣ
#endif

#ifdef FEATURETABLE
	static int feattablesize;
  /// featindexǿǤ
#endif

	static VarArray<const type *> serialtable;
  /// ꥢֹfeature֥Ȥбطݻơ֥Ǥ
  // Table for serial -> feature
	static tserial builtin_limit;
  /// ꥢֹΤĤޤǤȤ߹߷Ǥ뤫ɽޤ
public:
		type(const char *in, module *mod, bool iexported = true);
  /// <var>in</var>̾<br>
  /// <var>mod</var>°⥸塼̾<br>
  /// <var>iexported</var>expotedɤʲIsExported()򻲾ȡ<br>
                ~type();
	
	const char *GetKey() const { return name; }
  /// HashTableѤΥ᥽åɤǤȤname֤ޤ<ref>hash.h</ref>
	const char *GetSimpleName() const { return name; }
  /// name֤ޤ
	lstring GetSimpleProperName() const;
  /// ???
	const char *GetProperName() const { return propername; }
  /// propernameޤ
	const char *GetPrintName() const;
  /// ɽѤ̾ޤ
        const char *GetName() const  { return GetProperName(); }
  /// ̾ޤpropernameޤ
	bool GetDynamic() const { return dynamic; }
  /// ηdynamicɤʥ桼ưŪѹǤ뤫ɤˤĴ٤ޤ
	bool GetMultifile() const { return multifile; }
  /// ηmultifileɤʣΥեǽ뤫ɤˤĴ٤ޤ
	void SetDynamic() { dynamic = true; }
  /// ηdynamicʥ桼ưŪѹǤ褦ˡˤˤޤ
	void SetMultifile() { multifile = true; }
  /// ηmultifileʣΥեǽ⤿ˤˤޤ
	module *GetModule() const { return mod; }
  /// η°⥸塼ޤ
	tserial GetSerialNo() const { P5("type::GetSerialNo"); return serialno; }
  /// ꥢֹޤ
#ifdef FEATURETABLE
	int GetIndex(fserial fSN) const { ASSERT(fSN < feattablesize); return feattable[fSN]; }
#else
	int GetIndex(fserial fSN) const { return GetIndex(feature::Serial(fSN)); }
  /// ˻ꤷηΤʤǲܤ˰֤뤫Ĵ٤ޤǽֹ0Ϥޤޤ
#endif
	const feature *Feature(int n) const { return feats[n]; }
  /// <var>n</var>ܤfeatureޤ<var>n</var>0GetNFeatures()-1ޤǤǤ
	tserial GetAppTypeS(int n) const { return apptypes[n]->GetSerialNo(); }
  /// <var>n</var>ܤapptypeΥꥢֹޤ<var>n</var>0GetNFeatures()-1ޤǤǤ
	tserial GetAppTypeS(const feature *f) const { int idx = GetIndex(f->GetSerialNo()); return idx == -1 ? TS_INVALID : apptypes[idx]->GetSerialNo(); }
  /// feature<var>f</var>бapptypeeΥꥢֹޤ
	const type *GetAppType(int n) const { return apptypes[n]; }
  /// nܤapptypeޤ
	const type *GetAppType(const feature *f) const { int idx = GetIndex(f->GetSerialNo()); return idx == -1 ? NULL : apptypes[idx]; }
  /// feature<var>f</var>бapptypeޤ
	int GetNFeatures() const { return nf; }
  /// featureοޤ
	const type *Child(int n) const { return children[n]; }
  /// <var>n</var>ܤλtypeޤ<var>n</var>0GetNChildren()-1ޤǤǤ
	int GetNChildren() const { return nc; }
  /// Ҥηοޤ
	const type *Parent(int n) const { return parents[n]; }
  /// <var>n</var>ܤοtypeޤ
	int GetNParents() const { return np; }
  /// Ƥηοޤ
	int GetIndex(const feature *f) const;
		// Get the index of feature f
		// return -1 if not found

	bool SetAsChildOf(type *parent);
  /// type <var>parent</var>λҤꤷޤFix()ˤȤޤ󡣡
	int  AddFeature(feature *f, const type *apptype = bot);
  /// ηɲäޤFix()ˤȤޤ󡣡
	void SetAttribute(const FSP &);
  /// ???Fix()ˤȤޤ󡣡
	FSP GetAttribute(machine &m) const;
  /// ???
	const type *TypeUnify(const type *t) const;
  /// ηȷ<var>t</var>Ȥñ첽Ԥޤз̤η֤ԤNULL֤ޤ
		// Get the unified type between this and t
		// return NULL if no candidate is found
	bool IsSubType(const type *t) const { return TypeUnify(t) == this; }
  /// type<var>t</var>subtypeɤĴ٤ޤ
	bool IsSuperType(const type *t) const { return t->TypeUnify(this) == t; }
  /// type<var>t</var>supertypeɤĴ٤ޤ
	bool IsExtendable() const;
  /// ¾Υ⥸塼ǤηsubtypeǤ뤫ɤĴ٤ޤ

#ifdef XCONSTR
	code *GetConstraint() const { return constraint; }
  /// ???
	bool IsConstrained() const { return constrained; }
  /// constraint¸ߤ뤫ɤĴ٤ޤ
	bool IsFeatureConstrained() const { return feat_constrained; }
  /// constraint¸ߤ뤫ɤĴ٤ޤ
#endif
		
	const type *AcceptableType(const feature *f) const;
  /// ηȤηΤ٤Ƥsubtypefeature<var>f</var>ĤȤǤ뷿ޤ뷿<var>a</var>Ȥsubtype<var>b</var>ξĤȤǤȤϡ־ηǤ<var>a</var>֤ޤ
		// Get the nearest descendant type that involves feature f
		// return NULL if no candidate is found

	bool IsBuiltIn() const { return serialno < builtin_limit; }
  /// Ȥ߹ߺѤߤΤΤɤĴ٤ޤȤ߹ߤΤΤʤtrue桼Τʤfalse֤ޤ
//	bool IsSpecial();
	bool IsFinal() const;
  /// ηsubtype¸ߤʤɤĴ٤ޤ
	bool IsExported() const { return exported; }
  /// ¾Υ⥸塼ǻȤݤˡ⥸塼̾ޤ᤿̾Ȥ鷺˥Ǥ뤫ɤĴ٤ޤ
	bool IsDeleted() const { return deleted; }
  /// type֥ȤѤߤɤĴ٤ޤ
// Static member functions
	static const type* Serial(tserial ser) { ASSERT(ser < serialtable.GetSize()); return serialtable[ser]; }
  /// ꥢֹ椫бtype֥Ȥޤ
	static tserial GetSerialCount() { return (tserial)serialtable.GetSize(); }
  /// ꥢֹ椬ĤޤǳƤƤ뤫Ĥޤtype֥Ȥ¸ߤ뤫ޤ
	static void SetBuiltInLimit() { builtin_limit = GetSerialCount(); }
  /// λ¸ߤƤtype֥ȤBuiltInǤȤޤ
#ifdef FEATURETABLE
	static void ExpandFeatureTable(int newnf);
  /// feature tableĥޤ
#endif

	static void terminate();
  /// λνʸˤԤޤ
	bool Fix(bool displayError = true);		// return true if success
  /// ηꤷޤ
	bool Delete();	// return true if success
  /// type֥Ȥޤ
	friend void CompileTypes();
  /// ???(compile.cppˤ)
	void IncAppCounter() const { ((type *)this)->appcounter++; }
  /// appcounter䤷ޤ
	void DecAppCounter() const { ((type *)this)->appcounter--; }
  /// appcounter򣱸餷ޤ
};
/// </body></classdef>

// About compile ... these are in compile.cpp
/// <classdef>
/// <name> featcode </name>
/// <overview>ȷñ첽ΤΥ饹</overview>
/// <desc>
/// <p>
/// ȷñ첽κݤ˼¹Ԥ<ref>featureCode</ref>ǤȤޤ
/// </p>
/// </desc>
/// <see></see>
/// <body>
struct featcode
{
	tserial result;
  /// ñ첽η̤ηΥꥢֹǤ
	ushort index;
  /// ñ첽ΥåȤindexǤ<ref>type</ref>򻲾ȡ
	code *coercion;
  /// ѴԤν򤢤路ޤ
};
/// </body></classdef>

/////////////////////////////////////////////////////////////////////////
/// <funcdef>
/// <name> CompileType </name>
/// <overview> code</overview>
/// <desc>
/// <jpn>
/// <p>
/// 2Ĥηñ첽¹Ԥ뤿code<ref>coerceHash</ref>Ͽޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <args> ñ첽2ĤηؤΥݥ󥿡ʤޤϥꥢֹˤϤޤ</args>
/// <retval> ʤ </retval>
/// <remark></remark>
/// <see></see>
/// <body>
void CompileType(tserial unifiee, tserial unifier);
void CompileType(const type *unifiee, const type *unifier);
/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////
/// <vardef>
/// <name> featureCode </name>
/// <overview>featcode2</overview>
/// <desc>
/// <p>
/// featcode2Ǥ뷿Τ򻲾ȤȤɬפʾäƤޤ<br>
/// 1ܤtypeΥꥢֹ桢2ܤfeatureΥꥢֹǤ
/// </p>
/// </desc>
/// <see></see>
/// <body>
extern featcode **featureCode;
/// </body></vardef>

/////////////////////////////////////////////////////////////////////////
/// <vardef>
/// <name> coerceNone[] </name>
/// <overview> ⤷ʤcode </overview>
/// <desc>
/// <p>
/// ⤷ʤcodeǤ
/// </p>
/// </desc>
/// <see></see>
/// <body>
extern code coerceNone[];
/// </body></vardef>

/////////////////////////////////////////////////////////////////////////
/// <vardef>
/// <name> onlyFail[] </name>
/// <overview></overview>
/// <desc>
/// <p>
/// 
/// </p>
/// </desc>
/// <see></see>
/// <body>
extern code onlyFail[];
/// </body></vardef>


#ifdef XCONSTR
/////////////////////////////////////////////////////////////////////////
/// <vardef>
/// <name> finalize_constraint </name>
/// <overview></overview>
/// <desc>
/// <p>
/// 
/// </p>
/// </desc>
/// <see></see>
/// <body>
extern code *finalize_constraint;
/// </body></vardef>

#define FINALIZE_CONSTRAINT finalize_constraint
#endif

inline tserial feature::GetKey() const { return reptype->GetSerialNo(); }
inline const type *feature::GetBaseType() const { return type::Serial(basetype); }
inline const type *feature::GetRepType() const { return reptype; }
inline void feature::SetBaseType(const type *t) { basetype = t->GetSerialNo(); }
inline const feature *feature::Search(const type *t) { if( t == NULL ) return NULL; else return Search(t->GetSerialNo()); }

#ifdef TYPE_TABLE_HASH
//////////////////////////////////////////////////////////////////////////////
/// <classdef>
/// <name> coercode </name>
/// <overview>ƱΤñ첽ΤΥ饹</overview>
/// <desc>
/// <p>
/// ƱΤñ첽κݤ˼¹Ԥ<ref>coerceCode</ref>ǤȤޤ 
/// </p>
/// </desc>
/// <see></see>
/// <body>
struct coercode
{
	int	 key;
  /// 2ơ֥ȤȸΨΤǥϥåȤΤΥ<br>tserial<var>a</var>tserial<var>b</var>äб륭aba=0012,b=0034ʤkey=00120034ʼºݤ<ref>tserial</ref>16bitǤ
	tserial result_s;
  /// ñ첽η̤ηΥꥢֹǤñ첽ԤȤ<ref>TS_INVALID</ref>ޤ
	code *coercion;
  /// ѴԤν򤢤路ޤ
#ifdef XCONSTR
	code *coercion_var;
  /// ???
	code *constraint;
  /// ???
#endif
	code *unification;
  /// ñ첽Ԥν򤢤路ޤ

	int GetKey() { return key; }
  /// HashTableѤΥ᥽åɤǤȤkey֤ޤ<ref>hash.h</ref>
};
/// </body></classdef>

/////////////////////////////////////////////////////////////////////////
/// <vardef>
/// <name> coerceHash </name>
/// <overview> coercodeΥϥå </overview>
/// <desc>
/// <p>
/// ͤcoercodeǤϥåǤcoerceCodeϥåˤΤǤ<br>
/// 2Ĥηñ첽ΤɬפʾäƤޤ
/// </p>
/// </desc>
/// <see></see>
/// <body>
extern IntegerHash<coercode> coerceHash;
/// </body></vardef>

/////////////////////////////////////////////////////////////////////////
/// <vardef>
/// <name> coercode_fail </name>
/// <overview>ñ첽ԤȤcoercode</overview>
/// <desc>
/// <p>
/// ñ첽ԤȤνcoercode֥ȤǤ
/// </p>
/// </desc>
/// <see></see>
/// <body>
extern const coercode coercode_fail;
/// </body></vardef>

inline int t2key(tserial a, tserial b) { return (((a)<<16)+(b)); }
/////////////////////////////////////////////////////////////////////////
/// <funcdef>
/// <name> coerceRef </name>
/// <overview> coercode֥Ȥ򻲾Ȥ뤿δؿ</overview>
/// <desc>
/// <jpn>
/// <p>
/// tserial <var>a</var>, tserial <var>b</var>бcoercode֥Ȥ<ref>coerceHash</ref>ޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <args> 2Ĥtserial֥ȤϤޤ</args>
/// <retval> бcoercode֥ȤФ֤ޤ бΤʤNULL֤ޤ</retval>
/// <remark></remark>
/// <see></see>
/// <body>
inline const coercode * coerceRef(tserial a, tserial b) 
	{ return coerceHash.Find(t2key(a,b)); }
/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////
/// <funcdef>
/// <name> coerceData </name>
/// <overview> coercode֥Ȥ򻲾Ȥ뤿δؿбΤʤcoercode_fail֤</overview>
/// <desc>
/// <jpn>
/// <p>
/// tserial <var>a</var>, tserial <var>b</var>бcoercode֥Ȥޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <args> 2Ĥtserial֥ȤϤޤ</args>
/// <retval> бcoercode֥ȤФ֤ޤ бΤʤ<ref>coercode_fail</ref>֤ޤ</retval>
/// <remark></remark>
/// <see></see>
/// <body>
inline const coercode & coerceData(tserial a, tserial b) 
	{ coercode *f = coerceHash.Find(t2key(a,b)); return f ? *f : coercode_fail; }
/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////
/// <funcdef>
/// <name> coerceDataAdd </name>
/// <overview> coercode֥Ȥ뤿δؿ</overview>
/// <desc>
/// <jpn>
/// <p>
/// tserial <var>a</var>, tserial <var>b</var>бcoercode֥Ȥޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <args> 2Ĥtserial֥ȤϤޤ</args>
/// <retval> бΤ¸ߤФ֤ޤ<br> ̵бcoercode֥Ȥ򿷤äƤ֤ޤ </retval>
/// <remark></remark>
/// <see></see>
/// <body>
inline coercode & coerceDataAdd(tserial a, tserial b) 
{ 
	coercode *f = coerceHash.Find(t2key(a,b)); 
	if(!f) 
	{ 
		f = new coercode; 
		f->key=t2key(a,b); 
		f->coercion = f->coercion_var = f->constraint = f->unification = NULL;
		coerceHash.Add(f); 
	} 
	return *f; 
}
/// </body></funcdef>

/////////////////////////////////////////////////////////////////////////
/// <funcdef>
/// <name> coerceDataDelete </name>
/// <overview>coercode֥Ȥ뤿δؿ </overview>
/// <desc>
/// <jpn>
/// <p>
/// tserial <var>a</var>, tserial <var>b</var>бcoercode֥Ȥޤ
/// </p>
/// </jpn>
/// <eng>
/// <p>
/// not yet
/// </p>
/// </eng>
/// </desc>
/// <args> 2Ĥtserial֥ȤϤޤ</args>
/// <retval> true򡢼Ԥfalse֤ޤ </retval>
/// <remark></remark>
/// <see></see>
/// <body>
inline bool coerceDataDelete(tserial a, tserial b) 
{ 
	coerceHash.Delete(t2key(a,b)); 
	return true;
}
/// </body></funcdef>

#else
struct coercode
{
	tserial result_s;
	code *coercion;
#ifdef XCONSTR
	code *coercion_var;
	code *constraint;
#endif
	code *unification;
};

/////////////////////////////////////////////////////////////////////////
/// <vardef>
/// <name> coerceCode </name>
/// <overview>coercode2</overview>
/// <desc>
/// <p>
/// coercode2Ǥ
/// </p>
/// </desc>
/// <see>unification_overview</see>
/// <body>
extern coercode **coerceCode;
/// </body></vardef>

inline const coercode & coerceData(tserial a, tserial b) 
	{ return coerceCode[(a)][(b)]; }
inline coercode & coerceDataAdd(tserial a, tserial b) 
	{ return coerceCode[(a)][(b)]; }
inline bool coerceDataDelete(tserial a, tserial b)  {
	coerceCode[(a)][(b)].result_s = TS_INVALID; 
	coerceCode[(a)][(b)].coercion = coerceCode[(a)][(b)].coercion_var = coerceCode[(a)][(b)].constraint = coerceCode[(a)][(b)].unification = onlyFail;
}
#endif


lstring Quote(const char *str, bool AllowCapital = false);

} // namespace lilfes

#endif // __ftype_h
