//////////////////////////////////////////////////////////////////////
////
////  Copyright (c) 1999, MAKINO Takaki and MIYAO Yusuke
////  You may distribute under the terms of the Artistic License.
////
////  Name: stream.h
////  Author: MAKINO Takaki (mak@is.s.u-tokyo.ac.jp)
////          MIYAO Yusuke (yusuke@is.s.u-tokyo.ac.jp)
////  Time-stamp: <1999-11-22 15:21:22 yusuke>
////  Purpose: Implementation of streams
////  $Id: stream.h,v 1.7 2005/02/12 07:11:54 yusuke Exp $
////
//////////////////////////////////////////////////////////////////////




///  <ID>$Id: stream.h,v 1.7 2005/02/12 07:11:54 yusuke Exp $</ID>
//
///  <COLLECTION> lilfes-bytecode </COLLECTION>
//
///  <name>stream.h</name>
//
///  <overview>
///  <jpn>stream롼</jpn>
///  <eng>stream operation routine </eng>
///  </overview>


/// <notiondef>
///
/// <name> stream_overview </name>
/// <title> stream </title>
/// <desc>
///  stream.hϥǡ̤򰷤Τ˻Ȥޤ
/// </desc>
/// <see>  </see>
/// </notiondef>


#ifndef __stream_h
#define __stream_h

#include "lconfig.h"
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>

//#include <netinet/in.h>

#include "sockstream.h"
#include "procstream.h"
#include "basic.h"
#include "builtin.h"

namespace lilfes {

//////////////////////////////////////////////////////////////////////
//  Error on lilfes streams

class lilfes_stream_error {
private:
  const char* m;
public:
  const char* message() {
    return m;
  }
  lilfes_stream_error( const char* m_ ) : m( m_ ) {
  }
};

//////////////////////////////////////////////////////////////////////
//  Table of lilfes_stream

class lilfes_stream;
typedef size_t stream_id_type;

class lilfes_stream_table {
public:
  static const size_t MAX_STREAMS = 1024;

private:
  stream_id_type stream_id_count;
  /// ˻ȤΤȹͤ륹ȥ꡼ΥϥɥݻѿǤ
  std::vector< lilfes_stream* > stream_id_list;
  /// ȥ꡼ΥϥɥꥹȤǤ

public:
  lilfes_stream_table( size_t s = MAX_STREAMS );
  virtual ~lilfes_stream_table();

public:
  stream_id_type newStream( lilfes_stream* s );
  void deleteStream( stream_id_type id ) {
    if ( id < stream_id_list.size() ) {
      stream_id_list[ id ] = 0;
    }
  }
  lilfes_stream* findStream( stream_id_type id ) const {
    if ( id < stream_id_list.size() ) {
      return stream_id_list[ id ];
    } else {
      return 0;
    }
  }
};

//////////////////////////////////////////////////////////////////////
//  Base class for streams on LiLFeS

/// <classdef>
/// <name> lilfes_stream </name>
/// <overview> LiLFeS Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤϥȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

class lilfes_stream
{
protected:
  stream_id_type handle;
  /// ѤƤ륹ȥ꡼ΥϥɥݻѿǤ

  static lilfes_stream_table stream_table;

  static lilfes_stream* standard_in_ptr;
  static lilfes_stream* standard_out_ptr;
  static lilfes_stream* standard_err_ptr;

public:
  lilfes_stream(void);
  /// ȥ꡼९饹Υ󥹥ȥ饯Ǥ
  virtual ~lilfes_stream();
  /// ȥ꡼९饹Υǥȥ饯Ǥ

//   virtual bool Close(void) { return true; }
//   virtual bool IsOpened(void) const { return true; }
//   /// ȥ꡼ξ֤򼨤۴ؿǤ
//   virtual ssize_t Write(const void* ptr, size_t size) = 0;
//   /// ȥ꡼˥ǡ񤭹ߡ񤭹size֤貾۴ؿǤ
//   /// <var>ptr</var>񤭹ǡ<br>
//   /// <var>size</var>񤭹size[byte]
//   virtual bool SyncWrite(const void* ptr, size_t size);
//   /// ȥ꡼˥ǡ񤭹եå夹벾۴ؿǤTrueԤFalse֤ޤ
//   /// <var>ptr</var>񤭹ǡ<br>
//   /// <var>size</var>񤭹size[byte]
//   virtual ssize_t Read(void* ptr, size_t size) = 0;
//   /// ȥ꡼फǡɤ߹ߡɤ߹size֤貾۴ؿǤ
//   /// <var>ptr</var>ɤߤǡ<br>
//   /// <var>size</var>ɤߤsize[byte]
//   virtual bool SyncRead(void* ptr, size_t size);
//   /// ȥ꡼˥ǡɤ߹եå夹벾۴ؿǤTrueԤFalse֤ޤ
//   /// <var>ptr</var>ɤߤǡ<br>
//   /// <var>size</var>ɤߤsize[byte]
//   virtual bool eof(void) const = 0;
//   /// ȥ꡼बüƤ뤫ɤ򼨤貾۴ؿǤ
//   virtual void Flush(void) const = 0;
//   /// ȥ꡼ΥХåեեå夹貾۴ؿǤ

  virtual bool Close() { return true; }
  /// ȥ꡼Ĥޤ
  virtual bool IsOpened() {
    std::ios* str = GetCStream();
    return str && *str;
  }
  /// ȥ꡼ξ֤򼨤۴ؿǤ

  virtual std::ios* GetCStream() = 0;
  /// C++ Υȥ꡼९饹ؿǤ

  virtual int GetHandle(void) { return handle; }
  /// ȥ꡼ϥɥintǼ벾۴ؿǤ
  virtual bool GetHandleFSP(FSP) = 0;
  /// ȥ꡼ϥɥFSPǼ貾۴ؿǤ

  static lilfes_stream* GetStreamFSP(FSP&);
  /// FSPΥȥ꡼ϥɥ뤫餽бlilfes_streamΥݥ󥿤᥽åɤǤ
  /// <var>FSP</var>FSPϥɥ<br>
  static lilfes_stream* GetStream(int handle) { return stream_table.findStream( handle ); }
  /// intΥȥ꡼ϥɥ뤫餽бlilfes_streamΥݥ󥿤᥽åɤǤ
  /// <var>handle</var>intϥɥ<br>

  static lilfes_stream* Stdin(void) { return standard_in_ptr; }
  /// ɸϥȥ꡼бlilfes_streamݥ󥿤᥽åɤǤ
  static lilfes_stream* Stdout(void) { return standard_out_ptr; }
  /// ɸϥȥ꡼бlilfes_streamݥ󥿤᥽åɤǤ
  static lilfes_stream* Stderr(void) { return standard_err_ptr; }
  /// ɸ२顼ϥȥ꡼бlilfes_streamݥ󥿤᥽åɤǤ

//   static bool SwitchStdin(lilfes_stream *s);
//   static bool SwitchStdout(lilfes_stream *s);
//   static bool SwitchStderr(lilfes_stream *s);

};

//////////////////////////////////////////////////////////////////////
//  Null stream
/// <classdef>
/// <name> lilfes_null_stream </name>
/// <overview> LiLFeS Null Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤƶȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

class nullstreambuf : public std::streambuf {
public:
  typedef size_t size_type;
  typedef char char_type;

public:
  nullstreambuf() {}
  virtual ~nullstreambuf() {}

  int sync() {
    setp( 0, 0 );
    return 0;
  }

  int underflow() {
    return EOF;
  }

  int overflow( int c = EOF ) {
    sync();
    if ( c == EOF ) return EOF;
    return c;
  }
};

class nullstream : public std::ostream {
private:
  nullstreambuf bf_buf;

public:
  nullstream() : std::ostream( NULL ), bf_buf() {
    init( &bf_buf );
  }
  virtual ~nullstream() {
  }
};

class lilfes_null_stream : public lilfes_stream {
private:
  nullstream ns;

public:
  lilfes_null_stream() {}
  virtual ~lilfes_null_stream() {}
  std::ios* GetCStream() { return &ns; }
  bool GetHandleFSP(FSP);
};

//////////////////////////////////////////////////////////////////////
//  Standard streams
/// <classdef>
/// <name> lilfes_stdin_stream </name>
/// <overview> LiLFeS Stdin Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤɸϤΥȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>
class lilfes_stdin_stream : public lilfes_stream {
public:
  lilfes_stdin_stream() {}
  virtual ~lilfes_stdin_stream() {}

  bool Close() { return false; }

  std::ios* GetCStream() { return input_stream; }

  bool GetHandleFSP(FSP);
};

/// <classdef>
/// <name> lilfes_stdout_stream </name>
/// <overview> LiLFeS Stdout Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤɸϤΥȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>
class lilfes_stdout_stream : public lilfes_stream {
public:
  lilfes_stdout_stream() {}
  virtual ~lilfes_stdout_stream() {}

  bool Close() { return false; }

  std::ios* GetCStream() { return output_stream; }

  bool GetHandleFSP(FSP);
};

/// <classdef>
/// <name> lilfes_stderr_stream </name>
/// <overview> LiLFeS Stderr Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤɸ२顼ϤΥȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>
class lilfes_stderr_stream : public lilfes_stream {
public:
  lilfes_stderr_stream() {}
  virtual ~lilfes_stderr_stream() {}

  bool Close() { return false; }

  std::ios* GetCStream() { return error_stream; }

  bool GetHandleFSP(FSP);
};

//////////////////////////////////////////////////////////////////////
//  Stream for debugging
/// <classdef>
/// <name> lilfes_debug_stream </name>
/// <overview> LiLFeS Debug Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤƥǥХåѤΥȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

// class lilfes_debug_stream : public lilfes_stream {
// public:
//   bool Close(void) { return true; }
//   bool IsOpened(void) const { return true; }

//   ssize_t Write(const void* ptr, size_t size);
//   ssize_t Read(void*, size_t);
//   bool eof(void) const { return false; }
//   void Flush(void) const {}

//   bool GetHandleFSP(FSP);
// };

//////////////////////////////////////////////////////////////////////
//  Socket stream
/// <classdef>
/// <name> lilfes_socket_stream_base </name>
/// <overview> LiLFeS Socket Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤƥåȥȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

// class lilfes_socket_stream_base : public lilfes_stream {
// protected:
//   int sock;
// public:
//   lilfes_socket_stream_base(int sock_ = 0);
//   /// LiLFeS Socket Stream饹Υ󥹥ȥ饯Ǥ
//   /// <var>sock_</var>åֹ<br>
//   virtual ~lilfes_socket_stream_base() {}

//   bool Close(void);
//   bool IsOpened(void) const { return 0 != sock; }

//   ssize_t Write(const void *ptr, size_t size);
//   ssize_t Read(void *ptr, size_t size);
//   bool eof(void) const { return false; }
//   void Flush(void) const {}

//   //FSP GetHandleFSP(void);
// };

/// </body>
/// </classdef>

//////////////////////////////////////////////////////////////////////
//  Socket stream
/// <classdef>
/// <name> lilfes_socket_stream </name>
/// <overview> LiLFeS Socket Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤƥåȥȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

class lilfes_socket_stream : public lilfes_stream {
private:
  std::string hostname;
  int port_number;
  socket_stream sock;

public:
  lilfes_socket_stream(const std::string& hostname_, int port_number_)
    : hostname( hostname_ ), port_number( port_number_ ), sock( hostname, port_number ) {
  }
  virtual ~lilfes_socket_stream() {}

  bool Close() {
    sock.close();
    return sock;
  }

  std::ios* GetCStream() { return &sock; }

  bool GetHandleFSP(FSP);
};

/// </body>
/// </classdef>

//////////////////////////////////////////////////////////////////////
//  Socket stream
/// <classdef>
/// <name> lilfes_server_stream </name>
/// <overview> LiLFeS Server Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤServerȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

class lilfes_server_stream : public lilfes_stream {
private:
  int port_number;
  server_stream sock;

public:
  lilfes_server_stream(int port_number_)
  /// LiLFeS Server Stream饹Υ󥹥ȥ饯Ǥ
  /// <var>port_number_</var>ݡֹ<br>
    : port_number( port_number_ ), sock( port_number ) {
  }
  virtual ~lilfes_server_stream() {}

  bool Close() {
    sock.close();
    return sock;
  }

  std::ios* GetCStream() { return &sock; }

  virtual bool is_connected() {
    return sock.is_connected();
  }
  virtual int accept_connection() {
    return sock.accept_connection();
  }
  virtual void close_connection() {
    sock.close_connection();
  }

  bool GetHandleFSP(FSP);
};

/// </body>
/// </classdef>

//////////////////////////////////////////////////////////////////////
//  PTY stream
/// <classdef>
/// <name> lilfes_pty_stream </name>
/// <overview> LiLFeS PTY Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤƲüȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

class lilfes_pty_stream : public lilfes_stream {
private:
  char* command;
  int argc;
  char** argv;
  pty_stream pty;

public:
  lilfes_pty_stream(char* command_, int argc_, char** argv_)
  /// LiLFeS PTY Stream 饹Υ󥹥ȥ饯Ǥ
  /// <var>=command_</var>: ư륳ޥ̾<br>
  /// <var>=argc_</var>: ο<br>
  /// <var>軰=argv_</var>: <br>
    : command( command_ ), argc( argc_ ), argv( argv_ ), pty( command, argc, argv ) {
  }
  virtual ~lilfes_pty_stream() {}

  bool Close() {
    pty.close();
    return pty;
  }

  std::ios* GetCStream() { return &pty; }

  bool GetHandleFSP(FSP);
};

/// </body>
/// </classdef>

//////////////////////////////////////////////////////////////////////
//  Process stream
/// <classdef>
/// <name> lilfes_process_input_stream </name>
/// <overview> LiLFeS Process Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤƥѥפ̤ƥץ̿Ԥʤȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

class lilfes_process_input_stream : public lilfes_stream {
private:
  char* command;
  int argc;
  char** argv;
  process_istream process;

public:
  lilfes_process_input_stream(char* command_, int argc_, char** argv_ )
  /// LiLFeS PTY Stream 饹Υ󥹥ȥ饯Ǥ
  /// <var>=command_</var>: ư륳ޥ̾<br>
  /// <var>=argc_</var>: ο<br>
  /// <var>軰=argv_</var>: <br>
  /// <var>Ͱ=mode_</var>: ѥפμ(i/o)<br>
    : command( command_ ), argc( argc_ ), argv( argv_ ), process( command, argc, argv ) {
  }
  virtual ~lilfes_process_input_stream() {}

  bool Close() {
    process.close();
    return process;
  }

  std::ios* GetCStream() { return &process; }

  bool GetHandleFSP(FSP);
};

/// </body>
/// </classdef>

/// <classdef>
/// <name> lilfes_process_output_stream </name>
/// <overview> LiLFeS Process Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤƥѥפ̤ƥץ̿Ԥʤȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

class lilfes_process_output_stream : public lilfes_stream {
private:
  char* command;
  int argc;
  char** argv;
  process_ostream process;

public:
  lilfes_process_output_stream(char* command_, int argc_, char** argv_ )
  /// LiLFeS PTY Stream 饹Υ󥹥ȥ饯Ǥ
  /// <var>=command_</var>: ư륳ޥ̾<br>
  /// <var>=argc_</var>: ο<br>
  /// <var>軰=argv_</var>: <br>
  /// <var>Ͱ=mode_</var>: ѥפμ(i/o)<br>
    : command( command_ ), argc( argc_ ), argv( argv_ ), process( command, argc, argv ) {
  }
  virtual ~lilfes_process_output_stream() {}

  bool Close() {
    process.close();
    return process;
  }

  std::ios* GetCStream() { return &process; }

  bool GetHandleFSP(FSP);
};

/// </body>
/// </classdef>

//////////////////////////////////////////////////////////////////////
//  File stream
/// <classdef>
/// <name> lilfes_file_stream </name>
/// <overview> LiLFeS File Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤƥե륹ȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

class lilfes_file_stream : public lilfes_stream {
private:
  char* filename;
  char* openmode;
  char* option;
  std::iostream* fstr;

public:
  lilfes_file_stream(char*, char*, char* = "");
  /// LiLFeS File Stream饹Υ󥹥ȥ饯Ǥ
  /// <var>=filename_</var>ե̾<br>
  /// <var>=openmode_</var>եΥץ⡼ɡʲΥ⡼ɤǤޤ<br>
  ///<LI>"r" : ɤ߹Ѥǳե뤬̵ϽҸ줬Ԥ롣 
  ///<LI>"w" : 񤭹Ѥǳ¸ΥեϾ񤭤롣 
  ///<LI>"a" : ɲý񤭹Ѥǳ 
  ///<LI>"r+" : ɤ߽񤭤ǳե뤬̵Ϻ롣 
  ///<LI>"w+" : ɤ߽񤭤ǳ¸ΥեϾ񤭤롣 
  ///<LI>"a+" : ɤ߽񤭤ǳե񤭹ޤ롥 </LI>
  /// <var>軰=option_</var>̷ ("gz" ޤ "bz2")<br>
  ///<body>
  virtual ~lilfes_file_stream() {
    Close();
  }
  bool Close(void);

  std::ios* GetCStream() { return fstr; }

  bool GetHandleFSP(FSP);
};

//////////////////////////////////////////////////////////////////////
//  Sting stream
/// <classdef>
/// <name> lilfes_string_stream </name>
/// <overview> LiLFeS String Stream 饹</overview>
/// <desc>
/// <p>
/// LiLFeSˤʸ󥹥ȥ꡼򰷤饹Ǥ
/// </p>
/// </desc>
/// <see></see>
/// <body>

class lilfes_string_stream : public lilfes_stream {
private:
  std::stringstream ss;

public:
  lilfes_string_stream() {
  }
  /// LiLFeS Sting Stream饹Υ󥹥ȥ饯Ǥ
  virtual ~lilfes_string_stream() {}

  std::ios* GetCStream() { return &ss; }

  bool GetHandleFSP(FSP);
};

/// </body>
/// </classdef>

} // namespace lilfes

#endif // __stream_h
// end of stream.h
