/* $Id: cell.cpp,v 1.7 2005/05/18 10:29:34 ninomi Exp $ 
 *
 *    Copyright (c) 1997-1998, Makino Takaki
 *
 *    You may distribute this file under the terms of the Artistic License.
 *
 */

static const char rcsid[] = "$Id: cell.cpp,v 1.7 2005/05/18 10:29:34 ninomi Exp $";

#include "cell.h"
#include "machine.h"
#include <iostream>

using namespace std;

namespace lilfes {

//////////////////////////////////////////////////////////////////////////////
//
// class cell
//

ostream& operator<<(ostream& os, cell c)
{
//	ASSERT( c != NULL );
  switch( Tag(c) )
    {
    case T_STR:
      os << "STR " << c2STR(c)->GetName(); break;
    case T_VAR:
      os << "VAR " << c2VAR(c)->GetName(); break;
    case T_PTR:
    case T_PTR1:
    case T_PTR2:
    case T_PTR3:
      if( c2PTR(c) == CORE_P_INVALID )
        os << "PTR (INVALID)"; 
      else
        os << "PTR " << std::hex << GetSlot(c2PTR(c)) << "-" << (cp2int(GetOffset(c2PTR(c)))) << std::dec; 
      break;
    case T_INT:
      os << "INT " << c2INT(c); break;
    case T_FLO:
      os << "FLO " << c2FLO(c); break;
    case T_CHR:
      os << "CHR '" << (char) c2CHR(c) << "' (0x" << std::hex << c2CHR(c) << std::dec << ")"; break;
    case T_STG:
      os << "STG \"" << c2STG(c) << "\""; break;
    default:
      os << "Unknown Tag"; break;
    }
  return os;
}


template class VarArray<const strsymbol *>;

VarArray<const strsymbol *> strsymbol::serialtable;

template <> uint VarArray<const strsymbol *>::growstep = 512;
template <> uint VarArray<const strsymbol *>::growmask = 511;
template <> uint VarArray<const strsymbol *>::growbits = 9;

strsymbol::strsymbol(machine *mach, const char *ininame) 
{ 
	name = ininame;
	hash.Add(this);
	used = permanent = false;
	serialno = serialtable.GetSize();
#ifdef USE_STRING_GC
	if( (serialno&65535)==0 && serialno > 0)
		gc(mach);
#endif
	serialtable.Next() = this;
}

strsymbol::~strsymbol() 
{ 
	serialtable[serialno] = NULL;
	hash.Delete(this);
}

void strsymbol::gc(machine *mach)
{
	P1("String GC");

	//std::cout << "[String GC at count=" << serialtable.GetSize() << flush;
	
	int serialcount = serialtable.GetSize();
	for( int i=0; i < serialcount; i++ )
	{
		if( serialtable[i] != NULL )
			serialtable[i]->used = serialtable[i]->permanent;
	}

	for( core_p i=int2core_p(0); i < mach->GetRP(); i++ )
		if( IsSTG(mach->ReadLocal(i)) )
			serialtable[c2STGS(mach->ReadLocal(i))]->used = true;

	for( core_p i=int2core_p(0); i < mach->GetTP(); i++ )
		if( IsSTG(mach->ReadTrail(i)) )
			serialtable[c2STGS(mach->ReadTrail(i))]->used = true;

	for( int h=0; h < NHEAP; h++ )
	{
		core_p limit;
		if( h == NHEAP - 1 )
			limit = mach->GetSP();
		else if( h == GetSlot(mach->GetHP()) )
			limit = mach->GetHP();
		else
			limit = mach->ReadHeap(MakeCoreP(h, int2core_p(0)));

		//std::cout << "(heap" << h << "=" << limit - MakeCoreP(h, 0) << ")" << flush;
		for( core_p i=MakeCoreP(h, 0); i<limit; i++ )
			if( IsSTG(mach->ReadHeap(i)) )
				serialtable[c2STGS(mach->ReadHeap(i))]->used = true;
	}

	int cnt = 0, siz = 0;
	for( int i=0; i < serialcount; i++ )
	{
		if( serialtable[i] != NULL )
			if( serialtable[i]->used == false )
			{
				cnt++; siz += strlen(serialtable[i]->name)+8;
				delete serialtable[i];
			}
//			else
//				std::cout << "Remain: \"" << serialtable[i]->name << "\" (permanent = " << (serialtable[i]->permanent ? "true" : "false") << ")" << std::endl;
	}
	
	//std::cout << " " << cnt << "strings (" << siz << " bytes) collected. Remained strings = " << hash.GetNData() << "]" << std::endl;

}

} // namespace lilfes
