/* $Id: hash.cpp,v 1.6 2005/05/18 10:29:34 ninomi Exp $
 *
 *    Copyright (c) 1996-1998, Makino Takaki
 *
 *    You may distribute this file under the terms of the Artistic License.
 *
 */

static const char rcsid[] = "$Id: hash.cpp,v 1.6 2005/05/18 10:29:34 ninomi Exp $";

#include "hash.h"
#include "proc.h"
#include "parse.h"
#include "in.h"
#include "option.h"
#include <iostream>
#ifdef PROFILE_HASH
#include <typeinfo>
#endif

using namespace std;

namespace lilfes {


//int VarArray<feature>::growstep = 8;
//int VarArray<type>::growstep = 8;

template <> core_p Queue<core_p>::invalid = CORE_P_INVALID;
template <> Pair<core_p, code *> Queue< Pair<core_p, code *> >::invalid = Pair<core_p, code *>(CORE_P_INVALID, NULL);

template <class Tobj>
IntegerHash<Tobj>::IntegerHash(const char *
#ifdef PROFILE_HASH
                               iname
#endif // PROFILE_HASH
                               , unsigned inisize, unsigned max)
{
	ASSERT( inisize != 0 );
	ASSERT( (inisize & (inisize-1)) == 0 );
#ifdef PROFILE_HASH
	name = iname;
	add = col = 0;
#endif
	tablesize = inisize;
	usagemax = max;
	ndata = 0;
	table = new Tobj* [inisize];
	while( inisize )
		table[--inisize] = NULL;
}

template <class Tobj>
StringHash<Tobj>::StringHash(const char *
#ifdef PROFILE_HASH
                             iname
#endif // PROFILE_HASH
                             , unsigned inisize, unsigned max)
{
	ASSERT( inisize != 0 );
	ASSERT( (inisize & (inisize-1)) == 0 );
#ifdef PROFILE_HASH
	name = iname;
	add = col = 0;
#endif
	usagemax = max;
	tablesize = inisize;
	ndata = 0;
	table = new Tobj* [inisize];
	while( inisize )
		table[--inisize] = NULL;
}

template <class Tobj>
FSPHash<Tobj>::FSPHash(const char *
#ifdef PROFILE_HASH
                       iname
#endif // PROFILE_HASH
                       , unsigned inisize, unsigned max)
{
	ASSERT( inisize != 0 );
	ASSERT( (inisize & (inisize-1)) == 0 );
#ifdef PROFILE_HASH
	name = iname;
	add = col = 0;
#endif
	tablesize = inisize;
	usagemax = max;
	ndata = 0;
	table = new Tobj* [inisize];
	while( inisize )
		table[--inisize] = NULL;
}

template <class Tobj>
IntegerHash<Tobj>::IntegerHash(IntegerHash<Tobj> &t)
{
#ifdef PROFILE_HASH
	name = t.name;
	add = col = 0;
#endif
	tablesize = t.tablesize;
	ndata = t.ndata;
	table = new Tobj* [tablesize];
	for( uint i=0; i<tablesize; i++ )
		table[i] = t.table[i];
}

template <class Tobj>
StringHash<Tobj>::StringHash(StringHash<Tobj> &t)
{
#ifdef PROFILE_HASH
	name = t.name;
	add = col = 0;
#endif
	tablesize = t.tablesize;
	ndata = t.ndata;
	table = new Tobj* [tablesize];
	for( uint i=0; i<tablesize; i++ )
		table[i] = t.table[i];
}

template <class Tobj>
FSPHash<Tobj>::FSPHash(FSPHash<Tobj> &t)
{
#ifdef PROFILE_HASH
	name = t.name;
	add = col = 0;
#endif
	tablesize = t.tablesize;
	ndata = t.ndata;
	table = new Tobj* [tablesize];
	for( uint i=0; i<tablesize; i++ )
		table[i] = t.table[i];
}

template <class Tobj>
IntegerHash<Tobj>::~IntegerHash()
{
	ASSERT(table != NULL);

	delete[] table;

#ifdef PROFILE_HASH
	if( add )
		std::cout << "Hash statistics Int(" << typeid(Tobj).name() << ")[" << name << "]: " << col << "/" << add << "=" << (double)col/add << std::endl;
#endif
}

template <class Tobj>
StringHash<Tobj>::~StringHash()
{
	ASSERT(table != NULL);
	
	delete[] table;

#ifdef PROFILE_HASH
	if( add )
		std::cout << "Hash statistics Str(" << typeid(Tobj).name() << ")[" << name << "]: " << col << "/" << add << "=" << (double)col/add << std::endl;
#endif
}

template <class Tobj>
FSPHash<Tobj>::~FSPHash()
{
	ASSERT(table != NULL);
	
	delete[] table;

#ifdef PROFILE_HASH
	if( add )
		std::cout << "Hash statistics FSP(" << typeid(Tobj).name() << ")[" << name << "]: " << col << "/" << add << "=" << (double)col/add << std::endl;
	delete[] name;
#endif
}

template <class Tobj>
Tobj *IntegerHash<Tobj>::Add(Tobj *t)
{
#ifdef PROFILE_HASH
//	if( (ndata & 0x3ff) == 0 )
//		std::cout << "TABLESIZE = " << tablesize << ", ndata = " << ndata << "   " << col << "/" << add << "=" << (double)col/add << std::endl;
	add++;
#endif
	Tobj **p;
	ASSERT(t != NULL);
	ASSERT(table != NULL);
	
	unsigned long key = (unsigned long)(t->GetKey());
	int h = hash(key);
	
	if( (ndata<<4) > tablesize*usagemax )
		ExpandTable();
	
	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
#ifdef PROFILE_HASH
		col++;
#endif
		if( (unsigned long)((*p)->GetKey()) == key )
		{
			DMSG("Duplicate-key Add ignored");
			return *p;
		}
		h++;
	}
	ndata++;
	*p = t;
	return NULL;
}

template <class Tobj>
Tobj *StringHash<Tobj>::Add(Tobj *t)
{
#ifdef PROFILE_HASH
	add++;
#endif

	Tobj **p;
	ASSERT(t != NULL);
	ASSERT(table != NULL);
	
	const char *key = t->GetKey();
	int h = hash(key);
	
	if( (ndata<<4) > tablesize*usagemax )
		ExpandTable();
	
	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
#ifdef PROFILE_HASH
		col++;
#endif
		if( strcmp((*p)->GetKey(), key) == 0 ) //ƱĤΤĤäAddʤ
		{
			DMSG("Duplicate-key Add ignored");
			return *p;
		}
		h++;
	}
	ndata++;
	*p = t;
	return NULL;
}

template <class Tobj>
Tobj *IntegerHash<Tobj>::AddOrReplace(Tobj *t)
{
#ifdef PROFILE_HASH
	add++;
#endif
	Tobj **p;
	ASSERT(t != NULL);
	ASSERT(table != NULL);
	
	unsigned long key = (unsigned long)(t->GetKey());
	int h = hash(key);
	
	if( (ndata<<4) > tablesize*usagemax )
		ExpandTable();
	
	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
#ifdef PROFILE_HASH
		col++;
#endif
          //ƱĤΤ֤
		if( (unsigned long)((*p)->GetKey()) == key )
		{
			Tobj *oldp = *p;
			*p = t;
			return oldp;
		}
		h++;
	}
	ndata++;
	*p = t;
	return NULL;
}

template <class Tobj>
Tobj *StringHash<Tobj>::AddOrReplace(Tobj *t)
{
#ifdef PROFILE_HASH
	add++;
#endif

	Tobj **p;
	ASSERT(t != NULL);
	ASSERT(table != NULL);
	
	const char *key = t->GetKey();
	int h = hash(key);
	
	if( (ndata<<4) > tablesize*usagemax )
		ExpandTable();
	
	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
#ifdef PROFILE_HASH
		col++;
#endif
		if( strcmp((*p)->GetKey(), key) == 0 )
		{
			Tobj *oldp = *p;
			*p = t;
			return oldp;
		}
		h++;
	}
	ndata++;
	*p = t;
	return NULL;
}

template <class Tobj>
Tobj *FSPHash<Tobj>::Add(Tobj *t)
{
#ifdef PROFILE_HASH
	add++;
#endif

	Tobj **p;
	ASSERT(t != NULL);
	ASSERT(table != NULL);
	
	FSP key = t->GetKey();
	int h = hash(key);

//	std::cout << "FSPHash::Add hash = " << h << std::endl << key.DisplayLiLFeS();
	
	if( (ndata<<4) > tablesize*usagemax )
		ExpandTable();
	
	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
#ifdef PROFILE_HASH
		col++;
#endif
		if( (*p)->GetKey().b_equiv(key) )
		{
			DMSG("Duplicate-key Add ignored");
			return *p;
		}
		h++;
	}
	ndata++;
	*p = t;
	return NULL;
}

template <class Tobj>
Tobj *IntegerHash<Tobj>::Find(ulong key) const
{
	Tobj **p;
	ASSERT(table != NULL);
	
	int h = hash(key);
	
	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
		if( (ulong)(*p)->GetKey() == key )
		{
			return *p;
		}
		h++;
	}
	return NULL;
}

template <class Tobj>
Tobj *StringHash<Tobj>::Find(const char * key) const
{
	Tobj **p;
	ASSERT(table != NULL);
	
	int h = hash(key);
	
	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
		if( strcmp((*p)->GetKey(), key) == 0 )
		{
			return *p;
		}
		h++;
	}
	return NULL;
}

template <class Tobj>
Tobj *FSPHash<Tobj>::Find(FSP key) const
{
	Tobj **p;
	ASSERT(table != NULL);
	
	int h = hash(key);
	
//	std::cout << "FSPHash::Find hash = " << h << std::endl << key.DisplayLiLFeS();

	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
		if( (*p)->GetKey().b_equiv(key) )
		{
			return *p;
		}
		h++;
	}
	return NULL;
}

template <class Tobj>
Tobj *IntegerHash<Tobj>::Delete(ulong key) 
{
	Tobj **p;
	Tobj *ret = NULL;
	ASSERT(table != NULL);
	
	int h = hash(key);
	
	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
		if( (ulong)(*p)->GetKey() == key )
		{
			ret = *p;
			*p = NULL;
			ndata--;

			h++;
			while( table[h & (tablesize-1)] != NULL )
			{
				Tobj *x = table[h & (tablesize-1)];
				table[h & (tablesize-1)] = NULL;
				ndata--;
				Add(x);
				h++;
			}
			return ret;
		}
		h++;
	}
	return NULL;
}

template <class Tobj>
Tobj *StringHash<Tobj>::Delete(const char * key)
{
	Tobj **p;
	Tobj *ret = NULL;
	ASSERT(table != NULL);
	
	int h = hash(key);
	
	while( *(p=&table[h & (tablesize-1)]) != NULL )
	{
		if( strcmp((*p)->GetKey(), key) == 0 )
		{
			ret = *p;
			*p = NULL;
			ndata--;

			h++;
			while( table[h & (tablesize-1)] != NULL )
			{
				Tobj *x = table[h & (tablesize-1)];
				table[h & (tablesize-1)] = NULL;
				ndata--;
				Add(x);
				h++;
			}
			return ret;
		}
		h++;
	}
	return NULL;
}

template <class Tobj>
Tobj *IntegerHash<Tobj>::Delete(Tobj *data) { return Delete(data->GetKey()); }
template <class Tobj>
Tobj *StringHash<Tobj>::Delete(Tobj *data) { return Delete(data->GetKey()); }

template <class Tobj>
inline int IntegerHash<Tobj>::hash(ulong p)
{
#ifdef HAS_INT64
	return (p*0x8b214823LL)>>31;
#else
	return p * 0x29 + (p * 0x211 >> 14);
#endif
}// 1000 1011 0010 0001 0100 1000 0010 0011

template <class Tobj>
int StringHash<Tobj>::hash(const char * p)
{
	int ret = 0;
	ASSERT( p != NULL );
	while( *p )
	{
		ret = *p * 257 + ((ret*(0x9449+*p)) >> 7);
//		ret += ret + *p;
		p++;
	}
	return ret;
}

template <class Tobj>
int FSPHash<Tobj>::hash(FSP p)
{
	int ret = 0;

	machine &m = * p.GetMachine();
	core_p cp = p.GetAddress();

	cell c;
	while(1)
	{
		c = m.ReadHeap(cp++);
		if( IsPTR(c) )
		{
			if(c == PTR2c(CORE_P_INVALID))
				break;
			c = int2cell( cell2int(c) - cell2int(PTR2c(cp)) );
		}
		ret = cell2int(c) * 257 + ((ret*(0x94493057^cell2int(c))) >> 7);
	} 
	return ret;
}

template <class Tobj>
void IntegerHash<Tobj>::ExpandTable()
{
	P1("Hash::ExpandTable()");
#ifdef DEBUG
	std::cout << "TABLESIZE = " << tablesize << ", ndata = " << ndata << std::endl;
#endif
	ASSERT( tablesize != 0 );
	ASSERT( (tablesize & (tablesize-1)) == 0 );
	ASSERT( table != NULL );


#ifdef PROFILE_HASH
	std::cout << "ExpandTable Hash statistics Int(" << typeid(Tobj).name() << ")[" << name << "]: " << col << "/" << add << "=" << (double)col/add << std::endl;
	add = col = 0;
#endif
	
	Tobj **oldtable = table;
	int oldtablesize = tablesize;
	
	tablesize = tablesize << 1;
	table = new Tobj *[tablesize];
	ndata = 0;

	int inisize = tablesize;
	while( inisize )
		table[--inisize] = NULL;
	
	for( int i=0; i<oldtablesize; i++ )
	{
		if( oldtable[i] != NULL )
			Add(oldtable[i]);
	}
	delete[] oldtable;
}

template <class Tobj>
void StringHash<Tobj>::ExpandTable()
{
	P1("Hash::ExpandTable()");
#ifdef DEBUG
	std::cout << "TABLESIZE = " << tablesize << ", ndata = " << ndata << std::endl;
#endif
	ASSERT( tablesize != 0 );
	ASSERT( (tablesize & (tablesize-1)) == 0 );
	ASSERT( table != NULL );
	
	Tobj **oldtable = table;
	int oldtablesize = tablesize;
	
	tablesize = tablesize << 1;
	table = new Tobj *[tablesize];
	ndata = 0;

	int inisize = tablesize;
	while( inisize )
		table[--inisize] = NULL;
	
	for( int i=0; i<oldtablesize; i++ )
	{
		if( oldtable[i] != NULL )
			Add(oldtable[i]);
	}
	delete[] oldtable;
}

template <class Tobj>
void FSPHash<Tobj>::ExpandTable()
{
	P1("Hash::ExpandTable()");
#ifdef DEBUG
	std::cout << "TABLESIZE = " << tablesize << ", ndata = " << ndata << std::endl;
#endif
	ASSERT( tablesize != 0 );
	ASSERT( (tablesize & (tablesize-1)) == 0 );
	ASSERT( table != NULL );
	
	Tobj **oldtable = table;
	int oldtablesize = tablesize;
	
	tablesize = tablesize << 1;
	table = new Tobj *[tablesize];
	ndata = 0;

	int inisize = tablesize;
	while( inisize )
		table[--inisize] = NULL;
	
	for( int i=0; i<oldtablesize; i++ )
	{
		if( oldtable[i] != NULL )
			Add(oldtable[i]);
	}
	delete[] oldtable;
}



template <class T> 
Queue<T>::Queue(int isize)
{ 
	list = new T[isize]; 
	read = write = 0; 
	size = isize; 
}

template <class T> 
Queue<T>::~Queue() { delete[] list; }

template <class T> 
void Queue<T>::Enqueue(T data) { 
	if( write == size )
	{
		int newsize = size + QUEUE_INISIZE;
		T *newlist = new T[newsize];
		if( list != NULL )
		{
			memcpy(newlist, list, size * sizeof(T));
			delete list;
		}
		list = newlist;
		size = newsize;
	}
	list[write++] = data;
}

template <class T> 
T Queue<T>::Dequeue(void) 
{ 
	if( write == 0 )
		return invalid;
	ASSERT( read < write );

	int oread = read;
	if( ++read == write )
		read = write = 0;
	return list[oread];
}

template <class T> 
void Queue<T>::FlushAll(void)
{
	read = write = 0;
}


// template instantiation

template class IntegerHash<feature>;
template class HashTable<type>;
template class HashTable<strsymbol>;
template class IntegerHash<procedure>;
//template class HashTable<localvar>;
template class HashTable<pinfo_var>;
template class HashTable<parsevar>;
#ifdef PROFILE
template class HashTable<profbase>;
#endif
//template class VarArray<code>;
//template class VarArray<code_mem::codeblk *>;
template class Queue<core_p>;
template class Queue< Pair<core_p, code *> >;

template class VarArray<const feature *>;
template class VarArray<const type *>;
template class VarArray<module *>;

template class IntegerHash<Int>;
template class IntegerHash< SharedPtr >;
template class IntegerHash<codepair<mint> >;
template class IntegerHash<codepair<sserial> >;
template class IntegerHash<bitmappair<mint> >;
template class IntegerHash<bitmappair<sserial> >;
// template class IntegerHash<array_std::string::array_std::string_entry>;
// template class IntegerHash<array_inthash::array_integer_entry>;
// template class IntegerHash<array_cell::array_integer_entry>;
// template class IntegerHash<array_type::array_integer_entry>;	// Added by mitsuisi
// template class FSPHash<array_fs::array_fs_entry>;
template class StringHash<token_key>;
// template class StringHash<HashCell_str_int>;
// template class IntegerHash<HashCell_int_str>;
template class StringHash<module>;    //Added by kaityo
template class StringHash<lilfes_flag>;

#ifdef TYPE_TABLE_HASH
template class IntegerHash<coercode>;

IntegerHash<coercode> coerceHash("coerceHash",4096,2);
#endif

IntegerHash<feature> feature::hash("featureHash");
IntegerHash<procedure> procedure::hash("procHash");
HashTable<strsymbol> strsymbol::hash;


Bitmap::Bitmap(unsigned is) 
{ 
	size = is; 
        ar = new int32[(size+31)/32]; 
        memset(ar, 0, ((size+31)/32) * sizeof(int32) );
}

Bitmap::Bitmap(const Bitmap &b) 
{ 
	size = b.size; 
        ar = new int32[(size+31)/32]; 
        memcpy(ar, b.ar, ((size+31)/32) * sizeof(int32) );
}

Bitmap& Bitmap::operator =(const Bitmap &b) 
{ 
	delete[] ar;
	size = b.size; 
        ar = new int32[(size+31)/32]; 
        memcpy(ar, b.ar, ((size+31)/32) * sizeof(int32) );
        return *this;
}


void Bitmap::Merge(const Bitmap &b) 
{ 
	for(uint i=0; i<(size+31)/32; i++ )
		ar[i] |= b.ar[i];
}

bool Bitmap::Ident(const Bitmap &b)
{
	if( b.size != size )
	{
		ABORT("Ident sizes are not same");
		return false;
	}
	for( uint i=0; i<(size+31)/32; i++ )
		if( b.ar[i] != ar[i] )
			return false;
			
	return true;
}

bool Bitmap::IsAllZero()
{
	for( uint i=0; i<(size+31)/32; i++ )
		if( ar[i] != 0 )
			return false;
	
	return true;
}

void Bitmap::Output(ostream &o)
{
	std::cout << "<" << size << " bit(s)>";
	for( uint i=0; i<size; i++ )
		o << (Test(i) ? "1" : "_");
}


} // namespace lilfes




/*
 * $Log: hash.cpp,v $
 * Revision 1.6  2005/05/18 10:29:34  ninomi
 * gcc3.4Ǥ⥳ѥǤ褦ˤޤ
 *
 * Revision 1.5  2004/05/30 12:29:01  yusuke
 *
 * warning message 򤱤
 *
 * Revision 1.4  2004/04/23 10:08:42  yusuke
 *
 * strstream -> sstream
 * g++3 ǤѤǤ褦ˤؤ󤳤
 *
 * Revision 1.3  2004/04/06 11:15:27  tsuruoka
 * dos2unix, nkf -e
 *
 */
