/* $Id: lex.ll,v 1.12 2004/08/18 14:59:07 yusuke Exp $
 *
 *    Copyright (c) 1997-1998, Makino Takaki
 *
 *    You may distribute this file under the terms of the Artistic License.
 *
 */

%option c++
%option yyclass="lilfesFlexLexer"
%option outfile="lex.yy.c"

%{

//#define DEBUG_INCLUDE
static const char rcsid[] = "$Id: lex.ll,v 1.12 2004/08/18 14:59:07 yusuke Exp $";
  /*  */
#define IN_LEX_LL
#include "builtin.h"
#include "in.h"
#include "lilfes.h"
#include "rl_stream.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <pwd.h>
#include <unistd.h>
#include <cassert>
//#include <FlexLexer.h>

#define YY_SKIP_YYWRAP  /* to avoid collision with namespace */

using namespace lilfes;
using namespace std;

namespace lilfes {
int token_search(char *);

char *my_strdup(char *str);

void error_open_file(char *);
void success_open_file(char *);
}

#define yylval (*yylvalp)

%}
%x COMMENT

%%

<COMMENT>"*/"  BEGIN INITIAL; 
<COMMENT>\n { /* printf("%s %d:\n", infile, lineno); fflush(stdout);*/ 	_line_no++; }
<COMMENT>.  /* eat up multi-line comments */
"%/*"        { BEGIN COMMENT; comment_start_line = Lineno(); }

"%"$                 /* eat up comments   */ 
"%"[^/\n].*$         /* eat up comments   */ 
"%/"[^*\n].*$         /* eat up comments   */ 
[\t\f\v\r ]+   /* eat up whitespace */
\n         { /*printf("%s %d:\n", infile, lineno); fflush(stdout);*/ 	_line_no++; Prompt(); }

\.\/	{ return ATTR_DEF; }

\. { parsing = false; return '.';}

[\(\)\,\[\]\|\{\}] { parsing = true; return *yytext;}

\-?[0-9]*\.[0-9]+[Ee][-+][0-9]* /* ¿E+ */ {
		parsing = true;
		yylval.dval = atof(yytext); 
		return TOK_DOUBLE; 
	}

\-?[0-9]*\.[0-9]+ /* ¿ */ {
		parsing = true;
		yylval.dval = atof(yytext); 
		return TOK_DOUBLE; 
	}


\-?(0x)?[0-9]+ /*  */ {
		parsing = true;
		yylval.ival = (int)strtol(yytext, (char**)NULL, 0); 
		return TOK_INTEGER;
	}

\"([^\"\n]|\"\")*\" /* DoubleQuote 줿ʸ */ { 
		parsing = true;
		char *p,*q; 
		yylval.str = my_strdup(yytext+1); 
		yylval.str[strlen(yylval.str)-1]='\0'; 
		p = yylval.str; 
		while((q=strstr(p,"\"\"")) != NULL) 
		{ 
			memmove(q,q+1,strlen(q)); 
			p=q+1; 
		} 
		return(TOK_STRING); 
	}

\'([^\'\n]|\'\')+\'\( /* SingleQuote 줿 ( ǽʸ */ { 
		parsing = true;
		char *p,*q; 
		yylval.str = my_strdup(yytext+1); 
		yylval.str[strlen(yylval.str)-2]='\0'; 
		p = yylval.str; 
		while((q=strstr(p,"''")) != NULL) 
		{ 
			memmove(q,q+1,strlen(q)); 
			p=q+1; 
		} 
		return(TOK_STARTP); 
	}

\'([^\'\n]|\'\')+\'\: /*SingleQuoted modulename*/ {
		parsing = true;
                char *p,*q;
                yylval.str = my_strdup(yytext+1);
                yylval.str[strlen(yylval.str)-2]='\0';
                p = yylval.str;
                while((q=strstr(p,"''")) != NULL)
                {
                        memmove(q,q+1,strlen(q));
                        p=q+1;
                }
                return(TOK_MODULENAME);
        }


\'([^\'\n]|\'\')+\'\\ /* SingleQuote 줿 \ ǽʸ */ { 
		parsing = true;
		char *p,*q; 
		yylval.str = my_strdup(yytext+1); 
		yylval.str[strlen(yylval.str)-2]='\\'; 
		yylval.str[strlen(yylval.str)-1]='\0'; 
		p = yylval.str; 
		while((q=strstr(p,"''")) != NULL) 
		{ 
			memmove(q,q+1,strlen(q)); 
			p=q+1; 
		} 
		return(TOK_FEAT); 
	}

\'([^\'\n]|\'\')+\' /* SingleQuote 줿ʸ */ { 
		parsing = true;
		char *p,*q; 
		yylval.str = my_strdup(yytext+1); 
		yylval.str[strlen(yylval.str)-1]='\0'; 
		p = yylval.str; 
		while((q=strstr(p,"''")) != NULL) 
		{ 
			memmove(q,q+1,strlen(q)); p=q+1; 
		} 
		return(TOK_ATOM); 
	}

([$A-Za-z_]|[\x80-\xff].)([$A-Za-z0-9_]|[\x80-\xff].)*\\ /* \ ǽ */ {
		parsing = true;
		yylval.str = my_strdup(yytext); 
//		yylval.str[strlen(yylval.str)-1] = '\0'; 
		return(TOK_FEAT); 
	}

([a-z]|[\x80-\xff].)([$A-Za-z0-9_]|[\x80-\xff].)*\( /* ( ǽ */ {
		parsing = true;
		yylval.str = my_strdup(yytext); 
		yylval.str[strlen(yylval.str)-1] = '\0'; 
		return(TOK_STARTP);
	}

([A-Za-z]|[\x80-\xff].)([$A-Za-z0-9_]|[\x80-\xff].)*\: /*end with : */ { 
		parsing = true;
                yylval.str = my_strdup(yytext);
                yylval.str[strlen(yylval.str)-1] = '\0';
                return(TOK_MODULENAME);
        }

([a-z]|[\x80-\xff].)([$A-Za-z0-9_]|[\x80-\xff].)* /* ʸǻϤޤ */ {
		parsing = true;
		yylval.str = my_strdup(yytext); 
		if( !strcmp(yytext, "is") ) return TOK_IS; 
		if( !strcmp(yytext, "mod") ) return TOK_MOD; 
		return(TOK_ATOM);
	}

[_$A-Z]([$A-Za-z0-9_]|[\x80-\xff].)* /* ʸǻϤޤ */ {
		parsing = true;
		yylval.str = my_strdup(yytext); 
		return(TOK_VAR);
	}

[\!\#\&\-\+\/\:\;\*\<\=\>\?\@\\\`\~\^]+ /*  */ { 
		parsing = true;
		yylval.str = my_strdup(yytext); 
		return token_search(yylval.str);
}

\' { 
		inerr(this) << "Quoted atom not terminated" << endmsg; 
		return TOK_INVALID; 
	}
\" { 
		inerr(this) << "String not terminated" << endmsg;
		return TOK_INVALID; 
	}

. { 
		inerr(this) << "Internal error: unknown input" << endmsg; 
		return TOK_INVALID; 
	}

%%
#if 0
 [\!\#\&\-\+\/\:\;\*\<\=\>\?\@\\\`\~]+\( /* ( ǽ뵭 */ {yylval.str = my_strdup(yytext); yylval.str[strlen(yylval.str)-1] = '\0'; return(TOK_STARTP);} 
#endif

/*
main() {
  yylex();
}
*/

//void SetInteractive(int i) { interactive = i; yy_set_interactive(i); }


// #define DEBUG_INCLUDE

// lilfesFlexLexer *lexer;

namespace lilfes {

int yylex(YYSTYPE *yylvalp, void *parm_) 
{ 
	parser_control *parm = (parser_control *)parm_;
	P1("Parse"); 
	if( ! parm->mach->getLexer() )
		return YY_NULL;

	while( parm->mach->getLexer()->IsSkip() )
	{
#ifdef DEBUG_INCLUDE
		std::cout << "POP_FILE " 
		  << (parm->mach->getLexer()->Next() ? (const char *)(parm->mach->getLexer()->Next()->InFile()) : "(null)") 
		  << ":" 
		  << (parm->mach->getLexer()->Next() ? parm->mach->getLexer()->Next()->Lineno() : 0) 
		  << " <- " 
		  << parm->mach->getLexer()->InFile() 
		  << ":"
		  << parm->mach->getLexer()->Lineno()
		  << std::endl;
#endif
                //std::cerr << "1: " << parm->mach->getLexer()->InFile() << std::endl;
		if( parm->mach->getLexer()->GetModule() != module::CurrentModule() )
		{
                        //std::cerr << "1: finish!" << std::endl;
			module::CurrentModule()->Finish();
			module::SetCurrentModule(parm->mach->getLexer()->GetModule());
		}
		lilfesFlexLexer *next = parm->mach->getLexer()->Next();
		delete (lilfesFlexLexer *)(parm->mach->getLexer());
		parm->mach->setLexer(next);
		if( next == NULL )
			return YY_NULL;
	}

	if( parm->mach->getLexer()->FirstRun() ) {
                //cerr << "poe";
		parm->mach->getLexer()->Prompt();
        }

	parm->mach->getLexer()->set_yylvalp(yylvalp); 
	int ret = parm->mach->getLexer()->IsSkip() ? YY_NULL : parm->mach->getLexer()->yylex(); 
	if( ret != YY_NULL )
		return ret;
	else
	{
//		return YY_NULL;
		if( parm->mach->getLexer()->Next() == NULL || parm->mach->getLexer()->ProgInput() != parm->mach->getLexer()->Next()->ProgInput() )
		{
#ifdef DEBUG_INCLUDE
		std::cout << "Finish=YY_NULL " 
		  << parm->mach->getLexer()->InFile() 
		  << ":"
		  << parm->mach->getLexer()->Lineno()
		  << std::endl;
#endif
                        parm->mach->getLexer()->Skip();
			return YY_NULL;
		}
		else
		{
#ifdef DEBUG_INCLUDE
		std::cout << "Finish=EOF " 
		  << parm->mach->getLexer()->InFile() 
		  << ":"
		  << parm->mach->getLexer()->Lineno()
		  << std::endl;
#endif
			parm->mach->getLexer()->Skip();
			return TOK_EOF;
		}
	}
}

int eval(machine *m, const char *str, bool progvars)
{
#ifdef DEBUG_INCLUDE
		std::cout << "Start Eval " 
		  << lexer->InFile() 
		  << ":"
		  << lexer->Lineno()
		  << std::endl;
#endif
	std::istringstream *iss = new std::istringstream( str );
	lilfesFlexLexer *lex = new lilfesFlexLexer( NULL, iss, true, "eval", "", 1 );
	prog_list = prog_list_top = FSP(m);
	prog_vars = prog_vars_top = FSP(m);
#ifdef DEBUG_INCLUDE
		std::cout << "Eval Lexer created " 
		  << lex->InFile() 
		  << ":"
		  << lex->Lineno()
		  << std::endl;
#endif
	lex->SetProgVars(progvars);
//	while( lex->ProgInput() )
//	{
		m->parse(lex);
//	}
	prog_list.Coerce(nil);
	prog_vars.Coerce(nil);
	return 1;
}

void search_path::ExtractRelativePath(const char *str)
{
	lstring l = str;
	if( strcmp(str, "stdin") == 0 )
		l = "./";
	else {
		char *p;
		p = strrchr(l, '/');
		if( p == NULL )
			p = strrchr(l, '\\');

		if( p == NULL )
			l = "";
		else
			*(p+1) = '\0';		// cut off filename
	}
	RelativeSearchPath()->name = l;
}

void search_path::AddSearchPath(const char *path)
{
	if( path != NULL && *path != '\0' )
	{
		char *curaddpath;
		char *p = strchr(path, ':');
		if( p )
		{
			AddSearchPath(p+1);
			curaddpath = new char [p-path+2];
			strncpy(curaddpath, path, p-path);
			curaddpath[p-path] = '\0';
		}
		else
		{
			curaddpath = new char [strlen(path)+2];
			strcpy(curaddpath, path);
		}

		char c = curaddpath[strlen(curaddpath)-1];
		if( c != '/' && c != '\\' )
		{
			strcat(curaddpath, "/");
		}

		default_search_path = new search_path(curaddpath, default_search_path);

		delete[] curaddpath;
	}
}

void search_path::init()
{
	default_search_path = NULL;

	search_path::AddSearchPath(DEFAULT_SEARCH_PATH);
	search_path::AddSearchPath(getenv("LILFES_PATH"));
}

search_path* search_path::default_search_path;
search_path search_path::minimum_search_path("");
search_path search_path::relative_search_path("./");

static InitFunction IF(search_path::init);

void search_path::term()
{
}

static TermFunction TF(search_path::term);

namespace builtin {

bool get_search_path( machine& m, FSP arg1 )
{
	search_path::SetRelativePath("(current file relative)");
	search_path *s = search_path::SearchPath();
	while( s )
	{
		if(	arg1.Follow(hd).Unify(FSP(m,s->GetName() )) == false ||
			! (arg1 = arg1.Follow(tl)).IsValid() )
		{
			return false;
		}
		s = s->Next();
	}
	return arg1.Coerce(nil) ? true : false;
}

LILFES_BUILTIN_PRED_1(get_search_path, get_search_path);

} // namespace builtin

// const char *infile() { return lexer ? lexer->InFile() : "(closed)"; }
// int lineno() { return lexer ? lexer->Lineno() : -1; }
// int proginput() { return lexer ? lexer->ProgInput() : 0;} 


bool lilfesFlexLexer::Isatty() {
  if( IsSkip() && next )
    return next->Isatty();
  else
    return (s == &std::cin ||
            s == &rl_istream::rl_stdin) &&
      isatty(0);
}

void lilfesFlexLexer::Prompt()
{
  if ( s == &rl_istream::rl_stdin ) {
    rl_istream::rl_stdin.set_prompt( parsing? "| " : "> " );
  }
//	if( Isatty() ) 
//		std::cout << (parsing ? "| " : "> "); 
//	else
//		std::cout << (parsing ? "(|) " : "(>) "); 
//	else
//		std::cout << "(" << InFile() << ") ";
}

//#include <stdio.h>

void lilfesFlexLexer::Flush()
{
	parsing = false; 
	//YY_FLUSH_BUFFER();
	s->sync();
	while(s->rdbuf()->in_avail())
		if(s->get() == '\n')
			break;
	SetFirstRun();
}


int pushfile(machine *mach, const char *fname, int pi, const char *prefix)
{
#ifdef DEBUG_INCLUDE
	{
		char *buf = getcwd(NULL, 2048);
		std::cout << "Trying to open file " << fname << ", cwd = " << buf << std::endl;
		free(buf);
	}
#endif
	const search_path *search_path_ptr = search_path::SearchPath();
	if( fname[0] == '/' || (fname[0] == '-' && fname[1] == '\0' ))
	{
		search_path_ptr = search_path::MinimumSearchPath();
	}
	while( search_path_ptr )
	{
		lstring fn = search_path_ptr->GetName();
		fn = fn + fname;
		lstring fn1 = fn;
		if( *fn == '~' ) 
		{
			if( fn[1] == '/' )
			{
				char *home = getenv("HOME");
				if( home == NULL )
					home = "/";
				fn = lstring(home) + (fn+1);
			}
			else
			{
				char *p = strchr(fn, '/');
				if( p != NULL )
				{
					*p = '\0';
					p++;
				}
				struct passwd *pwd = getpwnam(fn+1);
				if( pwd == NULL ) goto CONTINUE;
				fn = lstring(pwd->pw_dir) + "/" + (p ? p : "");
			}
		}
#ifdef DEBUG_INCLUDE
		std::cout << "look: " << fn << std::endl;
#endif
		std::istream *f;
		bool shouldDelete;
		if( ! strcmp(fn, "-") )
		{
			f = &std::cin;
			shouldDelete = false;
		}
		else
		{
			f = new std::ifstream(fn);
			shouldDelete = true;
		}
		if( f->good() )
		{
#ifdef DEBUG_INCLUDE
			std::cout << "open: " << fn << std::endl;
#endif
			mach->setLexer(new lilfesFlexLexer(mach->getLexer(), f, shouldDelete, fn1, prefix, pi));

#ifdef DEBUG_INCLUDE
			std::cout << "opened: " << fn << std::endl;
#endif

	#ifdef DEBUG_INCLUDE
		std::cout << "PUSH_FILE success " 
		  << (mach->getLexer()->Next() ? (const char *)(mach->getLexer()->Next()->InFile()) : "(null)") 
		  << ":" 
		  << (mach->getLexer()->Next() ? (mach->getLexer()->Next()->Lineno()) : 0) 
		  << " -> " 
		  << fn << ":1" << std::endl;
	#endif

			return 1;
		}
#ifdef DEBUG_INCLUDE
		else
		{
			perror("fail");
		}
#endif
	CONTINUE:
		search_path_ptr = search_path_ptr->Next();
	}
#ifdef DEBUG_INCLUDE
		cout << "PUSH_FILE fail " << endl;
#endif
	return 0;
}

int readfile(machine *m)
{
	while( m->getLexer()->ProgInput() )
	{
		parse(m);
	}
	return 1;
}

lexer_t::lexer_t(lilfesFlexLexer *il) : l(il) 
{
	if(l) l->incRef();
}

lexer_t::lexer_t(const lexer_t &lx) : l(lx.l)
{
	if(l) l->incRef();
}

lexer_t::lexer_t(std::istream *is, const char *filename, int iproginput )
{
	l = new lilfesFlexLexer(is, filename, iproginput);
//	l->incRef();   // now it is done in the constructor of lilfesFlexLexer
}
lexer_t::~lexer_t()
{
	if( l && l->decRef() == 0 )
		delete l;
}

} // namespace lilfse

int yywrap(void)
{
	return 1;
}


/*
 * $Log: lex.ll,v $
 * Revision 1.12  2004/08/18 14:59:07  yusuke
 *
 * c_to_lilfes  vector< cell > б
 * load_module Ʊ⥸塼ʣ load  segmentation fault ˤʤΤ
 *
 * Revision 1.11  2004/05/30 08:40:27  yusuke
 *
 * memory leak check ΥץˤȤʤmemory leak 
 * 餷
 *
 * Revision 1.10  2004/05/18 08:14:57  yusuke
 *
 * eval/3 򤸤ä
 *
 * Revision 1.9  2004/04/23 08:38:41  yusuke
 *
 * std:: Ĥ
 *
 * Revision 1.8  2004/04/20 03:07:51  yusuke
 *
 * readline 򤯤ߤ
 *
 * Revision 1.7  2004/04/19 15:39:34  yusuke
 *
 * builtin Τ礯礦
 *
 * Revision 1.6  2004/04/08 06:23:29  tsuruoka
 * lexer_t: modified the initialization method of reference counting
 *
 */

