/* $Id: lstring.cpp,v 1.3 2004/04/23 15:36:26 yusuke Exp $ 
 *
 *    Copyright (c) 1997-1998, Makino Takaki
 *
 *    You may distribute this file under the terms of the Artistic License.
 *
 */

#include "lstring.h"
#include <sstream>
#include <string.h>

#ifdef USE_MEMALLOC
#include "memalloc.h"

lilfes_allocator<char> lstring_alloc;
#endif

namespace lilfes {

static const char rcsid[] = "$Id: lstring.cpp,v 1.3 2004/04/23 15:36:26 yusuke Exp $";

lstring::lstring(const char *s) 
{
	if( s == NULL )
		str = NULL;
	else {
#ifndef USE_MEMALLOC
		str = new char[strlen(s)+1];
#else
		str = lstring_alloc.allocate(strlen(s)+1);
#endif
		strcpy(str, s);
	}
}

lstring::lstring(const lstring &s) 
{
	if( s.str == NULL )
		str = NULL;
	else {
#ifndef USE_MEMALLOC
		str = new char[strlen(s.str)+1];
#else
		str = lstring_alloc.allocate(strlen(s.str)+1);
#endif
		strcpy(str, s.str);
	}
}

lstring::lstring(char c, int repeat) 
{
#ifndef USE_MEMALLOC
	str = new char[repeat+1];
#else
	str = lstring_alloc.allocate(repeat+1);
#endif
	memset(str, c, repeat);
	str[repeat] = '\0';
}

lstring::~lstring() 
{
#ifndef USE_MEMALLOC
	delete[] str;
#else
	if( str != NULL )
		lstring_alloc.deallocate(str, strlen(str)+1);
#endif
}

lstring &lstring::operator =(const lstring &s)
{
#ifndef USE_MEMALLOC
	delete[] str;
#else
	if( str != NULL )
		lstring_alloc.deallocate(str, strlen(str)+1);
#endif
	if( s.str == NULL )
		str = NULL;
	else {
#ifndef USE_MEMALLOC
		str = new char[strlen(s.str)+1];
#else
		str = lstring_alloc.allocate(strlen(s.str)+1);
#endif
		strcpy(str, s.str);
	}
	return *this;
}

lstring &lstring::operator =(const char *s)
{
#ifndef USE_MEMALLOC
	delete[] str;
#else
	if( str != NULL )
		lstring_alloc.deallocate(str, strlen(str)+1);
#endif
	if( s == NULL )
		str = NULL;
	else {
#ifndef USE_MEMALLOC
		str = new char[strlen(s)+1];
#else
		str = lstring_alloc.allocate(strlen(s)+1);
#endif
		strcpy(str, s);
	}
	return *this;
}

lstring operator+(const lstring &a, const lstring &b)
{
	lstring ret;
	int alen = strlen(a.str);
#ifndef USE_MEMALLOC
	ret.str = new char[alen+strlen(b.str)+1];
#else
	ret.str = lstring_alloc.allocate(alen+strlen(b.str)+1);
#endif
	strcpy( ret.str, a.str );
	strcpy( ret.str + alen, b.str);
	return ret;
}

lstring operator+(const char *a, const lstring &b)
{
	lstring ret;
	int alen = strlen(a);
#ifndef USE_MEMALLOC
	ret.str = new char[alen+strlen(b)+1];
#else
	ret.str = lstring_alloc.allocate(alen+strlen(b)+1);
#endif
	strcpy( ret.str, a );
	strcpy( ret.str + alen, b.str);
	return ret;
}

lstring operator+(const lstring &a, const char *b)
{
	lstring ret;
	int alen = strlen(a);
#ifndef USE_MEMALLOC
	ret.str = new char[alen+strlen(b)+1];
#else
	ret.str = lstring_alloc.allocate(alen+strlen(b)+1);
#endif
	strcpy( ret.str, a );
	strcpy( ret.str + alen, b );
	return ret;
}

lstring int2lstr(int i)
{
	std::ostringstream oss;
	oss << i << std::endl;
	return lstring(oss.str().c_str());
}

} // namespace lilfes
