/* $Id: main.cpp,v 1.22 2005/02/12 04:11:38 yusuke Exp $
 *
 *    Copyright (c) 1997, Makino Takaki
 *
 *    You may distribute this file under the terms of the Artistic License.
 *
 */

#include <liblilfes/lilfes.h>
#include <liblilfes/builtin.h>
#include <liblilfes/option.h>
#include <liblilfes/rl_stream.h>
#include <liblilfes/in.h>

#include <iostream>
#include <setjmp.h>
#include <exception>
#include <sstream>

static const char rcsid[] = "$Id: main.cpp,v 1.22 2005/02/12 04:11:38 yusuke Exp $";

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#else
#  include <io.h>
#endif

int main(int argc, char **argv)
{
	try {
		*lilfes::error_stream << lilfes::Version ;
		*lilfes::error_stream << lilfes::Copyright << std::endl;

		// Initialize everything
		lilfes::Initializer::PerformAll();

		// Initialize a machine object
		lilfes::machine m;

                // Main routine
                int return_code = 0;
                try {

                  // Apply command-line options to the machine (Not needed when you don't use command line)
                  std::istringstream dummy_stream("");
		  lilfes::lexer_t cl(&dummy_stream, "stdin");
		  m.setLexer(cl);
                  lilfes::lilfes_flag::ParseOptions(&m, argc, argv);
		  m.parse();
                  lilfes::module::SetCurrentModule(lilfes::module::UserModule());

#if 0
                  // Sample of declaring new type
                  lilfes::feature *f1 = new lilfes::feature("testfeat\\", lilfes::module::UserModule() /*, 0, true*/ );
                  lilfes::type *t1 = new lilfes::type("testtype", lilfes::module::UserModule() /*, true*/ );
                  t1->SetAsChildOf(lilfes::bot);
                  t1->AddFeature(f1 /*, lilfes::t_int*/);
                  t1->Fix();

                  // Sample of declaring new predicate type
                  lilfes::type *t = new lilfes::type("testpred", lilfes::module::UserModule() /*, true*/);
                  t->SetAsChildOf(lilfes::t_pred[1]);
                  t->Fix();

                  // sample of machine::eval 
                  lilfes::FSP val;
                  val = m.eval("[3, 4, testtype & testfeat\\ 5, 6, 7].");
                  if( val.IsValid() )
                    std::cout << val.DisplayAVM();
                  else
                    *error_stream << "Error eval!" << std::endl;

                  // sample of machine::call, FSP::coerce, FSP::follow
                  lilfes::FSP f(m);
                  f.Coerce(lilfes::module::UserModule()->Search("assert"));
                  f.Follow(lilfes::f_arg[1]).Coerce(t);
                  f.Follow(lilfes::f_arg[1]).Follow(lilfes::f_arg[1]).Unify(val.Follow(lilfes::hd));
                  std::cout << f.DisplayAVM();
                  bool ret = m.call(f);
                  *error_stream << (ret ? "Good" : "Bad" ) << std::endl;

                  // sample of direct builtin call
                  lilfes::FSP g(m);
                  g.Coerce(t);
                  g.Follow(lilfes::f_arg[1]).Unify(g);
                  lilfes::builtin::lilfes_assert( m, g );

                  std::cout << "Type '?- testpred(X).' to see results." << std::endl;
#endif // if 0

		  // sample of lexer parsing (parsing from stdin)
		  std::string s;
		  if( !lilfes::lilfes_flag::ExecOptions(m) ){
		    lilfes::lexer_t lexer(&lilfes::rl_istream::rl_stdin, "stdin");
		    m.parse(lexer);
		  }

                } // try

                catch ( lilfes::halt_exception& e ) {
                  return_code = e.GetReturnCode();
                }

		// Termination process
		lilfes::Terminator::PerformAll();
		
		return return_code;
	} 
	catch (std::bad_alloc)
	{
                *lilfes::error_stream << "Fatal error: Memory allocation failed, LiLFeS terminated." << std::endl;
		throw;
	}
	catch (std::exception &e)
	{
                *lilfes::error_stream << "Fatal error: " << e.what() << std::endl;
		throw;
	}
}


/*
 * $Log: main.cpp,v $
 * Revision 1.22  2005/02/12 04:11:38  yusuke
 *
 * version 1.2.6.
 * socket, process stream  iostream Ȥ
 * ˤȤʤä lilfes_XXX_stream μɡ
 * ФȤ ios Ĥ褦ˤ
 * write, read ʤɤ iostream  write, read Ȥ褦ˤ
 *  cin, cout, cerr ľܻȤäƤȤ
 * input_stream, output_stream, error_stream Ȥ褦ѹ
 *
 * Revision 1.21  2004/09/20 08:53:37  n165
 * ParseOptionsexit()ƤФʤ褦˽
 *
 * Revision 1.20  2004/09/16 09:02:15  n165
 * -e option enabled.
 *
 * Revision 1.19  2004/05/31 08:53:56  yusuke
 *
 * դΤʤǤ礯 halt ֤ exception  catch 줺 abort ƤޤФ򤷤夦
 *
 * Revision 1.18  2004/05/30 08:40:27  yusuke
 *
 * memory leak check ΥץˤȤʤmemory leak 
 * 餷
 *
 * Revision 1.17  2004/05/18 07:28:07  yusuke
 *
 * halt 򤸤ä
 * ll -> lilfes
 *
 * Revision 1.16  2004/05/11 09:07:10  yusuke
 *
 * ƤȤ褦פ򤱤
 *
 * Revision 1.15  2004/04/23 08:19:18  yusuke
 *
 * g++3 ǤѤǤ褦ˤ
 *
 * Revision 1.14  2004/04/20 03:07:51  yusuke
 *
 * readline 򤯤ߤ
 *
 * Revision 1.13  2004/04/19 08:46:44  yusuke
 *
 * builtinfs.cpp  new interface ˤ
 *
 * Revision 1.12  2004/04/19 08:04:31  yusuke
 *
 * xml, string, option, builtin  builtin 򤢤餷 interface ˤؤ󤫤
 *
 * Revision 1.11  2004/04/12 06:07:01  tsuruoka
 * fixed kanji code
 *
 * Revision 1.10  2004/04/11 05:42:27  yusuke
 *
 * gcc 3.2 Ǥ⥳ѥǤ褦ˤ
 *
 * Revision 1.9  2004/04/09 07:50:54  tsuruoka
 * to avoid Segmentation fault when trying to load a nonexistent file
 *
 * Revision 1.8  2004/04/08 10:16:26  tsuruoka
 * to properly handle commandline options
 *
 * Revision 1.7  2004/04/08 06:23:29  tsuruoka
 * lexer_t: modified the initialization method of reference counting
 *
 * Revision 1.6  2004/04/06 11:05:12  tsuruoka
 * dos2unix, nkf -e
 *
 */
