/**
 * @module  = misc
 * @cvs     = $Id: misc.cpp,v 1.2 2004/05/15 03:36:29 yusuke Exp $
 * @copyright = Copyright (c) 1997-1998, Makino Takaki
 * @copyright = You may distribute this file under the terms of the Artistic License.
 * @desc    = Miscellaneous
 * @jdesc   = ¾
 * ¾ȹ߽ҸǤ
*/

#include "builtin.h"
#include <sstream>

namespace lilfes {

/**
  * @predicate	= lilfes_version($X) 
  * @desc		= Return lilfes version.
  * @param		= +$X/bot: the variable which get the lilfes version
  * @example	= 
  > ?- lilfes_version($X).
  X: "LiLFeS 1.2          by Takaki Makino and Tsujii lab., Tokyo Univ.
  Last Compiled on May 14 2004, 20:22:09 [PROF=1]"

  Return lilfes version.
  * @end_example

  * @jdesc		= LiLFeSΥС֤ޤ
  * @jparam		= +$NAME/bot :LiLFeSΥСǼѿ
  * @jexample	=
  > ?- lilfes_version($X).
  X: "LiLFeS 1.2          by Takaki Makino and Tsujii lab., Tokyo Univ.
  Last Compiled on May 14 2004, 20:22:09 [PROF=1]"

  LiLFeSΥС֤ޤ
  * @end_jexample
*/

extern const char *Version;
extern const char *Copyright;
extern const char *LastCompiled;

namespace builtin {

bool lilfes_version( machine&, FSP arg1 )
{

        std::ostringstream oss;
	oss << Version << Copyright << std::endl;
	oss << "Last Compiled on " << LastCompiled
#if defined( DEBUG )
	<< " [DEBUG]"
#elif defined( DOASSERT )
	<< " [DOASSERT]"
#endif
#if defined( PROFILE )
	<< " [PROF=" << PROFILE << "]"
#endif
	<< std::ends;

        return arg1.Unify(oss.str().c_str());
}

LILFES_BUILTIN_PRED_1(lilfes_version, lilfes_version);

/**
  * @predicate	= lilfes_version($X, $Y)
  * @desc		= Return lilfes version.
  * @param		= +$X/bot: lilfes version
  * @param		= +$X/bot: last compiled time
  * @example	= 
  > ?- lilfes_version($X).
  X: "LiLFeS 1.2"
  Y: "May 14 2004, 20:22:09"

  Return lilfes version.
  * @end_example

  * @jdesc		= LiLFeSΥС֤ޤ
  * @jparam		= +$X/bot :LiLFeSΥСǼѿ
  * @jparam		= +$Y/bot :ѥ뤵줿֤Ǽѿ
  * @jexample	=
  > ?- lilfes_version($X).
  X: "LiLFeS 1.2"
  Y: "May 14 2004, 20:22:09"

  LiLFeSΥС֤ޤ
  * @end_jexample
*/

bool lilfes_version_2( machine&, FSP arg1, FSP arg2 )
{
  return arg1.Unify( Version ) && arg2.Unify( LastCompiled );
}

LILFES_BUILTIN_PRED_OVERLOAD_2(lilfes_version_2, lilfes_version_2, lilfes_version);

} // namespace builtin

} // namespace lilfes

