/**
 * @module  = module
 * @cvs     = $Id: module.cpp,v 1.16 2005/05/18 10:29:34 ninomi Exp $
 * @copyright = Copyright (c) 1997-1998, Makino Takaki
 * @copyright = You may distribute this file under the terms of the Artistic License.
 * @desc    = Module
 * @jdesc   = ⥸塼
 * Directives and predicates for module control
 * @japanese =
 * ⥸塼椹뤿Υǥ쥯ƥ֤Ȥ߹߽ҸǤ
 * @end_japanese
*/

/* $Id: module.cpp,v 1.16 2005/05/18 10:29:34 ninomi Exp $
 *
 *    
 *      made by kaityo
 *    You may distribute this file under the terms of the Artistic License.
 *
 */

static const char rcsid[] = "$Id: module.cpp,v 1.16 2005/05/18 10:29:34 ninomi Exp $";

#include "module.h"
#include "initial.h"
#include "option.h"
#include "builtin.h"
#include "in.h"
#include <stdio.h>

//#define DEBUG_MODULE

//#ifdef DEBUG
//#define DEBUG_MODULE
//#endif

#define MODLIST_GROW_BITS 3

namespace lilfes {

template <> uint VarArray<module *>::growstep = (1 << MODLIST_GROW_BITS)  ;
template <> uint VarArray<module *>::growmask = (1 << MODLIST_GROW_BITS)-1;
template <> uint VarArray<module *>::growbits =       MODLIST_GROW_BITS   ;

template <> StringHash<module> module::modulelist_hash;

module *module::m_user;
module *module::m_core;
module *module::m_builtin;
module *module::m_directive;
module *module::m_current;
lstring module::module_prefix("");

////////////////////////////////////////////////////////////////////////

void get_fileinfo(const char *filename, lstring &lilname, lstring &modulename);

////////////////////////////////////////////////////////////////////////
//
//  class module
//

module::module(const char *iname, new_module_type inmt /* = NMT_NONE */, bool Extendable /* = false */) 
	: name(""), defname(module_prefix + iname), ExtendableMode(Extendable)
{ 
	nmt = inmt;
	InterfaceMode = true;
	prev_module = NULL;
	first_import_modules = NULL;
	if( nmt == NMT_BUILTIN )
		Register();
}

void module::SetFirstImport(module** mods) {
	if (mods == 0) {
		first_import_modules = new module*[1];
		first_import_modules[0] = 0;
		return;
	}
	int len = 0; module** cur = mods;
	while(*cur != 0) { cur++; len++; }
	first_import_modules = new module*[len+1];
	int i; for (i=0; i<len; i++) first_import_modules[i] = mods[i];
	first_import_modules[i] = 0;
}

void module::Delete(type *t) 
{ 
	if( t->IsDeleted() ) 
		hash.Delete(t); 
	else 
		ABORT("INTERNAL ERROR"); 
}

const feature * module::SearchFeature(const char *str) const {
  ASSERT(this != NULL);
  return feature::Search(hash.Find(str));
}

bool module::IsUsable(const type *t) const
{ 
	return t == Search(t->GetSimpleName()); 
}

bool module::IsExporting(const type *t) const
{
	if( ! t->IsExported() )
		return false;
	if( t->GetModule() == this )
		return true;
	if( ! IsUsable(t) )
		return false;
	for( uint i=0; i<reexportlist.GetSize(); i++ )
		if( t->GetModule() == reexportlist[i] )
		{
			return true;
		}
	return false;
}

void module::ShowList(void)
{
	if( hash.GetNData() )
	{
		std::cout << "Exporting Symbols: " << std::endl;
		for ( uint i=0;i<hash.GetTableSize();i++ )
		{
			type *t = hash.Pick(i);
			if (t!=NULL && IsExporting(t))
				std::cout << t->GetName() << " ";
		}
		std::cout << std::endl;

		std::cout << "Non exporting symbols: " << std::endl;
		for ( uint i=0;i<hash.GetTableSize();i++ )
		{
			type *t = hash.Pick(i);
			if (t!=NULL && t->GetModule() == this && !IsExporting(t))
				std::cout << t->GetName() << " ";
		}

		std::cout << std::endl;
		std::cout << "All symbols: " << std::endl;

		for ( uint i=0;i<hash.GetTableSize();i++ )
		{
			type *t = hash.Pick(i);
			if (t!=NULL)
				std::cout << t->GetProperName() << " ";
		}
		std::cout << std::endl;
	}

	if( reexportlist.GetSize() )
	{
		std::cout << "Re-exporting modules: " << std::endl;
		for( uint i=0; i<reexportlist.GetSize(); i++ )
			std::cout << reexportlist[i]->GetName() << " ";
		std::cout << std::endl;
	}
	if( noexportlist.GetSize() )
	{
		std::cout << "Not-exporting modules: " << std::endl;
		for( uint i=0; i<noexportlist.GetSize(); i++ )
			std::cout << noexportlist[i]->GetName() << " ";
		std::cout << std::endl;
	}

	std::cout << "List of all modules: " << std::endl;
	for( uint i=0; i<modulelist_hash.GetTableSize(); i++ )
		if( modulelist_hash.Pick(i) )
		{
			module *p = modulelist_hash.Pick(i);
			std::cout << p->GetName() << " ";
		}
	std::cout << std::endl;
}

module** module::ExportList(void) const
{
	module** mods = new module*[reexportlist.GetSize()+1];
	uint i; for (i=0; i<reexportlist.GetSize(); i++) {
		mods[i] = reexportlist[i];
	}
	mods[i] = 0;
	return mods;
}

module** module::NoExportList(void) const
{
	module** mods = new module*[noexportlist.GetSize()+1];
	uint i; for (i=0; i<noexportlist.GetSize(); i++) {
		mods[i] = noexportlist[i];
	}
	mods[i] = 0;
	return mods;
}

void module::Import(module *mod, bool bexport)
{
	ASSERT(mod != this);

#ifdef DEBUG_MODULE
		std::cout << "Module " << GetName() << "(" << GetDefName() << ") imports " << mod->GetName() << " (reexport " << bexport << ")" << std::endl;
#endif

	if( bexport )
	{
		for( uint i=0; i<reexportlist.GetSize(); i++ )
			if( reexportlist[i] == mod )
			{
#ifdef DEBUG_MODULE
				std::cout << "SAME MODULE IMPORT; ignored" << std::endl;
#endif
				return;	
			}
		reexportlist.Next() = mod;
		for( uint i=0; i<mod->reexportlist.GetSize(); i++ )
		{
			for( uint j=0; j<reexportlist.GetSize(); j++ )
			{
				if( reexportlist[j] == mod->reexportlist[i] )
					goto CONTINUE;
			}
			reexportlist.Next() = mod->reexportlist[i];
		CONTINUE:
			;
		}
	} else {
		for( uint i=0; i<noexportlist.GetSize(); i++ )
			if( noexportlist[i] == mod )
			{
#ifdef DEBUG_MODULE
				std::cout << "SAME MODULE IMPORT; ignored" << std::endl;
#endif
				return;	
			}
		noexportlist.Next() = mod;
	}
	
	for( uint i=0; i<mod->hash.GetTableSize(); i++ )
	{
		type *t = mod->hash.Pick(i);
		if( t != NULL && mod->IsExporting(t) )
		{
#ifdef DEBUG_MODULE
			if( DEB )
			if( mod != BuiltinModule() && mod != CoreModule() )
			std::cout << ">>" << t->GetName() << std::endl;
#endif
			type *oldt = hash.AddOrReplace(t);

			if( oldt != NULL )
			{
				if( oldt->GetModule() == this )
				{
					hash.AddOrReplace(oldt);		// Replace again
					RUNWARN( "The imported type " << t->GetProperName() << " cannot override the same-named intrinsic type in module " << GetName());
				}
				else if( oldt->GetModule() != t->GetModule() )
				{
					RUNWARN( "The imported type " << t->GetProperName() << " overrides previously imported type " << oldt->GetProperName() << " in module " << GetName());
				}
			}
		}
	}
}


void module::SetCurrentModule(module *mod) 
{ 
	ASSERT(mod != NULL);
#ifdef DEBUG_MODULE
	std::cout << "current module changed " << (m_current == NULL ? "(null)" : (const char *)m_current->name) << " -> " << mod->name << std::endl;
#endif
	m_current = mod; 
}

InitFunction module::IF(module::init, 5);
InitFunction module::IF2(module::init2, 300);

void module::init()		// This initialization should be done 
						// ahead of init() in ftype.cpp
{
	m_core = new module("core", NMT_BUILTIN, true);
	m_user = new module("user", NMT_BUILTIN, true);
	m_builtin = new module("builtin", NMT_BUILTIN);
	m_directive = new module("builtin:directive", NMT_BUILTIN);
	m_current = m_user;

}

void module::init2()
{
	m_builtin->Import(m_directive, true);
	m_user->Import(m_core, false);
	m_user->Import(m_builtin, false);
}


void module::term()		// This initialization should be done 
						// ahead of init() in ftype.cpp
{
	for(uint i=0; i<modulelist_hash.GetTableSize(); i++ )
	{
		module *m = modulelist_hash.Pick(i);
		if( m )
		{
			delete m;
		}
	}
}

TermFunction module::TF(module::term, 5);

////////////////////////////////////////////////////////////////////////
//
//  Built-in functions for module manipulation
//

//////////////////////////////////////////////////////////////////////////////
// ensure_loaded/1 : include a file


module* module::new_module(machine &m, const char *biname, lstring modname, new_module_type nmt, module *prevmod /* = NULL */, module** import_module /* = NULL */)
{
#ifdef DEBUG_MODULE
	std::cout << "new_module " << modname << " (type " << nmt << ") in " << biname << std::endl;
	if( prevmod ) 
		std::cout << " ( prevmod = " << prevmod->GetName() << " )" << std::endl;
#endif

        // Unnecessary?  removed by yusuke (2004/09/15)
        /*
	if( nmt == NMT_ENSURE || nmt == NMT_ENSURE_INITIAL || nmt == NMT_ENSURE_IN_PROG || nmt == NMT_ENSURE_WITHOUT_BUILTIN || nmt == NMT_ENSURE_WITHOUT_BUILTIN_IN_PROG)
	{	// ensure Ϥξ
		module *mod = module::SearchModule(CurrentModule()->GetModulePrefix()+modname);
//		if( mod && mod->GetModuleType() == NMT_BUILTIN )
		if( mod )	// ⥸塼뤬¸ߤƤ
		{
			module::CurrentModule()->Import( mod );	// ݡȤ
			//return modname;
                        return mod;
		}
	}
        */

	module *prev_current = module::CurrentModule();
	module *mod = NULL;
	
	lstring lilname, modulename;
	get_fileinfo(modname, lilname, modulename);

	if( nmt != NMT_INCLUDE && nmt != NMT_READ_LILFES )
	{	// ⥸塼˴طҸξ
		mod = new module( modulename, nmt ); 			// ⥸塼
		mod->Import( module::CoreModule(), false );		// Core ⥸塼Ͼ˥ݡ
		if( nmt != NMT_READ_MODULE &&
		    nmt != NMT_ENSURE_WITHOUT_BUILTIN &&
		    nmt != NMT_ENSURE_WITHOUT_BUILTIN_IN_PROG)
		{												// ξʳ Builtin 򥤥ݡ
			if ( prev_current->first_import_modules == 0) {		// first_import ꤬ʤ
				mod->Import( module::BuiltinModule(), false );	// ̤ builtin 򥤥ݡ
			} else {											// ꤬
				mod->Import( module::DirectiveModule(), false );	// Directive ϥݡ
				mod->SetFirstImport(prev_current->first_import_modules);	// first_import Ѥ
				module** cur = prev_current->first_import_modules;
				while(*cur != NULL)								// first_import 򤹤
					mod->Import( *cur++, false);
			}
		}
		else											// Builtin 򥤥ݡȤʤ
		{
			mod->Import( module::DirectiveModule(), false );	// Directive ϥݡ
			mod->SetFirstImport(import_module);			// import_module 򥻥å (NULL ξǤ⤽Τޤޥå)
			if (import_module != NULL) {
				while(*import_module != NULL)
					mod->Import( *import_module++, false);
			}
		}
	}

	if( nmt == NMT_CREATE_MODULE )						// create_module ξϤǤ
	{
		module::SetCurrentModule( mod );
		//return mod ? lstring(mod->GetName()) : modulename;
                return mod;
	}

	if( nmt != NMT_ENSURE_INITIAL && m.getLexer() )	{		// ʳ
		search_path::ExtractRelativePath(m.getLexer()->InFile());		// Хѥ򸡺
        } else {
          search_path::SetRelativePath("./");
        }

	if( nmt == NMT_READ_LILFES || nmt == NMT_READ_LILFES_WITH_PREFIX || nmt == NMT_READ_MODULE )
	{
		/* readfile() ƤŸ */
#ifdef DEBUG_INCLUDE
	        std::cout << "READ_FILE" <<  fname << std::endl;
#endif
		if(! pushfile( &m, lilname, 1, (nmt == NMT_READ_LILFES_WITH_PREFIX ? "__" : "")) )
		{
			INERR(&m, "Cannot open file '" << modname << "' in " << biname);
			delete mod;
			m.Fail();
			//return lstring("");
                        return NULL;
		}
		else
		{
			module::SetCurrentModule( mod );
			mod->SetPrevModule(prevmod ? prevmod : prev_current);
			readfile( &m );
#ifdef DEBUG_INCLUDE
		        std::cout << "END_READ_FILE\n";
#endif
		}
	}
	else
	{
		if( (! pushfile( &m, lilname, ( (nmt == NMT_ENSURE_IN_PROG || nmt == NMT_ENSURE_WITHOUT_BUILTIN_IN_PROG) ? 2 : 0) ) ) )
		{
			if( nmt != NMT_INITIALIZE )
				INERR(&m, "Cannot open file '" << modname << "' in " << biname);
			delete mod;
			m.Fail();
			//return lstring("");
                        return NULL;
		}
		else
		{
			if( mod != NULL )
			{
				module::SetCurrentModule( mod );
				mod->SetPrevModule(prevmod ? prevmod : prev_current);
			}
		}
	}

//		if( nmt == NMT_INITIALIZE || nmt == NMT_ENSURE_IN_PROG )
//		{
//			m.getLexer()->SetDepth(1);
//		}
	//return mod ? lstring(mod->GetName()) : modulename;
        return mod;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

namespace builtin {

/**
  * @predicate  = ensure_loaded(+ModulePath)
  * @desc       = Load a module .
  * @param      = +ModulePath/ std::string  : Module path
  * @example    =
  > :- ensure_loaded("lillib/lists").
  * @end_example

  * @jdesc      = ⥸塼ɤ߹ߤޤ
  * @jparam     = +ModulePath/ std::string: ⥸塼Υѥ
  * @jexample   =
  > :- ensure_loaded("lillib/lists").
  * @end_jexample
*/
bool ensure_loaded( machine& m, FSP arg1 )
{
	if( ! arg1.IsString() )
	{
          INERR(&m, "ensure_loaded/1 requires a file name");
          return false;
	}

  module::new_module(m, "ensure_loaded/1", arg1.ReadString(), (m.getLexer()->ProgInput() ? module::NMT_ENSURE_IN_PROG : module::NMT_ENSURE) );
  return true;
}

LILFES_BUILTIN_DIRECTIVE_1(ensure_loaded, ensure_loaded);

/**
  * @predicate  = module(+NAME)
  * @desc       = Declare module name.This is used at the top of module.
  * @param      = +NAME/ std::string  : Module name
  * @example    = :- module("list:reverse").
  * @end_example

  * @jdesc      = ⥸塼̾ޤνҸϥ⥸塼ƬǻѤޤ
  * @jparam     = +Module/ std::string: ⥸塼̾
  * @jexample   = :- module("list:reverse").
  * @end_jexample
*/
bool module_directive( machine& m, FSP arg1 )
{
			// Ϳ줿⥸塼̾ǥ⥸塼ߡ
	if( lilfes::module::CurrentModule()->IsRegistered() )
	{
		INERR(&m, "Module declaration should be appeared on the top of a module, only once");
	}
	else if( ! arg1.IsString() )
	{
		INERR(&m, "The argument of module/1 should be a std::string");
		lilfes::module::CurrentModule()->Register();
	}
	else if( *arg1.ReadString() == '\0' )
	{
		INERR(&m, "The argument of module/1 should not be an empty std::string");
		lilfes::module::CurrentModule()->Register();
	}
	else
	{
		lilfes::module *mod =  lilfes::module::SearchModule(arg1.ReadString());

		if(mod)		// Ʊ̾⥸塼¸
		{
			if( ! strcmp( arg1.ReadString(), "core" ) || ! strcmp( arg1.ReadString(), "builtin" )  )
			{
                          INERR(&m, "Built-in module \"" << arg1.ReadString() << "\" cannot be redefined");
                          m.getLexer()->Skip();
                          return false;
			}
			if( strcmp( mod->GetDefName(), lilfes::module::CurrentModule()->GetDefName() ) && lilfes_flag::Search("warn-module-override")->GetIntVal() )
			{
				INWARN(&m, "Same module name \"" << arg1.ReadString() << "\" with different paths; the old module will be invalidated");
				INWARN(&m, mod->GetDefName() );
				INWARN(&m, lilfes::module::CurrentModule()->GetDefName() );
			}
			else
				if( lilfes::module::CurrentModule()->GetModuleType() == lilfes::module::NMT_ENSURE 
				 || lilfes::module::CurrentModule()->GetModuleType() == lilfes::module::NMT_ENSURE_IN_PROG
				 || lilfes::module::CurrentModule()->GetModuleType() == lilfes::module::NMT_ENSURE_INITIAL
				 || lilfes::module::CurrentModule()->GetModuleType() == lilfes::module::NMT_ENSURE_WITHOUT_BUILTIN
				 || lilfes::module::CurrentModule()->GetModuleType() == lilfes::module::NMT_ENSURE_WITHOUT_BUILTIN_IN_PROG )
				{
					lilfes::module *prevmod = lilfes::module::CurrentModule()->GetPrevModule();
					if( prevmod )
					{
						if( lilfes::module::CurrentModule()->GetModuleType() == lilfes::module::NMT_ENSURE 
						 || lilfes::module::CurrentModule()->GetModuleType() == lilfes::module::NMT_ENSURE_INITIAL
						 || lilfes::module::CurrentModule()->GetModuleType() == lilfes::module::NMT_ENSURE_IN_PROG)
							{prevmod->Import(mod);}
						lilfes::module::CurrentModule()->SetPrevModule(NULL);
					}
					m.getLexer()->Skip();
					return true;
				}
		}
		lilfes::module::CurrentModule()->Register(arg1.ReadString());
	}
  return true;
}

LILFES_BUILTIN_DIRECTIVE_1(module_directive,module);

/**
  * @predicate  = module_interface
  * @desc       = Declare the begining of the module interface part. See how to make module.
  * @example    = :- module_interface.

  * @end_example

  * @jdesc      = ⥸塼륤󥿥եγϤޤ ⥸塼κ򻲾ȤƤ
  * @jexample   = :- module_interface.
  * @end_jexample
*/
bool module_interface( machine& )
{
  lilfes::module::CurrentModule()->SetInterfaceMode(true);
  return true;
}

LILFES_BUILTIN_DIRECTIVE_0(module_interface, module_interface);

/**
  * @predicate  = module_implementation
  * @desc       = Declare the begining of the module interface part. See how to make module.
  * @example    = :- module_interface.

  * @end_example

  * @jdesc      = ⥸塼γϤޤ ⥸塼κ򻲾ȤƤ
  * @jexample   = :- module_interface.
  * @end_jexample
*/
bool module_implementation( machine& )
{
  lilfes::module::CurrentModule()->SetInterfaceMode(false);
  return true;
}

LILFES_BUILTIN_DIRECTIVE_0(module_implementation, module_implementation);

/**
  * @predicate  = module_extendable
  * @desc       = Declare that this module can be defined as a type of a subtype.
  * @example    = > :- module_extendable.

  * @end_example

  * @jdesc      = Υ⥸塼줿 subtype ¾Υ⥸塼뤳ȤǽǤ뤳Ȥޤ
  * @jexample   = > :- module_extendable.
  * @end_jexample
*/
bool module_extendable( machine& )
{
  lilfes::module::CurrentModule()->SetExtendableMode(true);
  return true;
}

LILFES_BUILTIN_DIRECTIVE_0(module_extendable, module_extendable);

} // namespace builtin

/**
  * @predicate  = get_module(+Type, -Module)
  * @desc       = Get the module name that contains an specified type.
  * @param      = +Type/ bot  : Type of the object
  * @param      = -Module/ string : Module name
  * @example    =
  > ?- get_module(list, X).
  X: "core"
  * @end_example

  * @jdesc      = Type ν° Module ޤ
  * @jparam     = +Type/ bot : оݤη
  * @jparam     = -Module/ string : Type °⥸塼̾
  * @jexample   =
  > ?- get_module(list, X).
  X: "core"
  * @end_jexample
*/
bool get_module( machine& m, FSP arg1, FSP arg2 )
{
  return arg2.Unify(FSP(m,arg1.GetType()->GetModule()->GetName()));
}

LILFES_BUILTIN_PRED_2(get_module, get_module)

/**
  * @predicate  = is_exported(+Type)
  * @desc       = Can discover if a module can accessed from another.
  * @param      = +Type/ bot : Type of the target
  * @example    =
  > ?- is_exported(list).
  yes
  * @end_example

  * @jdesc      = Type ¾Υ⥸塼뤫鸫뤫ɤȽꤷޤ
  * @jparam     = +Type/ bot : оݤη
  * @jexample   =
  > ?- is_exported(list).
  yes
  * @end_jexample
*/
bool is_exported( machine&, FSP arg1 )
{
  return arg1.GetType()->IsExported();
}

LILFES_BUILTIN_PRED_1(is_exported, is_exported)

/**
  * @predicate  = change_module(+NAME)
  * @desc       = Change the present module.
  * @param      = +NAME/ string  : Module name
  * @example    = > :- change_module("lillib:lists").
  * @end_example

  * @jdesc      = ߤΥ⥸塼ѹޤ
  * @jparam     = +NAME/ string: ⥸塼̾
  * @jexample   = > :- change_module("lillib:lists").
  * @end_jexample
*/
bool change_module( machine&, FSP arg1 )
{
	if (! arg1.IsString() ) {
          RUNERR("change_module/1 requires a module name");
          return false;
	}
	module* mod = module::SearchModule( arg1.ReadString() );
	module* prevmod = module::CurrentModule();
	if (mod == NULL) {
          RUNWARN("Warning: module '"<<arg1.ReadString() << "' has not been loaded yet");
          return false;
	}

//	prevmod->Import(mod);
	mod->SetPrevModule(prevmod);
	module::SetCurrentModule( mod );
//	mod->SetInterfaceMode(true);
	mod->SetInterfaceMode(false);
  return true;
}

LILFES_BUILTIN_PRED_1(change_module, change_module)

/**
  * @predicate  = return_module
  * @desc       = Restore the previous module
  * @param      = +NAME/ string  : Module name
  * @example    = > :- return_module.
  * @end_example

  * @jdesc      = Υ⥸塼ޤ
  * @jexample   = > :- return_module.
  * @end_jexample
*/
bool return_module( machine& )
{
	module* mod = module::CurrentModule()->GetPrevModule();
	if (mod == 0) {
          RUNERR("return_module/0 : current module does not have previous module.");
          return false;
	}
	module::SetCurrentModule(mod);
  return true;
}

LILFES_BUILTIN_PRED_0(return_module, return_module)

/**
  * @predicate  = current_module(-NAME)
  * @desc       = Get name of present module.
  * @param      = -NAME/ string  : Module name
  * @example    =
  * @end_example

  * @jdesc      = ߤΥ⥸塼̾ޤ
  * @jparam     = -NAME/ string: ⥸塼̾
  * @jexample   =
  * @end_jexample
*/
bool current_module( machine& m, FSP arg1 )
{
  return arg1.Unify(FSP(m, module::CurrentModule()->GetName()));
}

LILFES_BUILTIN_PRED_1(current_module, current_module)

#define EXT_LENGTH 4

void get_fileinfo(const char *filename, lstring &lilname, lstring &modulename)
//get filename return filename with ".lil" and extract it's modulename
{
	lstring buf = filename;
	bool backslash = false;
	char *p;

	if( ! strcmp(filename, "-") )
	{
		lilname = "-";
		modulename = "stdin";
		return;
	}

	p = strrchr(buf, '/');
	if( p == NULL )
	{
		p = strrchr(buf, '\\');
		if( p != NULL )
			backslash = true;
		else
			p = buf;
	}

	if( ! strrchr(p, '.') )
	{
		lilname = buf + ".lil";
	}
	else
	{
		lilname = buf;
		*p = '\0';
	}

	int i = 0;
	while( buf[i] )
	{
		if( buf[i] == (backslash ? '\\' : '/') )
			buf[i] = ':';
		i++;
	}
	if( buf[0] == ':' )
		modulename = buf+1;
	else
		modulename = buf;
}


void module::Finish() 
{ 
	ASSERT( this == CurrentModule() );
#ifdef DEBUG_MODULE
	std::cout << "Finish " << name << std::endl;
#endif
	if( IsRegistered() ) 
	{
		if( prev_module ) 
		{ 
#ifdef DEBUG_MODULE
			std::cout << "Import " << prev_module->name << std::endl;
#endif
			if( GetModuleType() != NMT_READ_MODULE && prev_module != this )
				if( GetModuleType() != module::NMT_ENSURE_WITHOUT_BUILTIN
				 && GetModuleType() != module::NMT_ENSURE_WITHOUT_BUILTIN_IN_PROG )
					prev_module->Import(this); 
			prev_module = NULL;
		}
	}
	else 
	{
		delete this; 
		m_current = NULL;
	}
}


void module::UseCurrentModule()
{
	if( ! module::CurrentModule()->IsRegistered() )
	{
		RUNERR("The first line must be a module directive");
		module::CurrentModule()->Register();
	}
}


#define LEFT_ALIGN 35
void module::Register(const char *str /*= NULL*/) 
{ 
	if( IsRegistered() ) 
		ABORT("Duplicated module name register"); 

	if( str == NULL ) 
		str = defname; 
	name = module_prefix + str; 
	modulelist_hash.AddOrReplace(this); 
}





} // namespace lilfes

