/**
 * @module  = option
 * @cvs     = $Id: option.cpp,v 1.25 2005/06/16 03:11:09 yusuke Exp $
 * @copyright = Copyright (c) 1997-1998, Makino Takaki
 * @copyright = You may distribute this file under the terms of the Artistic License.
 * @desc    = Option operations
 * Predicates for operating command-line options
 * @jdesc   = ץ
 * @japanese  = ޥɥ饤LiLFeSץǻѤǤ륪ץ˴ؤҸǤ
 * @end_japanese
*/

#include "option.h"
#include "errors.h"
#include "builtin.h"
#include "in.h"
#include <string>
#include <sstream>
#include <stdlib.h>
#ifdef HAVE_LIMITS_H
#  include <limits.h>
#endif

using namespace std;

namespace lilfes {

  static const char rcsid[] = "$Id: option.cpp,v 1.25 2005/06/16 03:11:09 yusuke Exp $";

  /////////////////////////////////////////////////////////////////////////////
  // LiLFeS flags
  
  StringHash<lilfes_flag> lilfes_flag::flagCollection;
  lilfes_flag* lilfes_flag::flagmap[UCHAR_MAX];
  
  /////////////////////////////////////////////////////////////////////////////
  // LiLFeS quiet flag

  bool lilfes_quiet_flag::SetFlagFromFSP(FSP f){
    if( ! f.IsInteger()) {
      RUNERR("LiLFeS Flag \"quiet\" requires 1 or 0");
      return false;
    }

#ifdef PROFILE
    int b = f.ReadInteger();
    if(b == 0){ profiler::SetQuiet(false); return true; }
    else if(b == 1){ profiler::SetQuiet(true); return true; }
    else{ 
            RUNERR("LiLFeS Flag \"quiet\" requires 1 or 0");
	    return false; 
    }
#else // PROFILE
    return true;
#endif // PROFILE
  }

  FSP lilfes_quiet_flag::GetFlagToFSP(machine &m){
#ifdef PROFILE
    int i = (int)profiler::GetQuiet();
#else // PROFILE
    int i = 1;
#endif // PROFILE
    return FSP(m, i);
  }
    
  /////////////////////////////////////////////////////////////////////////////
  // LiLFeS help flag

  int lilfes_help_flag::help_flag = 0;

  int lilfes_help_flag::optcmp(const void *a, const void *b) { return static_cast< const string* >( a )->compare( *static_cast< const string* >( b ) ); }
  
  void lilfes_help_flag::printHelp(string &str){
    extern const char *LastCompiled;
    string last(LastCompiled);
    str =  "Last Compiled on " + last;
    
#if defined( WITH_TNT )
	str += " [TNT]";
#endif
#if defined( WITH_GDBM )
        str += " [GDBM]";
#endif
#if defined( WITH_ASSOC )
        str += " [ASSOC]";
#endif
#if defined( WITH_ITSDB )
        str += " [ITSDB]";
#endif

#if defined( DEBUG )
	str += " [DEBUG]";
#elif defined( DOASSERT )
	str += " [DOASSERT]";
#endif

#if defined( NOMEMCHK )
        str += " [NOMEMCHK]";
#endif
#if defined( XCONSTR )
        str += " [XCONSTR]";
#endif
#if defined( INLINE_WRITEHEAP )
	str += " [IN_WH]";
#endif
#if defined( TYPE_TABLE_HASH )
	str += " [TTHASH]";
#endif
#if defined( PROFILE_HASH )
        str += " [PROFHASH]";
#endif
#if defined( PROF_UNIFY )
        str += " [PROFUNIFY]";
#endif
#if defined( PROF_COPY )
        str += " [PROFCOPY]";
#endif
#if defined( FORCE_DYNAMIC_UNIFY )
	str += " [FORCE_DU]";
#endif
#if defined( FORCE_NORMAL_COPY )
	str += " [FORCE_NC]";
#endif
#if defined( USE_STRING_GC )
        str += " [STRING_GC]";
#endif
#if defined( AMAVL_MODE )
	str += " [AMAVL]";
#endif
#if defined( PROFILE )
	char buf[256];
	sprintf(buf, "%d", PROFILE);
	string str1(buf);
	str += " [PROF=" + str1 + "]";
#endif
	str += "\n";


	str += "Usage:\n" ;

	int n = flagCollection.GetNData();

	string* s = new string[n];
	uint c = 0;

	for( uint i=0; i<flagCollection.GetTableSize(); i++ )
	{
		if( flagCollection.Pick(i) )
		{
			std::ostringstream oss;
			flagCollection.Pick(i)->PrintUsage(oss); 
			s[c++] = oss.str();
		}
	}

	qsort(s, c, sizeof(string), optcmp);

	for( uint i=0; i<c; i++ )
	{
		str += s[i];
	}
	delete[] s;
  }

  bool lilfes_help_flag::SetFlagFromString(const char *)
  {

    help_flag = 1;
    return false;

  }

  FSP lilfes_help_flag::GetFlagToFSP(machine &m){
    string s;
    lilfes_help_flag::printHelp(s);

    return FSP(m, s.c_str());
  }

//////////////////////////////////////////////////////////////////////////////
// LiLFeS load flag

//   static int stoc;
//   static int stoc_max;
//   static lstring* stov;
//   static module::new_module_type *ston;

  std::vector< std::string > lilfes_load_flag::stov;
  std::vector< module::new_module_type > lilfes_load_flag::ston;

  bool lilfes_load_flag::SetFlagFromString(const char *s)
  {
    if( s == NULL || *s == '\0' )
      {
        RUNERR( "-l option requires a file name" );
        return false;
      }

//	std::cout << "hoge0 " << s << std::endl;
    //ston[stoc] = nmt;
    //stov[stoc] = s;
//	std::cout << "hoge1 " << stov[stoc] << std::endl;
    //stoc++;

    stov.push_back( s );
    ston.push_back( nmt );

    return true;
  }

  FSP lilfes_load_flag::GetFlagToFSP(machine &m){

    FSP f( m );
    FSP froot( m, f.GetAddress() );
    for ( size_t i = 0; i < stov.size(); ++i ) {
      FSP g( m, stov[i].c_str() );
      f.Follow( hd ).Unify( g );
      f = f.Follow( tl );
    }
    f.Coerce( nil );

    return froot;
  }

//////////////////////////////////////////////////////////////////////////////
// LiLFeS path flag

  bool lilfes_path_flag::SetFlagFromFSP(FSP f){
 
    if( !f.IsString() ){
      RUNERR("LiLFeS Flag \"include-path\" requires 1 or 0");
      return false;
    }

    lstring path = f.ReadString();
    search_path::AddSearchPath(path);

    return true;
  }

  bool lilfes_path_flag::SetFlagFromString(const char *s)
  {
    if( s == NULL || *s == '\0' )
      {
        RUNERR( "-I option requires a pathname" );
        return false;
      }

    search_path::AddSearchPath(s);
    return true;
  }

  FSP lilfes_path_flag::GetFlagToFSP(machine &m){
  
    FSP f( m );
    FSP froot( m, f.GetAddress() );

    search_path *s = search_path::SearchPath();
    while( s )
      {
        f.Follow(hd).Unify(FSP(m,s->GetName() ));
        f = f.Follow(tl);
        s = s->Next();
      }
  
    f.Coerce( nil );
    return froot;
  }

//////////////////////////////////////////////////////////////////////////////
// LiLFeS int flag

  bool lilfes_int_flag::SetFlagFromFSP(FSP f)
  {
    if( ! f.IsInteger() )
      {
        RUNERR("LiLFeS Flag \"" << name << "\" requires integer");
        return false;
      }
    else if( !is_changeable )
      {
        RUNERR("LiLFeS Flag \"" << name << "\" is read-only");
        return false;
      }
    else
      {
        val = f.ReadInteger();
        return true;
      }
  }

  bool lilfes_int_flag::SetFlagFromString(const char *s)
  {
    if( s == NULL )
      {
        val = 1;
        return false;
      }
    if( *s == '=' )
      s++;
		
    if( *s == '+' )
      val = 1;
    else if( isdigit(*s) || (*s == '-' && isdigit(s[1])))
      val = atoi(s);
    else
      {
        val = 1;
        return false;
      }
    return true;
  }

//////////////////////////////////////////////////////////////////////////////
// LiLFeS arg flag

  std::vector< std::string > lilfes_arg_flag::comargs;

  bool lilfes_arg_flag::SetFlagFromStringArray(int argc, 
                                               char **argv){
    if( argc <= 0 ){
      RUNERR( "-a option requires arguments" );
      return false;
    }
  
    for(int j=0; j<argc; j++){
      comargs.push_back( argv[ j ] );
    }

    return true;
  }

  FSP lilfes_arg_flag::GetFlagToFSP(machine &m){

    FSP f( m );
    FSP froot( m, f.GetAddress() );
    for ( size_t i = 0; i < comargs.size(); ++i ) {
      FSP g( m, comargs[i].c_str() );
      f.Follow( hd ).Unify( g );
      f = f.Follow( tl );
    }
    f.Coerce( nil );
  
    return froot;
  }

//////////////////////////////////////////////////////////////////////////////
// LiLFeS eval flag

  vector< string > lilfes_eval_flag::cmd_list;

  bool lilfes_eval_flag::SetFlagFromString(const char *s){

    if( s == NULL ){
      RUNERR( "-e option requires arguments" );
      return false;
    }

    cmd_list.push_back( s );
    return true;

  }

  FSP lilfes_eval_flag::GetFlagToFSP(machine &m){

    RUNERR("LiLFeS Flag \"eval\" is only available from command-line.");
    return FSP(m);

  }

//////////////////////////////////////////////////////////////////////////////
// LiLFeS package directory flag

string lilfes_package_dir_flag::LILFES_PACKAGE_DIR = PKGDATADIR;

bool lilfes_package_dir_flag::SetFlagFromFSP( FSP f ) {
  if ( ! f.IsString() ) {
    RUNERR( "package-dir option requires string" );
    return false;
  }
  LILFES_PACKAGE_DIR = f.ReadString();
  return true;
}

bool lilfes_package_dir_flag::SetFlagFromString(const char *s){

  if( s == NULL ){
    RUNERR( "package-dir option requires arguments" );
    return false;
  }

  LILFES_PACKAGE_DIR = s;
  return true;

}

FSP lilfes_package_dir_flag::GetFlagToFSP(machine &m){

  return FSP(m, LILFES_PACKAGE_DIR.c_str());

}

//////////////////////////////////////////////////////////////////////

static void init3()
{
#ifdef DEBUG
	std::cout << "Initialize " << __FILE__ << " (init3)" << std::endl;
#endif

	new lilfes_int_flag ('n',"no_initialize"  , 0, false, 
		"Do not interpret global initialize files");
	new lilfes_int_flag ('u',"no_user_initialize"  , 0, false, 
		"Do not interpret user initialize files");
	new lilfes_int_flag ('w',"warning"         , 0, true,
		"Display all warnings");
	new lilfes_quiet_flag ('q',"quiet"           ,
		"Suppress verbose messages");
	new lilfes_int_flag (    "warn-module-override" , 0, true,
		"Warn module override (same module name with different paths)");
	new lilfes_load_flag('l',"load",
		"Load module", module::NMT_ENSURE_INITIAL);
	new lilfes_path_flag('I',"include-path",
		"Add search path of modules");
	new lilfes_arg_flag('a',"argv",
		"Set command-line arguments");
	new lilfes_eval_flag('e',"eval",
		"Execute commands from command-line");
	new lilfes_help_flag('h',"help",
		"Show this help");
	new lilfes_package_dir_flag("package-dir",
		"Directory of LiLFeS data");

#ifdef DEBUG
	std::cout << "End Initialize " << __FILE__ << " (init3)" << std::endl;
#endif
}

static InitFunction IF3(init3, 200);

/**
  * @predicate	= lilfes_flag(+OPT,-ARG) 
  * @desc	= Get an argument of option. 
  * @param	= +OPT/std::string :option mode
  * @param	= -ARG/bot :argument of option
  * @note	= List of option
  * @note	= <div align="conter">
  * @note	=   <table>
  * @note	=     <tr>
  * @note	=       <td>abbreviation</td>
  * @note	=       <td>name</td>
  * @note	=       <td>context</td>
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'n'</td>
  * @note	=       <td>"no_initialize"</td>
  * @note	=       <td>Do not interpret global initialize files</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'u'</td>
  * @note	=       <td>"no_user_initialize"</td>
  * @note	=       <td>Do not interpret user initialize files</td>
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td height="10" width="23">'w'</td>
  * @note	=       <td>"warning"</td>
  * @note	=       <td>Display all warnings</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'q'</td>
  * @note	=       <td>"quiet"</td>
  * @note	=       <td>Suppress verbose messages</td> 
  * @note	=    </tr>
  * @note	=     <tr>
  * @note	=       <td></td>
  * @note	=       <td>"warn-module-override"</td>
  * @note	=       <td>Warn module override (same module name with different paths)</td>
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'l'</td>
  * @note	=       <td>&quot;load&quot;</td>
  * @note	=       <td>Load module</td>
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'I'</td>
  * @note	=       <td>"include-path"</td>
  * @note	=       <td>Add search path of modules</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=     <tr>
  * @note	=       <td>'a'</td>
  * @note	=       <td>"argv"</td>
  * @note	=       <td>Set command-line arguments</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'e'</td>
  * @note	=       <td>"eval"</td>
  * @note	=       <td>Exec a LiLFeS program from command-line</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'h'</td>
  * @note	=       <td>"help"</td>
  * @note	=       <td>Show this help</td> 
  * @note	=     </tr>
  * @note	=   </table>
  * @note	= </div>
  * @example	= 
  * @end_example
  * @jdesc	= ץΰޤ
  * @jparam	= +OPT/std::string :ץ⡼
  * @jparam	= -ARG/bot :ץΰ
  * @jnote	= ץΥꥹ
  * @jnote	= <div align="conter">
  * @jnote	=   <table>
  * @jnote	=     <tr>
  * @jnote	=       <td>ά</td>
  * @jnote	=       <td>ץ̾</td>
  * @jnote	=       <td></td>
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'n'</td>
  * @jnote	=       <td>"no_initialize"</td>
  * @jnote	=       <td>Хʽեᤷޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'u'</td>
  * @jnote	=       <td>"no_user_initialize"</td>
  * @jnote	=       <td>桼νեᤷޤ</td>
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td height="10" width="23">'w'</td>
  * @jnote	=       <td>"warning"</td>
  * @jnote	=       <td>Ƥηٹɽޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'q'</td>
  * @jnote	=       <td>"quiet"</td>
  * @jnote	=       <td>ĹʥåνϤޤ</td> 
  * @jnote	=    </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td></td>
  * @jnote	=       <td>"warn-module-override"</td>
  * @jnote	=       <td>⥸塼overrideʰۤʤѥ̾Ʊ̾Υ⥸塼ɤ߹ळȡˤٹ𤷤ޤ</td>
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'l'</td>
  * @jnote	=       <td>&quot;load&quot;</td>
  * @jnote	=       <td>⥸塼ɤ߹ߤޤ</td>
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'I'</td>
  * @jnote	=       <td>"include-path"</td>
  * @jnote	=       <td>⥸塼θѥɲäޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'a'</td>
  * @jnote	=       <td>"argv"</td>
  * @jnote	=       <td>ޥɥ饤ꤷޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'e'</td>
  * @jnote	=       <td>"eval"</td>
  * @jnote	=       <td>ޥɥ饤󤫤LiLFeSץ¹Ԥޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'h'</td>
  * @jnote	=       <td>"help"</td>
  * @jnote	=       <td>ץΥإפɽޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=   </table>
  * @jnote	= </div>
  * @jexample	= 
  * @end_jexample
*/
/**
  * @predicate	= prolog_flag(+OPT,-ARG) 
  * @desc	= Get an argument of option. 
  * @param	= +OPT/std::string :option mode
  * @param	= -ARG/bot :argument of option
  * @note	= List of option
  * @note	= <div align="conter">
  * @note	=   <table>
  * @note	=     <tr>
  * @note	=       <td>abbreviation</td>
  * @note	=       <td>name</td>
  * @note	=       <td>context</td>
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'n'</td>
  * @note	=       <td>"no_initialize"</td>
  * @note	=       <td>Do not interpret global initialize files</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'u'</td>
  * @note	=       <td>"no_user_initialize"</td>
  * @note	=       <td>Do not interpret user initialize files</td>
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td height="10" width="23">'w'</td>
  * @note	=       <td>"warning"</td>
  * @note	=       <td>Display all warnings</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'q'</td>
  * @note	=       <td>"quiet"</td>
  * @note	=       <td>Suppress verbose messages</td> 
  * @note	=    </tr>
  * @note	=     <tr>
  * @note	=       <td></td>
  * @note	=       <td>"warn-module-override"</td>
  * @note	=       <td>Warn module override (same module name with different paths)</td>
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'l'</td>
  * @note	=       <td>&quot;load&quot;</td>
  * @note	=       <td>Load module</td>
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'I'</td>
  * @note	=       <td>"include-path"</td>
  * @note	=       <td>Add search path of modules</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=     <tr>
  * @note	=       <td>'a'</td>
  * @note	=       <td>"argv"</td>
  * @note	=       <td>Set command-line arguments</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'e'</td>
  * @note	=       <td>"eval"</td>
  * @note	=       <td>Exec a LiLFeS program from command-line</td> 
  * @note	=     </tr>
  * @note	=     <tr>
  * @note	=       <td>'h'</td>
  * @note	=       <td>"help"</td>
  * @note	=       <td>Show this help</td> 
  * @note	=     </tr>
  * @note	=   </table>
  * @note	= </div>
  * @example	= 
  * @end_example
  * @jdesc	= ץΰޤ
  * @jparam	= +OPT/std::string :ץ⡼
  * @jparam	= -ARG/bot :ץΰ
  * @jnote	= ץΥꥹ
  * @jnote	= <div align="conter">
  * @jnote	=   <table>
  * @jnote	=     <tr>
  * @jnote	=       <td>ά</td>
  * @jnote	=       <td>ץ̾</td>
  * @jnote	=       <td></td>
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'n'</td>
  * @jnote	=       <td>"no_initialize"</td>
  * @jnote	=       <td>Хʽեᤷޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'u'</td>
  * @jnote	=       <td>"no_user_initialize"</td>
  * @jnote	=       <td>桼νեᤷޤ</td>
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td height="10" width="23">'w'</td>
  * @jnote	=       <td>"warning"</td>
  * @jnote	=       <td>Ƥηٹɽޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'q'</td>
  * @jnote	=       <td>"quiet"</td>
  * @jnote	=       <td>ĹʥåνϤޤ</td> 
  * @jnote	=    </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td></td>
  * @jnote	=       <td>"warn-module-override"</td>
  * @jnote	=       <td>⥸塼overrideʰۤʤѥ̾Ʊ̾Υ⥸塼ɤ߹ळȡˤٹ𤷤ޤ</td>
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'l'</td>
  * @jnote	=       <td>&quot;load&quot;</td>
  * @jnote	=       <td>⥸塼ɤ߹ߤޤ</td>
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'I'</td>
  * @jnote	=       <td>"include-path"</td>
  * @jnote	=       <td>⥸塼θѥɲäޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'a'</td>
  * @jnote	=       <td>"argv"</td>
  * @jnote	=       <td>ޥɥ饤ꤷޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'e'</td>
  * @jnote	=       <td>"eval"</td>
  * @jnote	=       <td>ޥɥ饤󤫤LiLFeSץ¹Ԥޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=     <tr>
  * @jnote	=       <td>'h'</td>
  * @jnote	=       <td>"help"</td>
  * @jnote	=       <td>ץΥإפɽޤ</td> 
  * @jnote	=     </tr>
  * @jnote	=   </table>
  * @jnote	= </div>
  * @jexample	= 
  * @end_jexample
*/
namespace builtin {

bool lilfes_flag_2( machine& m, FSP arg1, FSP arg2 )
{
  if( ! arg1.IsString() )
    {
          RUNERR("Error: lilfes_flag/2 requires a std::string as a first argument");
          return false;
    }
  else
    {
      lilfes::lilfes_flag *lf = lilfes::lilfes_flag::Search(arg1.ReadString());
      if( lf == NULL )
	{
	  RUNERR("Error: Unknown lilfes flag \"" << arg1.ReadString() << "\"");
	  return false;
	}
      else
	{
	  return arg2.Unify(lf->GetFlagToFSP(m));
	}
    }
}
  
  LILFES_BUILTIN_PRED_2(lilfes_flag_2, lilfes_flag);
  LILFES_BUILTIN_PRED_2(lilfes_flag_2, prolog_flag);
}


//////////////////////////////////////////////////////////////////////////////
// lilfes_flag/3 : 

/**
  * @predicate	= lilfes_flag(+OPT, -OLD, -NEW) 
  * @desc	= Set option.
  * @param	= +OPT/string : option mode
  * @param	= -OLD/bot : an old value of the option
  * @param	= -NEW/bot : a new value of the option
  * @example	= 
  * @end_example
  * @jdesc      = ץΰ򿷤ꤷޤ
  * @jparam	= +OPT/string : ץΥ⡼
  * @jparam	= -OLD/bot : θŤ
  * @jparam	= -NEW/bot : ο
  * @jexample	= 
  * @end_jexample
*/
/**
  * @predicate	= prolog_flag(+OPT, -OLD, -NEW) 
  * @desc	= Set option.
  * @param	= +OPT/string : option mode
  * @param	= -OLD/bot : an old value of the option
  * @param	= -NEW/bot : a new value of the option
  * @example	= 
  * @end_example
  * @jdesc      = ץΰ򿷤ꤷޤ
  * @jparam	= +OPT/string : ץΥ⡼
  * @jparam	= -OLD/bot : θŤ
  * @jparam	= -NEW/bot : ο
  * @jexample	= 
  * @end_jexample
*/
namespace builtin {

bool lilfes_flag_3( machine& m, FSP arg1, FSP arg2, FSP arg3 )
{
  if( !lilfes_flag_2(m, arg1, arg2)) 
    return false;

  if( ! arg1.IsString() )
    {
      RUNERR("Error: lilfes_flag/3 requires a std::string as a first argument");
      return false;
    }
  else
    {
      lilfes_flag *lf = lilfes_flag::Search(arg1.ReadString());
      if( lf == NULL )
	{
	  RUNERR("Error: Unknown lilfes flag \"" << arg1.ReadString() << "\"");
	  return false;
	}
      else
	{
	  return lf->SetFlagFromFSP(arg3);
	}
    }
}

LILFES_BUILTIN_PRED_OVERLOAD_3(lilfes_flag_3, lilfes_flag_3, lilfes_flag);
LILFES_BUILTIN_PRED_OVERLOAD_3(lilfes_flag_3, prolog_flag_3, prolog_flag);

} // namespace builtin

///////////////////////////////////////////////////////////////////////////////

/**
  * @predicate	= get_argc(-ARGC) 
  * @desc	= Get the number of command-line arguments.
  * @param	= -ARGC/integer : the number of arguments
  * @example	= 
  * % lilfes -a arg1 arg2
  * > ?- get_argc(X).
  * X: 2
  * @end_example
  * @jdesc	= ޥɥ饤οޤ
  * @jparam	= 
  * @jexample	= 
  * @end_jexample
*/
namespace builtin{
bool get_argc( machine& m, FSP arg1 )
{
  FSP result = FSP( m, lilfes_arg_flag::getArgc() );
  
  if( ! arg1.Unify( result )){
    return false;
  }
  return true;
}

LILFES_BUILTIN_PRED_1(get_argc, get_argc);
}

///////////////////////////////////////////////////////////////////////////////

/**
  * @predicate	= get_argv(+IND, -ARG) 
  * @desc	= Get a commandline argument.
  * @param	= +IND/integer : index
  * @param	= -ARG/string : commandline argument
  * @example	= 
  * % lilfes -a arg1 arg2
  * > ?- get_argv(1, X).
  * X: "arg2"
  * @end_example
  * @jdesc	= ޥɥ饤ޤ
  * @jparam	= 
  * @jexample	= 
  * @end_jexample
*/
namespace builtin{
  bool get_argv( machine& m, FSP arg1, FSP arg2 )
  {
    if ( ! arg1.IsInteger() ) {
      RUNWARN( "get_argv/2: arg1 requires integer" );
      return false;
    }

    int i = arg1.ReadInteger();

    if( i >= lilfes_arg_flag::getArgc() ){
      RUNWARN( "get_argv/2: arg1 is too big" );
      return false;
    }

    FSP result = FSP( m, lilfes_arg_flag::getArgv( i ).c_str() );

    if( ! arg2.Unify( result )){
      return false;
    }
    return true;
  }

  LILFES_BUILTIN_PRED_2(get_argv, get_argv);
}

void lilfes_flag::PrintUsage(ostream &out)
{
	out << "   ";
	if( opt )
		out << "-" << opt;
	else
		out << "  ";

	if( name )
		out << " --" << name;
	out << "  " << usage << std::endl;
}

void lilfes_flag::ParseOptions(machine *mach, int argc, char **argv)
{
#ifdef DEBUG
	std::cout << "ParseOptions" << argc << std::endl;
#endif

	int warn = 0;

	//stoc = 0;

#ifdef DEBUG
	std::cout << "ParseOptions..." << argc << std::endl;
#endif

	//stov = new lstring [argc];
	//ston = new module::new_module_type[argc];
	

#ifdef DEBUG
	std::cout << "ParseOptions   " << argc << std::endl;
#endif

	//comargc = 1;
	//comargv = argv;
	// argc, argv

	for( int i=1; i<argc; i++ )
	{
#ifdef DEBUG
	std::cout << "ParseOptions " << i << "/" << argc << "  " << argv[i] << std::endl;
#endif
		if( argv[i][0] == '-' && argv[i][1] )
		{
		  if( argv[i][1] == '-' )  /// ץ󤬡"--ʤȤ"ξ
		    {
		      char *p = strchr(argv[i]+2, '='); //Ǥ¦˽и'='ؤΥݥ󥿤
		      if( p ) //'='С
			{
			  *p = '\0';
			  lilfes_flag *f = lilfes_flag::Search(argv[i]+2);
			  if( f == NULL )
			    RUNERR( "Unknown option: --" << argv[i]+2 );
			  else
			    f->SetFlagFromString(p+1);
			}
		      else
			{
			  lilfes_flag *f = lilfes_flag::Search(argv[i]+2);

			  if( f == NULL )
			    RUNERR( "Unknown option: --" << argv[i]+2 );
			  else if( strcmp( (argv[i]+2), "argv" ) == 0 ){
			    //lilfes_arg_flag::SetFlagFromStringArray(argc, argv, i);
			    //i += comargc - 1;
			    lilfes_arg_flag::SetFlagFromStringArray(argc-i-1, argv+i+1);
                            break;
			  }
			  else {
			    bool ret = f->SetFlagFromString(argv[i+1]);
			    if( ret )
			      i++;
			  }
			}
		    }
		  else  /// ץ󤬡"-ʤȤ"ξ
		    {
		      lilfes_flag *f = lilfes_flag::GetFlagMap(argv[i][1]);
		      if( f == NULL )
			RUNERR( "Unknown option: -" << argv[i][1] );
		      else if( argv[i][1] == 'a' ){
			lilfes_arg_flag::SetFlagFromStringArray(argc-i-1, argv+i+1);
                        break;
		      }
		      else if( argv[i][2] )
			f->SetFlagFromString(argv[i]+2);
		      else
			{
			  bool ret = f->SetFlagFromString(argv[i+1]);
			  if( ret ) // Υץʸ񤷤ϡĤνȤФƤμء
						i++;
			}
		    }
		}
		else
		    {
		      if( !warn )
			{
			  RUNWARN( "obsolete commandline" );
			  warn = 1;
			}
// 		      ston[stoc] = module::NMT_INCLUDE;
// 		      stov[stoc] = (lstring)(argv[i]);
// 		      stoc++;
                      lilfes_load_flag::pushNewModule( argv[ i ], module::NMT_INCLUDE );
		    }
	}

#ifdef DEBUG
	std::cout << "ParseOptions q" << std::endl;
#endif
// 	stoc_max = stoc;
// 	while(stoc--)
// 	{
// 		module::new_module(*mach, "command line", stov[stoc], ston[stoc], module::UserModule());
// 	}
        for ( int i = lilfes_load_flag::getNumLoadModule() - 1; i >= 0; --i ) {
          module::new_module(*mach, "command line",
                             lilfes_load_flag::getLoadModule( i ).c_str(),
                             lilfes_load_flag::getLoadModuleType( i ));
        }

#ifdef DEBUG
	std::cout << "ParseOptions r" << std::endl;
#endif
	if( ! lilfes_flag::Search("no_user_initialize")->GetIntVal() )
	{
		char *init = getenv("LILFESRC");
		if( init == NULL )
			init = LOCAL_INITIALIZATION_FILE;
		module::new_module(*mach, "initialization phase", init, module::NMT_INITIALIZE, module::UserModule());
	}

#ifdef DEBUG
	std::cout << "ParseOptions s" << std::endl;
#endif
	if( ! lilfes_flag::Search("no_initialize")->GetIntVal() )
	{
		module::new_module(*mach, "initialization phase", GLOBAL_INITIALIZATION_FILE, module::NMT_INITIALIZE, module::UserModule());
	}

#ifdef DEBUG
	std::cout << "ParseOptions t" << std::endl;
#endif
	//delete[] stov;
	//delete[] ston;

#ifdef DEBUG
	std::cout << "ParseOptions x" << std::endl;
#endif
}

int lilfes_flag::ExecOptions(machine &m){
  
  if( lilfes_eval_flag::getCommandList().empty() && lilfes_help_flag::getHelpFlag() == 0 ) { return 0; }
  else {
    if( lilfes_help_flag::getHelpFlag() ) { 
      string s;
      lilfes_help_flag::printHelp(s);
      *error_stream << s;
    }
    if( ! lilfes_eval_flag::getCommandList().empty() ) {
      for ( vector< string >::const_iterator it = lilfes_eval_flag::getCommandList().begin();
            it != lilfes_eval_flag::getCommandList().end();
            ++it ) {
        std::istringstream iss(*it);
        lilfes::lexer_t lexer(&iss, "eval_flag");
        m.setLexer(lexer);
        m.parse();
      }
    }
    return 1;
  }
  return 0;

}

} // namespace lilfes



