/**
 * @module  = prof
 * @cvs     = $Id: profile.cpp,v 1.13 2005/06/30 18:47:46 yusuke Exp $
 * @copyright = Copyright (c) 1997, Makino Takaki
 * @copyright = You may distribute this file under the terms of the Artistic License.
 * @desc    = Profiling
 * @jdesc   = ץե
 * Builtin predicates for measuring execution times
 * @japanese =
 * ¹Ի֤¬ҸǤ
 * @end_japanese
*/

/* $Id: profile.cpp,v 1.13 2005/06/30 18:47:46 yusuke Exp $ 
 *
 *    
 *
 *    You may distribute this file under the terms of the Artistic License.
 *
 */

#include "profile.h"
#include "hash.h"
#include "initial.h"
#include "builtin.h"
#include <iostream>
#include <iomanip>
#include <cctype>
#include <cstdio>

static const char rcsid[] = "$Id: profile.cpp,v 1.13 2005/06/30 18:47:46 yusuke Exp $";

#define exception _exception
#include <math.h>
#undef exception

#ifdef HAVE_SYS_TIME_H
#include "sys/time.h"
#endif
#ifdef HAVE_SYS_RESOURCE_H
#include "sys/resource.h"
#endif
#ifdef HAVE_WINDOWS_H
#include <windows.h>
#endif

using namespace std;

//#define DEBUG_PROF

namespace lilfes {

#ifdef PROFILE

#ifndef PASSIVE_PROFILE
static timebase pertick = 0;
#endif

#ifdef PASSIVE_PROFILE

#define TICK 10000		/* ticks; microsec */

const timebase profiler::tickunit = TICK / 1000000.0;
struct sigaction profiler::oldact_SIGALRM;
int profiler::profile_block_count = 0;

#define USE_SIGACTION
#define USE_SIGALRM_HANDLER

#endif

// Global variables.

static HashTable<profbase> profhash;

prof *prof::current = NULL;
profuser * profuser::root = NULL;
bool profiler::quiet = false;

ostream& profbase::output(ostream& os)
{
	profbase &p = *this;
//#ifdef PASSIVE_PROFILE
//	timebase alltime = p.alltime + p.exctime;
//	timebase runtime = p.alltime;
//#else
	timebase alltime = p.alltime;
	timebase runtime = p.alltime - p.exctime;
	if( alltime < 0 )
		alltime = 0;
	if( runtime < 0 )
		runtime = 0;
//#endif
	timebase pertime = runtime / p.passcnt;
//	os << "(" << alltime << "-" << p.exctime << "=" << runtime << ")" << std::endl;
	return os << std::setw(9) << p.passcnt
	          << std::setw(9) << (long)floor(alltime) << "." << std::setw(4) << setfill('0') << floor((alltime-floor(alltime))*10000) << setfill(' ')
	          << std::setw(9) << (long)floor(runtime) << "." << std::setw(4) << setfill('0') << floor((runtime-floor(runtime))*10000) << setfill(' ')
	          << std::setw(8) << (long)floor(pertime) << "." << std::setw(9) << setfill('0') << (long)(floor((pertime-floor(pertime))*1000000000)) << setfill(' ')
	          << "  " << p.name << std::endl;
}

prof::prof(const char *n)
{
#ifdef DEBUG_PROF
	std::cout << "PROF START: " << n << std::endl;
#endif
	parent = current;
	current = NULL;
	my_base = profhash.Find(n);
	if( my_base == NULL )
	{
		my_base = new profbase(n);
		profhash.Add(my_base);
	}
	else
		my_base->Enter();
#ifdef PASSIVE_PROFILE
	start_tick = 0;
#else
	start_tick = GetTick();
#endif
	current = this;
}

prof::~prof()
{
#ifdef DEBUG_PROF
	std::cout << "PROF END: " << my_base->GetKey() << std::endl;
#endif
#ifdef PASSIVE_PROFILE
	current = NULL;
	if( my_base->Leave() == 0 )
		my_base->AddTime(start_tick);
	if( parent != NULL ) {
		parent->AddTime(start_tick);
		parent->my_base->ExcTime(start_tick);
#ifdef DEBUG_PROF
		if( start_tick > 0.0 )
		std::cout << "PROF: " << parent->my_base->GetKey() << " -> " << my_base->GetKey() << ": " << start_tick << std::endl;
#endif
	}
#else
	timebase end_tick = GetTick();
	timebase tick = end_tick - start_tick;
	my_base->AddTime(tick - pertick);
	if( parent != NULL ) {
		parent->my_base->ExcTime(tick);
#ifdef DEBUG_PROF
		std::cout << "PROF: " << parent->my_base->GetKey() << " -> " << my_base->GetKey() << ": " << tick << std::endl;
#endif
	}
#endif
	current = parent;
}

timebase prof::GetTick()
{
#ifdef MSDOS
#ifdef WIN95
	return GetTickCount() * 0.001;
#else
	FILETIME c, e, k, u;
	GetProcessTimes(GetCurrentProcess(), &c, &e, &k, &u);
	
	return u.dwLowDateTime * 0.0000001 + u.dwHighDateTime * (0.0000001 * 4294967296.0);
#endif
#else		// UNIX
	struct rusage ru;
	getrusage(RUSAGE_SELF, &ru);
	
	return ru.ru_utime.tv_sec + ru.ru_utime.tv_usec * 0.000001;
#endif
}

profiler::profiler()
{
	p = new prof("Total");

#ifdef PASSIVE_PROFILE
//#if defined(USE_SIGACTION)
	struct sigaction act;
	
//# if defined(USE_SIGALRM_HANDLER)
	act.sa_handler = &SIGALRM_Handler;
#ifdef SA_RESTART
	act.sa_flags = SA_RESTART;			// Not to stop FLEX action
#else
	act.sa_flags = 0;
#endif
	sigemptyset(&act.sa_mask);
	sigaction(SIGPROF, &act, &oldact_SIGALRM);
//# endif // USE_SIGALRM_HANDLER
//#endif // USE_SIGACTION

#ifdef USE_TIMER_SETTIME
	int ret = timer_create(CLOCK_REALTIME, NULL, &timerid);
	if( ret != 0 )
	{
		RUNERR( "Fail to create timer" );
                perror(""); return;
	}

	struct timespec tspec;

	clock_getres(CLOCK_REALTIME, &tspec);

	struct itimerspec itspec;

	itspec.it_interval.tv_sec  = 0;
	itspec.it_interval.tv_nsec = (TICK % 1000000) * 1000;
	itspec.it_value   .tv_sec  = 0;
	itspec.it_value   .tv_nsec = (TICK % 1000000) * 1000;

	*output_stream << "Profile clock resolution:" << itspec.it_interval.tv_sec + itspec.it_interval.tv_nsec * 0.000000001 << std::endl;

	ret = timer_settime(timerid, 0, &itspec, NULL);

	if( ret != 0 )
	{
		RUNERR( "Fail to set timer" ); return;
	}

	timer_gettime(timerid, &itspec);
	*output_stream << "System clock resolution:" << itspec.it_interval.tv_sec + itspec.it_interval.tv_nsec * 0.000000001 << std::endl;
#else
	struct itimerval it;

	it.it_value.tv_sec = 0;
	it.it_value.tv_usec = TICK % 1000000;
	it.it_interval.tv_sec = 0;
	it.it_interval.tv_usec = TICK % 1000000;
	int ret = setitimer(ITIMER_PROF, &it, NULL);
	if( ret != 0 )
	{
		RUNERR( "Fail to create timer" );
                perror(""); return;
	}
#endif
#endif
}

profiler::~profiler()
{
	if( ! quiet )
	{
		if( ! IsProfiling() )
			*error_stream << "********** Profile consistency error!! ***********" << std::endl;

		if( profuser::GetRoot() )
		{
			*error_stream << "********** Following Profiles are not completed!! ***********" << endl;
			profuser *p = profuser::GetRoot();
			while( p )
			{
				*error_stream << "User profile: " << p->GetName() << endl;
				profuser *q = p;
				p = p->GetNext();
				delete q;
			}
		}

	#ifdef USE_TIMER_SETTIME
		timer_delete(timerid);
	#endif
		uint i;
		
		delete p;
	#ifdef WIN95
		*error_stream << "Profiling result (Win95 ticks; not accurate)" << std::endl;
	#else
		*error_stream << "Profiling result" << std::endl;
	#endif
		*error_stream << "PassCount    Total Time  Exclude Time   Time Per Pass    Name" << std::endl;

		profbase **arr = new profbase *[profhash.GetNData()];
		profbase **p = arr;
		for( i=0; i<profhash.GetTableSize(); i++ )
		{
			profbase *b = profhash.Pick(i);
			if( b != NULL )
			{
				*p++ = b;
			}
		}
		qsort(arr, profhash.GetNData(), sizeof(profbase *), profbase::pbcomp);
		for( i=0; i<profhash.GetNData(); i++ )
		{
			arr[i]->output(*error_stream);
			delete arr[i];
		}
		delete[] arr;
	}
}


static profiler *profr;
/*
timebase tm = 0;
long timecnt = 0;
long overrun = 0;
*/
#ifdef PASSIVE_PROFILE
void profiler::SIGALRM_Handler(int)
{
//	long ov = timer_getoverrun(profr->timerid);
	if( IsProfiling() )
	{
		prof *p = prof::current;
		if( p != NULL )
		{
			p->AddTime(tickunit);// * (1 + ov));
		}

		int c = 0;
		profuser *pu = profuser::root;
		while( pu != NULL )
		{
			if( ++c > 100000 )
			{
				ABORT("INFINITE LOOP in profiler::SIGALRM_Handler");
			}
			pu->AddTime(tickunit);// * (1 + ov));
			pu = pu->next;
		}
	}
/*
	else
	{
		tm += tickunit;// * (1 + ov);
	}
*/
//	overrun += ov;
}
#endif

int COMPAREFUNC profbase::pbcomp(const void *a, const void *b)
{ 
	return strcmp((*(profbase **)a)->name, (*(profbase **)b)->name);
}


prof *pit;

static void init() 
{ 
#ifdef DEBUG
	std::cout << "Initialize " << __FILE__ << std::endl;
#endif
	profr = new profiler; 
#if PROFILE >= 2
	pit = new prof("Initialize");
#endif
}
static void term() 
{ 
#if PROFILE >= 2
	delete pit;
#endif
	delete profr; 
}
static InitFunction IF(init, 1);
static TermFunction TF(term, 1);

#if PROFILE >= 2
static void init2() 
{ 
#ifdef DEBUG
	std::cout << "Initialize " << __FILE__ << " (init2)" << std::endl;
#endif
	delete pit;
}
static void term2() 
{ 
	pit = new prof("Terminate");
}
static InitFunction IF2(init2, 65535);
static TermFunction TF2(term2, 65535);
#endif

#ifndef PASSIVE_PROFILE
static void init3() 
{
#ifdef DEBUG
	std::cout << "Initialize " << __FILE__ << " (init3)" << std::endl;
#endif
	int i;
	timebase start = prof::GetTick();
	for( i=1; i<1000; i++ )
	{
		prof::GetTick();
	}
	pertick = (prof::GetTick() - start) * (1.0/i);
#ifdef DEBUG
	*error_stream << "Profiler PerTick = " << pertick << std::endl;
#endif
}

static InitFunction IF3(init3, 2);
#endif

//////////////////////////////////////////////////////////////////////

profuser::profuser(const char *n)
{
//	cout << "PROF: " << n << endl;
	my_base = profhash.Find(n);
	if( my_base == NULL )
	{
		my_base = new profbase(n);
		profhash.Add(my_base);
	}
	else
		my_base->Enter();
#ifdef PASSIVE_PROFILE
	start_tick = 0;
#else
	start_tick = prof::GetTick();
#endif
	next = root;
	root = this;
}

profuser::~profuser()
{
//	cout << "~PROF: " << my_base->GetName() << endl;
	profiler::DisableProfiling();
	int c=0;
	profuser **p = &root;
	while( *p != this )
	{
		if( ++c > 10000 )
		{
			cerr << "PROF ERROR: INFINITE LOOP" << endl;
			profuser *x = root;
			for( int i=0; i<30; i++ )
			{
				cerr << x->my_base->GetName() << "(" << hex << (ptr_int)x << dec << ") -> ";
				x = x->next;
			}
			cerr << "...." << endl;
			ABORT("INFINITE LOOP profuser list");
		}
		if( *p == NULL )
		{
			ABORT("CANNOT REMOVE LIST FROM profuser list");
		}
		p = &((*p)->next);
	}
	*p = next;
	
#ifdef PASSIVE_PROFILE
	if( my_base->Leave() == 0 )
		my_base->AddTime(start_tick);
#else
	timebase end_tick = GetTick();
	timebase tick = end_tick - start_tick;
	my_base->AddTime(tick - pertick);
#endif
	profiler::EnableProfiling();
}

profuser *profuser::search_internal(const char *name)
{ 
	if( !strcmp(my_base->GetName(), name) )
	{
		return this;
	}
	else
	{
		if( next == NULL )
			return NULL;
		else
			return next->search_internal(name);
	}
}

void start_prof(const char* s) {
    profiler::DisableProfiling();
    new profuser(s);
    profiler::EnableProfiling();
}

int stop_prof(const char* s) {
    profiler::DisableProfiling();

    profuser *p = profuser::Search(s);
    if( p == NULL ) {
        RUNERR("Profile \"" << s << "\" is not started");
        profiler::EnableProfiling();
        return 0;
    }
#ifdef PASSIVE_PROFILE
    timebase f = p->GetTime();
#else
    timebase f = prof::GetTick() - p->GetTime();
#endif
    delete p;
    int time = (int) (f*1000);
    profiler::EnableProfiling();
    return time;
}

int total_prof(const char* s) {
    profiler::DisableProfiling();

    profbase *b = profhash.Find(s);
    if( b == NULL )	{
        RUNERR("Profile \"" << s << "\" is not defined");
        profiler::EnableProfiling();
        return 0;
    }
    timebase f = b->GetAllTime();
    int time = (int)(f*1000);
    profiler::EnableProfiling();
    return time;
}

unsigned long total_count_prof(const char* s) {
    profiler::DisableProfiling();

    profbase *b = profhash.Find(s);
    if( b == NULL )	{
        RUNERR("Profile \"" << s << "\" is not defined");
        profiler::EnableProfiling();
        return 0;
    }
    unsigned long c = b->GetPassCount();
    profiler::EnableProfiling();
    return c;
}
    
//////////////////////////////////////////////////////////////////////

namespace builtin {

/**
  * @predicate	= start_prof(+NAME) 
  * @desc	= make new profiling part and start profiling.see profile.
  * @param	= +NAME/string :profile part name
  * @example	= 
  * @end_example

  * @jdesc	= ˥ץե֤ץե󥰤Ϥޤ
  * @jparam	= +NAME/string :ץե֤̾
  * @jexample	= 
  * @end_jexample
*/
bool start_prof( machine&, FSP arg1 )
{
	profiler::DisableProfiling();
	if( ! arg1.IsString() )
	{
		RUNERR("start_prof/1 requires a string");
		profiler::EnableProfiling();
                return false;
	}
	else
	{
		new profuser(arg1.ReadString());
		profiler::EnableProfiling();
                return true;
	}
}

/**
  * @predicate	= stop_prof(+NAME) 
  * @desc	= Stop profiling.
  * @param	= +NAME/string :profile part name
  * @example	= 
  * @end_example

  * @jdesc	= ץե󥰤ߤޤ
  * @jparam	= +NAME/string :ץե֤̾
  * @jexample	= 
  * @end_jexample
*/
bool stop_prof( machine&, FSP arg1 )
{
	profiler::DisableProfiling();
	if( ! arg1.IsString() )
	{
		RUNERR("stop_prof/1 requires a string");
		profiler::EnableProfiling();
		return false;
	}
	else
	{
		profuser *p = profuser::Search(arg1.ReadString());
		if( p == NULL )
		{
			RUNERR("Profile \"" << arg1.ReadString() << "\" is not started");
			profiler::EnableProfiling();
			return true;
		}
		else
		{
			delete p;
			profiler::EnableProfiling();
			return true;
		}
	}
}

/**
  * @predicate	= stop_prof(+NAME,-TIME) 
  * @desc	= Get result of profiling.
  * @param	= +NAME/string :profile part name
  * @param	= -TIME/integer :result time
  * @example	= 
  * @end_example

  * @jdesc	= ץե󥰤ߤη̤ޤ
  * @jparam	= +NAME/string :ץե֤̾
  * @jparam	= -TIME/integer :η̤λ
  * @jexample	= 
  * @end_jexample
*/
bool stop_prof_2( machine& m, FSP arg1, FSP arg2 )
{
	profiler::DisableProfiling();

	if( ! arg1.IsString() )
	{
		RUNERR("stop_prof/2 requires a string");
		profiler::EnableProfiling();
		return false;
	}
	else
	{
		profuser *p = profuser::Search(arg1.ReadString());
		if( p == NULL )
		{
			RUNERR("Profile \"" << arg1.ReadString() << "\" is not started");
			profiler::EnableProfiling();
			return true;
		}
		else
		{
#ifdef PASSIVE_PROFILE
			timebase f = p->GetTime();
#else
			timebase f = prof::GetTick() - p->GetTime();
#endif
			delete p;
			if( ! arg2.Unify(FSP(m,(mint)(f*1000))) )
			{
				profiler::EnableProfiling();
				return false;
			}
			else
			{
				profiler::EnableProfiling();
				return true;
			}
		}
	}
}

/**
  * @predicate	= total_prof(+NAME,-TIME) 
  * @desc	= Get total time of profiling.
  * @param	= +NAME/string :profile part name
  * @param	= -TIME/integer :result time
  * @example	= 
  * @end_example

  * @jdesc	= ץե󥰤֤ޤ 
  * @jparam	= +NAME/string :ץե֤̾
  * @jparam	= -TIME/integer :η̤λ
  * @jexample	= 
  * @end_jexample
*/
bool total_prof( machine& m, FSP arg1, FSP arg2 )
{
	profiler::DisableProfiling();

	if( ! arg1.IsString() )
	{
		RUNERR("total_prof/2 requires a string");
		profiler::EnableProfiling();
                return false;
	}
	else
	{
		profbase *b = profhash.Find(arg1.ReadString());
		if( b == NULL )
		{
			profiler::EnableProfiling();
                        return false;
		}
		else
		{
			timebase f = b->GetAllTime();
			if( ! arg2.Unify(FSP(m,(mint)(f*1000))) )
			{
				profiler::EnableProfiling();
                                return false;
			}
			else
			{
				profiler::EnableProfiling();
                                return true;
			}
		}
	}
}

/**
  * @predicate	= total_prof(+NAME,-TIME,-COUNT) 
  * @desc	= Get total time and count of profiling.
  * @param	= +NAME/string : profile part name
  * @param	= -TIME/integer : total time
  * @param	= -COUNT/integer : total count
  * @example	= 
  * @end_example

  * @jdesc	= ץե󥰤֤̲ޤ 
  * @jparam	= +NAME/string : ץե֤̾
  * @jparam	= -TIME/integer : η̤λ
  * @jparam	= -COUNT/integer : ̲
  * @jexample	= 
  * @end_jexample
*/
bool total_prof_3( machine&, FSP arg1, FSP arg2, FSP arg3 )
{
	profiler::DisableProfiling();

	if( ! arg1.IsString() )
	{
		RUNERR("total_prof/3 requires a string");
		profiler::EnableProfiling();
                return false;
	}
	else
	{
		profbase *b = profhash.Find(arg1.ReadString());
		if( b == NULL )
		{
			profiler::EnableProfiling();
                        return false;
		}
		else
		{
			timebase f = b->GetAllTime();
                        unsigned long c = b->GetPassCount();
			if( arg2.Unify((mint)(f*1000)) &&
                            arg3.Unify((mint)c) )
                          {
				profiler::EnableProfiling();
                                return true;
			}
			else
			{
				profiler::EnableProfiling();
                                return false;
			}
		}
	}
}

} // namespace builtin

#else // PROFILE

void start_prof(const char*) {
}
int stop_prof(const char*) {
  return 0;
}
int total_prof(const char*) {
  return 0;
}
unsigned long total_count_prof(const char*) {
  return 0;
}

namespace builtin {

bool start_prof( machine&, FSP ) {
  return true;
}

bool stop_prof( machine&, FSP ) {
  return true;
}

bool stop_prof_2( machine&, FSP, FSP arg2 ) {
  return arg2.Unify( 0 );
}

bool total_prof( machine&, FSP, FSP arg2 ) {
  return arg2.Unify( 0 );
}

bool total_prof_3( machine&, FSP, FSP arg2, FSP arg3 ) {
  return arg2.Unify( 0 ) && arg3.Unify( 0 );
}

} // namespace builtin

#endif // PROFILE

namespace builtin {

LILFES_BUILTIN_PRED_1(start_prof, start_prof);
LILFES_BUILTIN_PRED_1(stop_prof, stop_prof);
LILFES_BUILTIN_PRED_OVERLOAD_2(stop_prof_2, stop_prof_2, stop_prof);
LILFES_BUILTIN_PRED_2(total_prof, total_prof);
LILFES_BUILTIN_PRED_OVERLOAD_3(total_prof_3, total_prof_3, total_prof);

} // namespace builtin

} // namespace lilfes

