/**
 * @module  = system
 * @cvs     = $Id: system.cpp,v 1.9 2004/09/13 12:47:35 n165 Exp $
 * @copyright = Copyright (c) 1997-1998, Makino Takaki
 * @copyright = You may distribute this file under the terms of the Artistic License.
 * @desc    = System operations
 * Predicates for operating a system
 * @jdesc   = ƥ
 * @japanese =
 * ƥԤʤҸǤ
 * @end_japanese
*/

#include "builtin.h"
#include <string>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif // HAVE_SYS_TIME_H
#ifdef HAVE_TIME_H
#include <time.h>
#endif // HAVE_TIME_H

namespace lilfes {

namespace builtin {

/**
  * @predicate	= getenv(+STR,-ENV) 
  * @desc	= Return content of enviroment variable STR.
  * @param	= +STR/string : name of enviroment variable
  * @param	= -ENV/string : content of enviroment variable
  * @deprecated = use 'environ/2'
  * @example	= 
  * @end_example
  * @jdesc	= Ķѿͤޤ
  * @jparam	= +STR/string : ѿ̾
  * @jparam	= -ENV/string : ѿ
  * @jdeprecated = 'environ/2' ѤƤ
  * @jexample	= 
  * @end_jexample
*/
  bool getenv( machine&, FSP arg1, FSP arg2 ) {
    cell c = arg1.ReadCell();
    if( !IsSTG(c) ) {
      RUNWARN( "getenv/2 requires a string" );
      return false;
    }

    char *s = c2STG(c);
    char *e = ::getenv(s);

    if (e == NULL) {
      return false;
    }
    return arg2.Unify( e );
  }

  LILFES_BUILTIN_PRED_2(getenv, getenv);

/**
  * @predicate	= environ(+STR,-ENV) 
  * @desc	= Return content of enviroment variable STR.
  * @param	= +STR/string : name of enviroment variable
  * @param	= -ENV/string : content of enviroment variable
  * @example	= 
  * @end_example
  * @jdesc	= Ķѿͤޤ
  * @jparam	= +STR/string : ѿ̾
  * @jparam	= -ENV/string : ѿ
  * @jexample	= 
  * @end_jexample
*/
  LILFES_BUILTIN_PRED_2(getenv, environ);

///////////////////////////////////////////////////////////////////////////////

/**
  * @predicate = getdate(-$DATE)
  * @param     = -$DATE/string : current date
  * @desc      = get the current date
  * @deprecated = use 'now/1'
  * You can get the current date with this predicate.
  * @jparam     = -$DATE/string : ߤ
  * @jdesc      = ߤޤ
  * @jdeprecated = 'now/1' ѤƤ
  * @japanese  =
  * ߤ򤢤魯ʸ뤳Ȥޤ
  * @end_japanese
*/

  bool getdate( machine&, FSP arg1 ) {
#if defined( HAVE_GETTIMEOFDAY ) && defined( HAVE_CTIME )
    struct timeval current_time;
    gettimeofday( &current_time, NULL );
    std::string str( ctime( &current_time.tv_sec ) );
    return arg1.Unify( str.substr( 0, str.size() - 1 ).c_str() );
#else // HAVE_GETTIMEOFDAY && HAVE_CTIME
    RUNWARN( "getdata: gettimeofday is not implemented in this system" );
    return arg1.Unify( "" );
#endif // HAVE_GETTIMEOFDAY && HAVE_CTIME
  }

  LILFES_BUILTIN_PRED_1(getdate, getdate);

/**
  * @predicate = now(-$DATE)
  * @param     = -$DATE/string : current date
  * @desc      = get the current date
  * You can get the current date with this predicate.
  * @jparam     = -$DATE/string : ߤ
  * @jdesc      = ߤޤ
  * @japanese  =
  * ߤ򤢤魯ʸ뤳Ȥޤ
  * @end_japanese
*/

  LILFES_BUILTIN_PRED_1(getdate, now);

///////////////////////////////////////////////////////////////////////////////

/**
  * @predicate	= system(+COMMAND) 
  * @desc	= Execute external command
  * @param	= +COMMAND/string : command
  * @example	= 
  * @end_example
  * @jdesc	= ޥɤ¹Ԥޤ
  * @jparam	= +COMMAND/string : ޥ̾
  * @jexample	= 
  * @end_jexample
*/

  bool system( machine&, FSP arg1 )
  {
    cell c1 = arg1.ReadCell();
    if( !IsSTG(c1) ) {
      RUNWARN( "Warning: system/1 requires a string" );
      return false;
    }

    char *s1 = c2STG(c1);
    return ::system(s1) == 0;
  }

  LILFES_BUILTIN_PRED_1(system, system);

}




} // namespace lilfes










